@@ -65,11 +65,13 @@ struct dcm_ib_qp {
 	struct mcm_wrc_info	 wrc;	   /* local WC info */
 	struct mcm_wrc_info	 wrc_rem;  /* remote WR info */
 	DAPL_OS_LOCK		 lock;	   /* Proxy WR and WC queues */
+	uint8_t			 p2p_data; /* Max number of bytes to pass to proxy in the WR */
 	uint8_t			 ep_map;   /* Peer EP mapping, MXS, MSS, HST */
 	uint32_t		 seg_sz;   /* Peer MXS Proxy-in segment size */
 	char			 *wr_buf_rx; /* mcm_wr_rx_t entries, devices without inline data  */
 	struct ibv_mr		 *wr_buf_rx_mr;
 #endif
+	uint8_t			 mtu;      /* RC QP MTU, cm exchange, min(local,peer) */
 };
 
 #define DCM_CQ_TX 		0x1
@@ -149,7 +151,8 @@ typedef struct _ib_cm_msg
 	uint8_t			sportx; /* extend to 24 bits */
 	uint8_t			dportx; /* extend to 24 bits */
 	uint8_t			rtns; 	/* retransmissions */
-	uint8_t			resv[2];
+	uint8_t			mtu;	/* MTU */
+	uint8_t			resv[1];
 	union dcm_addr		saddr;
 	union dcm_addr		daddr;
 	union dcm_addr		saddr_alt;
@@ -167,6 +170,11 @@ typedef struct _ib_named_attr
 	 const char *mtu;
 	 const char *port;
 	 const char *port_num;
+	 const char *iw_ext;
+	 const char *ib_ext;
+	 const char *i_data;
+	 const char *f_add;
+	 const char *c_swap;
 
 } ib_named_attr_t;
 
@@ -242,7 +250,7 @@ typedef uint16_t		ib_hca_port_t;
 #define DCM_ACK_RETRY	7  /* 3 bits, 7 * 4.2  == 30 seconds */
 #define DCM_RNR_TIMER	12 /* 5 bits, 12 =.64ms, 28 =163ms, 31 =491ms */
 #define DCM_RNR_RETRY	7  /* 3 bits, 7 == infinite */
-#define DCM_IB_MTU	2048
+#define DCM_IB_MTU 	4096	/* new default MTU size */
 
 /* Global routing defaults */
 #define DCM_GLOBAL	0       /* global routing is disabled */
