@@ -525,7 +525,9 @@ dapli_post_ext(IN DAT_EP_HANDLE ep_handle,
 	 * requests, which aren't allowed to race with each other.
 	 */
 	dat_status = dapls_dto_cookie_alloc(&ep_ptr->req_buffer,
-					    DAPL_DTO_TYPE_EXTENSION,
+					    op_type == OP_RDMA_WRITE_IMM ?
+						DAPL_DTO_TYPE_EXTENSION_IMM:
+						DAPL_DTO_TYPE_EXTENSION,
 					    user_cookie, &cookie);
 	if (dat_status != DAT_SUCCESS)
 		goto bail;
