@@ -285,7 +285,7 @@ enum ibv_mtu dapl_ib_mtu(int mtu)
 	case 4096:
 		return IBV_MTU_4096;
 	default:
-		return IBV_MTU_1024;
+		return IBV_MTU_4096;
 	}
 }
 
@@ -303,7 +303,7 @@ const char *dapl_ib_mtu_str(enum ibv_mtu mtu)
 	case IBV_MTU_4096:
 		return "4096";
 	default:
-		return "1024";
+		return "4096";
 	}
 }
 
@@ -407,16 +407,30 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 		tp->na.mode = "DIRECT";
 		tp->na.read = "TRUE";
 #endif
-
-	/* query verbs for this device and port attributes */
-	if (ibv_query_device(hca_ptr->ib_hca_handle, &dev_attr) ||
-	    ibv_query_port(hca_ptr->ib_hca_handle,
-			   hca_ptr->port_num, &port_attr))
-		return (dapl_convert_errno(errno, "ib_query_hca"));
+#ifdef _OPENIB_MCM_
+	if (MFO_EP(&hca_ptr->ib_trans.addr) && tp->self.node) {
+		if (dapli_mix_query_device(tp, &dev_attr) ||
+			dapli_mix_query_port(tp, hca_ptr->port_num, &port_attr))
+				return (dapl_convert_errno(errno, "mix_query_hca"));
+	}
+	else
+#endif
+		/* query verbs for this device and port attributes */
+		if (ibv_query_device(hca_ptr->ib_hca_handle, &dev_attr) ||
+				ibv_query_port(hca_ptr->ib_hca_handle,
+						hca_ptr->port_num, &port_attr))
+			return (dapl_convert_errno(errno, "ib_query_hca"));
 
 	dev_attr.max_qp_wr = DAPL_MIN(dev_attr.max_qp_wr,
 				      dapl_os_get_env_val("DAPL_WR_MAX", dev_attr.max_qp_wr));
 
+	/* MTU to active by default, reset if env set and <= active_mtu */
+	if (getenv("DAPL_IB_MTU"))
+		tp->ib_cm.mtu = DAPL_MIN(port_attr.active_mtu,
+					 dapl_ib_mtu(dapl_os_get_env_val("DAPL_IB_MTU", DCM_IB_MTU)));
+	else
+		tp->ib_cm.mtu = port_attr.active_mtu;
+
 #ifdef _OPENIB_MCM_
 	/* Adjust for CCL Proxy; limited sge's, no READ support, reduce QP and RDMA limits */
 	dev_attr.max_sge = DAPL_MIN(dev_attr.max_sge, DAT_MIX_SGE_MAX);
@@ -428,9 +442,16 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 
 	if (ia_attr != NULL) {
 		(void)dapl_os_memzero(ia_attr, sizeof(*ia_attr));
-		strncpy(ia_attr->adapter_name,
-		        ibv_get_device_name(tp->ib_dev),
-		        DAT_NAME_MAX_LENGTH - 1);
+#ifdef _OPENIB_MCM_
+		/* MIC at MFO mode - Initialized at Open */
+		if (MFO_EP(&hca_ptr->ib_trans.addr) && tp->self.node) {
+			strncpy(ia_attr->adapter_name, hca_ptr->name, DAT_NAME_MAX_LENGTH);
+		}
+		else
+#endif
+			strncpy(ia_attr->adapter_name, ibv_get_device_name(tp->ib_dev),
+					DAT_NAME_MAX_LENGTH - 1);
+
 		ia_attr->adapter_name[DAT_NAME_MAX_LENGTH - 1] = '\0';
 		ia_attr->vendor_name[DAT_NAME_MAX_LENGTH - 1] = '\0';
 		ia_attr->ia_address_ptr = (DAT_IA_ADDRESS_PTR) &hca_ptr->hca_address;
@@ -465,23 +486,36 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 		ia_attr->max_message_size = port_attr.max_msg_sz;
 		ia_attr->max_rdma_size = port_attr.max_msg_sz;
 		/* iWARP spec. - 1 sge for RDMA reads */
+		ia_attr->max_iov_segments_per_rdma_read = dev_attr.max_sge;
+#ifdef _OPENIB_MCM_
+		if (!MFO_EP(&hca_ptr->ib_trans.addr))
+#endif
 		if (hca_ptr->ib_hca_handle->device->transport_type == IBV_TRANSPORT_IWARP)
 			ia_attr->max_iov_segments_per_rdma_read = 1;
-		else
-			ia_attr->max_iov_segments_per_rdma_read = dev_attr.max_sge;
 		ia_attr->max_iov_segments_per_rdma_write = dev_attr.max_sge;
 		ia_attr->num_transport_attr = 0;
 		ia_attr->transport_attr = NULL;
 		ia_attr->num_vendor_attr = 0;
 		ia_attr->vendor_attr = NULL;
 #ifdef DAT_EXTENSIONS
-		ia_attr->extension_supported = DAT_EXTENSION_IB;
-		ia_attr->extension_version = DAT_IB_EXTENSION_VERSION;
+		if (hca_ptr->ib_hca_handle->device->transport_type == IBV_TRANSPORT_IWARP) {
+			ia_attr->extension_supported = DAT_EXTENSION_NONE;
+			tp->na.ib_ext = "FALSE";
+			tp->na.i_data = "FALSE";
+			tp->na.f_add  = "FALSE";
+			tp->na.c_swap = "FALSE";
+		} else {
+			ia_attr->extension_supported = DAT_EXTENSION_IB;
+			ia_attr->extension_version = DAT_IB_EXTENSION_VERSION;
+			tp->na.ib_ext = "TRUE";
+			tp->na.i_data = "TRUE";
+			tp->na.f_add  = "TRUE";
+			tp->na.c_swap = "TRUE";
+		}
 #endif
 		/* save key device attributes for CM exchange */
 		tp->ib_cm.rd_atom_in  = dev_attr.max_qp_rd_atom;
 		tp->ib_cm.rd_atom_out = dev_attr.max_qp_init_rd_atom;
-		tp->ib_cm.mtu = DAPL_MIN(port_attr.active_mtu, tp->ib_cm.mtu);
 		tp->ib_cm.ack_timer = DAPL_MAX(dev_attr.local_ca_ack_delay, tp->ib_cm.ack_timer);
 
 		/* set provider/transport specific named attributes */
@@ -489,8 +523,14 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 		tp->na.mtu = dapl_ib_mtu_str(tp->ib_cm.mtu);
 		tp->na.port = dapl_ib_port_str(port_attr.state);
 		tp->na.port_num = dapl_ib_port_num_str(hca_ptr->port_num);
-		if (!tp->guid)
-			tp->guid = ntohll(ibv_get_device_guid(tp->ib_dev));
+		if (!tp->guid) {
+#ifdef _OPENIB_MCM_
+			if (MFO_EP(&hca_ptr->ib_trans.addr) && tp->self.node)
+				tp->guid = ntohll(dev_attr.node_guid);
+			else
+#endif
+				tp->guid = ntohll(ibv_get_device_guid(tp->ib_dev));
+		}
 
 		sprintf(tp->guid_str, "%04x:%04x:%04x:%04x",
 			(unsigned) (tp->guid >> 48) & 0xffff,
@@ -498,6 +538,12 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 			(unsigned) (tp->guid >> 16) & 0xffff,
 			(unsigned) (tp->guid >>  0) & 0xffff);
 
+#ifdef _OPENIB_MCM_
+		/* initialized at open */
+		if (MFO_EP(&hca_ptr->ib_trans.addr) && tp->self.node)
+			goto skip_ib;
+#endif
+
 		if (hca_ptr->ib_hca_handle->device->transport_type != IBV_TRANSPORT_IB)
 			goto skip_ib;
 
@@ -860,28 +906,47 @@ DAT_NAMED_ATTR ib_attrs[] = {
 void dapls_query_provider_specific_attr(IN DAPL_IA * ia_ptr,
 					IN DAT_PROVIDER_ATTR * attr_ptr)
 {
+	int i;
+
 	attr_ptr->num_provider_specific_attr = SPEC_ATTR_SIZE(ib_attrs);
 	attr_ptr->provider_specific_attr = ib_attrs;
 
 	dapl_log(DAPL_DBG_TYPE_UTIL,
 		 " prov_attr: %p sz %d\n", ib_attrs, SPEC_ATTR_SIZE(ib_attrs));
 
-	/* update common attributes from providers */
-	ib_attrs[1].value = ia_ptr->hca_ptr->ib_trans.na.dev;
-	ib_attrs[2].value = ia_ptr->hca_ptr->ib_trans.na.mode;
-	ib_attrs[3].value = ia_ptr->hca_ptr->ib_trans.na.read;
-	ib_attrs[4].value = ia_ptr->hca_ptr->ib_trans.guid_str;
-	ib_attrs[5].value = ia_ptr->hca_ptr->ib_trans.na.mtu;
-	ib_attrs[6].value = ia_ptr->hca_ptr->ib_trans.na.port;
-	ib_attrs[7].value = ia_ptr->hca_ptr->ib_trans.na.port_num;
+	for (i=0; i<SPEC_ATTR_SIZE(ib_attrs); i++) {
+		/* update attributes from IA query results */
+		if (!strcmp(ib_attrs[i].name, "DAT_IB_DEVICE_NAME"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.dev;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_CONNECTIVITY_MODE"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.mode;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_RDMA_READ"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.read;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_NODE_GUID"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.guid_str;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_TRANSPORT_MTU"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.mtu;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_PORT_STATUS"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.port;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_PORT_NUMBER"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.port_num;
+		else if (!strcmp(ib_attrs[i].name, "DAT_EXTENSION_INTERFACE"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.ib_ext;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_IMMED_DATA"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.i_data;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_FETCH_AND_ADD"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.f_add;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_CMP_AND_SWAP"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.na.c_swap;
 #ifdef _OPENIB_MCM_
-{
-	int i = attr_ptr->num_provider_specific_attr;
-	ib_attrs[i-3].value = ia_ptr->hca_ptr->ib_trans.fam_str;
-	ib_attrs[i-2].value = ia_ptr->hca_ptr->ib_trans.mod_str;
-	ib_attrs[i-1].value = ia_ptr->hca_ptr->ib_trans.ver_str;
-}
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_PROXY_CPU_FAMILY"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.fam_str;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_PROXY_CPU_MODEL"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.mod_str;
+		else if (!strcmp(ib_attrs[i].name, "DAT_IB_PROXY_VERSION"))
+			ib_attrs[i].value = ia_ptr->hca_ptr->ib_trans.ver_str;
 #endif
+	}
 }
 
 /*
