@@ -166,9 +166,12 @@ int  mcm_create_pi_cq(struct dcm_ib_qp *m_qp, int len);
 void mcm_destroy_pi_cq(struct dcm_ib_qp *m_qp);
 
 /* MIC eXchange (MIX) operations, mix.c */
+int  dapli_mix_mode(ib_hca_transport_t *tp, char *name);
 int  dapli_mix_open(ib_hca_transport_t *tp, char *name, int port, int query);
 void dapli_mix_close(ib_hca_transport_t *tp);
 int  dapli_mix_get_attr(ib_hca_transport_t *tp, dat_mix_prov_attr_t *pr_attr);
+int  dapli_mix_query_device(ib_hca_transport_t *tp, struct ibv_device_attr *dev_attr);
+int  dapli_mix_query_port(ib_hca_transport_t *tp, unsigned long port_num, struct ibv_port_attr *port_attr);
 int  dapli_mix_listen(dp_ib_cm_handle_t cm, uint16_t sid);
 int  dapli_mix_listen_free(dp_ib_cm_handle_t cm);
 int  dapli_mix_qp_create(ib_qp_handle_t m_qp, struct ibv_qp_init_attr *attr,
@@ -178,6 +181,9 @@ int  dapli_mix_cq_create(ib_cq_handle_t m_cq, int cq_len);
 int  dapli_mix_cq_free(ib_cq_handle_t m_cq);
 int  dapli_mix_cq_wait(ib_cq_handle_t m_cq, int time);
 int  dapli_mix_cq_poll(ib_cq_handle_t m_cq, struct ibv_wc *wc);
+int  dapli_mix_pz_create(DAPL_IA * ia_ptr, DAPL_PZ *m_pz);
+DAT_RETURN dapls_ib_pd_free(IN DAPL_PZ * pz);
+int  dapli_mix_pz_free(DAPL_PZ *m_pz);
 int  dapli_mix_cm_req_out(dp_ib_cm_handle_t m_cm, ib_qp_handle_t m_qp);
 int  dapli_mix_cm_rtu_out(dp_ib_cm_handle_t m_cm);
 void dapli_mix_cm_dreq_out(dp_ib_cm_handle_t m_cm);
