@@ -180,7 +180,7 @@ DAT_RETURN dapls_ib_open_hca(IN IB_HCA_NAME hca_name,
 			     IN DAPL_HCA * hca_ptr,
 			     IN DAPL_OPEN_FLAGS flags)
 {
-	struct ibv_device **dev_list;
+	struct ibv_device **dev_list = NULL;
 	struct dat_mcm_addr *mcm_ia = (struct dat_mcm_addr *) &hca_ptr->hca_address;
 	struct ibv_port_attr port_attr;
 	int i, nd = 0;
@@ -190,33 +190,92 @@ DAT_RETURN dapls_ib_open_hca(IN IB_HCA_NAME hca_name,
 		 PROVIDER_NAME, hca_name, hca_ptr,
 		 flags & DAPL_OPEN_QUERY ? "QUERY MODE":"STD MODE");
 
+	/* set RC tunables via enviroment or default */
+	hca_ptr->ib_trans.ib_cm.ack_retry =
+	    dapl_os_get_env_val("DAPL_ACK_RETRY", DCM_ACK_RETRY);
+	hca_ptr->ib_trans.ib_cm.ack_timer =
+	    dapl_os_get_env_val("DAPL_ACK_TIMER", DCM_ACK_TIMER);
+	hca_ptr->ib_trans.ib_cm.rnr_retry =
+	    dapl_os_get_env_val("DAPL_RNR_RETRY", DCM_RNR_RETRY);
+	hca_ptr->ib_trans.ib_cm.rnr_timer =
+	    dapl_os_get_env_val("DAPL_RNR_TIMER", DCM_RNR_TIMER);
+	hca_ptr->ib_trans.ib_cm.global =
+	    dapl_os_get_env_val("DAPL_GLOBAL_ROUTING", DCM_GLOBAL);
+	hca_ptr->ib_trans.ib_cm.hop_limit =
+	    dapl_os_get_env_val("DAPL_HOP_LIMIT", DCM_HOP_LIMIT);
+	hca_ptr->ib_trans.ib_cm.tclass =
+	    dapl_os_get_env_val("DAPL_TCLASS", DCM_TCLASS);
+	hca_ptr->ib_trans.ib_cm.mtu =
+	    dapl_ib_mtu(dapl_os_get_env_val("DAPL_IB_MTU", DCM_IB_MTU));
+
+	/* initialize CM list, LISTEN, SND queue, PSP array, locks */
+	if ((dapl_os_lock_init(&hca_ptr->ib_trans.lock)) != DAT_SUCCESS)
+		goto err;
+
+	if ((dapl_os_lock_init(&hca_ptr->ib_trans.llock)) != DAT_SUCCESS)
+		goto err;
+
+	if ((dapl_os_lock_init(&hca_ptr->ib_trans.slock)) != DAT_SUCCESS)
+		goto err;
+
+	if ((dapl_os_lock_init(&hca_ptr->ib_trans.plock)) != DAT_SUCCESS)
+		goto err;
+
+	if ((dapl_os_lock_init(&hca_ptr->ib_trans.cqlock)) != DAT_SUCCESS)
+		goto err;
+
+	/* initialize CM and listen lists on this HCA uCM QP */
+	dapl_llist_init_head(&hca_ptr->ib_trans.list);
+	dapl_llist_init_head(&hca_ptr->ib_trans.llist);
+	dapl_llist_init_head(&hca_ptr->ib_trans.cqlist);
+
+	/* Get MIC mode and EP mappings */
+	if (dapli_mix_mode(&hca_ptr->ib_trans, hca_name))
+		goto err;
+
+	/* MIC EP with Full offload, no local verbs device */
+	if (MFO_EP(&hca_ptr->ib_trans.addr)) {
+		if (dapli_mix_open(&hca_ptr->ib_trans, hca_name,
+				   hca_ptr->port_num,
+				   flags & DAPL_OPEN_QUERY)) {
+			dapl_log(DAPL_DBG_TYPE_ERR,
+				" open_hca: SCIF init ERR on %s\n", hca_name);
+			goto err;
+		}
+		hca_ptr->ib_hca_handle = hca_ptr->ib_trans.ib_ctx;
+		goto cm_init;
+	}
+
 	/* Get list of all IB devices, find match, open */
 	dev_list = ibv_get_device_list(&nd);
 	if (!dev_list) {
 		dapl_dbg_log(DAPL_DBG_TYPE_ERR,
 			     " open_hca: ibv_get_device_list() failed\n",
 			     hca_name);
-		return DAT_INTERNAL_ERROR;
+		goto err;
 	}
 	dapl_log(DAPL_DBG_TYPE_UTIL, " open_hca %p: %d devices found\n", hca_ptr, nd);
+	hca_ptr->ib_trans.ib_dev = NULL;
 	for (i = 0; i < nd; ++i) {
-		if (!strcmp(dev_list[i]->name, hca_name)) {
+		if (!hca_ptr->ib_trans.guid &&
+		    dev_list[i]->transport_type == IBV_TRANSPORT_IB)
+			hca_ptr->ib_trans.guid = ibv_get_device_guid(dev_list[i]);
+
+		if (!strcmp(dev_list[i]->name, hca_name))
 			hca_ptr->ib_trans.ib_dev = dev_list[i];
-			goto found;
-		}
+	}
+	if (hca_ptr->ib_trans.ib_dev == NULL) {
+		dapl_log(DAPL_DBG_TYPE_ERR, " open_hca: device %s not found\n", hca_name);
+		dat_status = DAT_PROVIDER_NOT_FOUND;
+		goto err2;
 	}
 
-	dapl_log(DAPL_DBG_TYPE_ERR, " open_hca: device %s not found\n", hca_name);
-	dat_status = DAT_PROVIDER_NOT_FOUND;
-	goto err;
-
-found:
 	hca_ptr->ib_hca_handle = ibv_open_device(hca_ptr->ib_trans.ib_dev);
 	if (!hca_ptr->ib_hca_handle) {
 		dapl_log(DAPL_DBG_TYPE_ERR,
 			 " open_hca: dev open failed for %s\n",
 			 ibv_get_device_name(hca_ptr->ib_trans.ib_dev));
-		goto err;
+		goto err2;
 	}
 	hca_ptr->ib_trans.ib_ctx = hca_ptr->ib_hca_handle;
 	dapls_config_verbs(hca_ptr->ib_hca_handle);
@@ -229,11 +288,11 @@ found:
 			 ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
 			 strerror(errno));
 		dat_status = DAT_INVALID_ADDRESS;
-		goto bail;
+		goto err2;
 	} else {
 		if (port_attr.state != IBV_PORT_ACTIVE) {
 			dat_status = DAT_INVALID_ADDRESS;
-			goto bail;
+			goto err2;
 		}
 		hca_ptr->ib_trans.addr.lid = htons(port_attr.lid);
 		hca_ptr->ib_trans.lid = htons(port_attr.lid);
@@ -248,55 +307,20 @@ found:
 			 ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
 			 strerror(errno));
 		dat_status = DAT_INVALID_ADDRESS;
-		goto bail;
+		goto err2;
 	}
 
-	/* set RC tunables via enviroment or default */
-	hca_ptr->ib_trans.max_inline_send =
-	    dapl_os_get_env_val("DAPL_MAX_INLINE", INLINE_SEND_IB_DEFAULT);
-	hca_ptr->ib_trans.ack_retry =
-	    dapl_os_get_env_val("DAPL_ACK_RETRY", DCM_ACK_RETRY);
-	hca_ptr->ib_trans.ack_timer =
-	    dapl_os_get_env_val("DAPL_ACK_TIMER", DCM_ACK_TIMER);
-	hca_ptr->ib_trans.rnr_retry =
-	    dapl_os_get_env_val("DAPL_RNR_RETRY", DCM_RNR_RETRY);
-	hca_ptr->ib_trans.rnr_timer =
-	    dapl_os_get_env_val("DAPL_RNR_TIMER", DCM_RNR_TIMER);
-	hca_ptr->ib_trans.global =
-	    dapl_os_get_env_val("DAPL_GLOBAL_ROUTING", DCM_GLOBAL);
-	hca_ptr->ib_trans.hop_limit =
-	    dapl_os_get_env_val("DAPL_HOP_LIMIT", DCM_HOP_LIMIT);
-	hca_ptr->ib_trans.tclass =
-	    dapl_os_get_env_val("DAPL_TCLASS", DCM_TCLASS);
-	hca_ptr->ib_trans.mtu =
-	    dapl_ib_mtu(dapl_os_get_env_val("DAPL_IB_MTU", DCM_IB_MTU));
-
-	if (dapli_mix_open(&hca_ptr->ib_trans, hca_name,
-			   hca_ptr->port_num, flags & DAPL_OPEN_QUERY)) {
-		dapl_log(DAPL_DBG_TYPE_ERR,
-			 " open_hca: SCIF init ERR for %s\n",
-			 ibv_get_device_name(hca_ptr->ib_trans.ib_dev));
-		goto bail;
+	/* max inline value when running on MIC is returned from host */
+	if (dapl_ib_inline_data(hca_ptr->ib_hca_handle)) {
+		hca_ptr->ib_trans.ib_cm.max_inline =
+			dapl_os_get_env_val("DAPL_MAX_INLINE",
+					    sizeof(mcm_wr_rx_t));
 	}
+	else
+		hca_ptr->ib_trans.ib_cm.max_inline = 0;
 
-	if (flags & DAPL_OPEN_QUERY)
-		goto done;
-
-	/* initialize CM list, LISTEN, SND queue, PSP array, locks */
-	if ((dapl_os_lock_init(&hca_ptr->ib_trans.lock)) != DAT_SUCCESS)
-		goto bail;
-	
-	if ((dapl_os_lock_init(&hca_ptr->ib_trans.llock)) != DAT_SUCCESS)
-		goto bail;
-	
-	if ((dapl_os_lock_init(&hca_ptr->ib_trans.slock)) != DAT_SUCCESS)
-		goto bail;
-
-	if ((dapl_os_lock_init(&hca_ptr->ib_trans.plock)) != DAT_SUCCESS)
-		goto bail;
-
-	if ((dapl_os_lock_init(&hca_ptr->ib_trans.cqlock)) != DAT_SUCCESS)
-		goto bail;
+	dapl_log(DAPL_DBG_TYPE_UTIL, " open_hca: max inline data set to %d\n",
+		hca_ptr->ib_trans.ib_cm.max_inline);
 
 	/* EVD events without direct CQ channels, CNO support */
 	hca_ptr->ib_trans.ib_cq =
@@ -305,7 +329,7 @@ found:
 		dapl_log(DAPL_DBG_TYPE_ERR,
 			 " open_hca: ibv_create_comp_channel ERR %s\n",
 			 strerror(errno));
-		goto bail;
+		goto err2;
 	}
 	dapls_config_comp_channel(hca_ptr->ib_trans.ib_cq);
 
@@ -315,23 +339,29 @@ found:
 		dapl_log(DAPL_DBG_TYPE_ERR,
 			 " open_hca: ERR: create_empty_cq = %s\n",
 			 strerror(errno));
-		goto bail;
+		goto err2;
 	}
 
-	/* initialize CM and listen lists on this HCA uCM QP */
-	dapl_llist_init_head(&hca_ptr->ib_trans.list);
-	dapl_llist_init_head(&hca_ptr->ib_trans.llist);
-	dapl_llist_init_head(&hca_ptr->ib_trans.cqlist);
+	if (dapli_mix_open(&hca_ptr->ib_trans, hca_name,
+			   hca_ptr->port_num, flags & DAPL_OPEN_QUERY)) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 " open_hca: SCIF init ERR for %s\n", hca_name);
+		goto err2;
+	}
 
-	/* create uCM qp services */
+cm_init:
+	if (flags & DAPL_OPEN_QUERY)
+		goto done;
+
+	/* create MCM qp services */
 	if (mcm_service_create(hca_ptr))
-		goto bail;
+		goto err2;
 
 	if (create_os_signal(hca_ptr)) {
 		dapl_log(DAPL_DBG_TYPE_ERR,
 			 " open_hca: failed to init cr pipe - %s\n",
 			 strerror(errno));
-		goto bail;
+		goto err2;
 	}
 
 	/* create thread to process inbound connect request */
@@ -342,13 +372,12 @@ found:
 	if (dat_status != DAT_SUCCESS) {
 		dapl_log(DAPL_DBG_TYPE_ERR,
 			 " open_hca: failed to create thread\n");
-		goto bail;
+		goto err2;
 	}
 
 	dapl_log(DAPL_DBG_TYPE_UTIL,
-		     " open_hca: MCM devname %s port %d, dev_IP %s ep_map %s\n",
-		     ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
-		     hca_ptr->port_num, 
+		     " open_hca: devname %s port %d, dev_IP %s ep_map %s\n",
+		     hca_name, hca_ptr->port_num,
 		     inet_ntoa(((struct sockaddr_in *)
 			       &hca_ptr->hca_address)->sin_addr),
 		     mcm_map_str(hca_ptr->ib_trans.addr.ep_map));
@@ -358,33 +387,39 @@ found:
 		dapl_os_sleep_usec(1000);
 
 done:
+	if (dev_list)
+		ibv_free_device_list(dev_list);
+
 	/* save LID, GID, QPN, PORT address information, for ia_queries */
 	/* Set AF_INET6 to insure callee address storage of 28 bytes */
 	hca_ptr->ib_trans.hca = hca_ptr;
 	hca_ptr->ib_trans.addr.family = AF_INET6;
 	hca_ptr->ib_trans.addr.qp_type = IBV_QPT_UD;
-	memcpy(&hca_ptr->hca_address, 
-	       &hca_ptr->ib_trans.addr, 
+	memcpy(&hca_ptr->hca_address,
+	       &hca_ptr->ib_trans.addr,
 	       sizeof(struct dat_mcm_addr));
 
-	dapl_dbg_log(DAPL_DBG_TYPE_UTIL,
+	dapl_log(DAPL_DBG_TYPE_UTIL,
 		 "%s open: dev %s port %d, GID %s, LID %x qpn %x sl %d %s\n",
 		 PROVIDER_NAME, hca_name, hca_ptr->port_num,
 		 inet_ntop(AF_INET6, &mcm_ia->gid, gid_str, sizeof(gid_str)),
 		 ntohs(mcm_ia->lid), ntohl(mcm_ia->qpn),
 		 mcm_ia->sl, mcm_map_str(mcm_ia->ep_map));
 
-	ibv_free_device_list(dev_list);
 	return DAT_SUCCESS;
-bail:
+
+err2:
 	mcm_service_destroy(hca_ptr);
-	ibv_close_device(hca_ptr->ib_hca_handle);
+	if (dev_list) {
+		if (hca_ptr->ib_hca_handle)
+			ibv_close_device(hca_ptr->ib_hca_handle);
+		ibv_free_device_list(dev_list);
+	}
+err:
 	hca_ptr->ib_hca_handle = IB_INVALID_HANDLE;
 	hca_ptr->ib_trans.ib_ctx = NULL;
 	hca_ptr->ib_trans.ib_dev = NULL;
-
-err:
-	ibv_free_device_list(dev_list);
+	hca_ptr->ib_trans.hca = NULL;
 	return dat_status;
 }
 
@@ -441,8 +476,10 @@ done:
 	}
 
 	if (hca_ptr->ib_hca_handle != IB_INVALID_HANDLE) {
-		if (ibv_close_device(hca_ptr->ib_hca_handle))
-			return (dapl_convert_errno(errno, "ib_close_device"));
+		if (!MFO_EP(&hca_ptr->ib_trans.addr) || hca_ptr->ib_trans.self.node == 0) {
+			if (ibv_close_device(hca_ptr->ib_hca_handle))
+				return (dapl_convert_errno(errno, "ib_close_device"));
+		}
 		hca_ptr->ib_hca_handle = IB_INVALID_HANDLE;
 	}
 
@@ -500,10 +537,10 @@ static void mcm_service_destroy(IN DAPL_HCA *hca)
 
 static int mcm_service_create(IN DAPL_HCA *hca)
 {
-        struct ibv_qp_init_attr qp_create;
+	struct ibv_qp_init_attr qp_create;
 	ib_hca_transport_t *tp = &hca->ib_trans;
 	struct ibv_recv_wr recv_wr, *recv_err;
-        struct ibv_sge sge;
+	struct ibv_sge sge;
 	int i, mlen = sizeof(dat_mcm_msg_t);
 	int hlen = sizeof(struct ibv_grh); /* hdr included with UD recv */
 	char *rbuf;
@@ -524,12 +561,12 @@ static int mcm_service_create(IN DAPL_HCA *hca)
 		return 0;
 
 	tp->pd = ibv_alloc_pd(hca->ib_hca_handle);
-        if (!tp->pd) 
-                goto bail;
-        
-        dapl_log(DAPL_DBG_TYPE_UTIL,
-                 " create_service: pd %p ctx %p handle 0x%x\n",
-                  tp->pd, tp->pd->context, tp->pd->handle);
+	if (!tp->pd) 
+		goto bail;
+	
+	dapl_log(DAPL_DBG_TYPE_UTIL,
+		 " create_service: pd %p ctx %p handle 0x%x\n",
+		  tp->pd, tp->pd->context, tp->pd->handle);
 
     	tp->rch = ibv_create_comp_channel(hca->ib_hca_handle);
 	if (!tp->rch) 
@@ -540,7 +577,7 @@ static int mcm_service_create(IN DAPL_HCA *hca)
 	tp->scq = ibv_create_cq(hca->ib_hca_handle, tp->cqe, hca, NULL, 0);
 	if (!tp->scq) 
 		goto bail;
-        
+	
 	tp->rcq = ibv_create_cq(hca->ib_hca_handle, tp->cqe, hca, tp->rch, 0);
 	if (!tp->rcq) 
 		goto bail;
@@ -554,12 +591,12 @@ static int mcm_service_create(IN DAPL_HCA *hca)
 	qp_create.recv_cq = tp->rcq;
 	qp_create.cap.max_send_wr = qp_create.cap.max_recv_wr = tp->qpe;
 	qp_create.cap.max_send_sge = qp_create.cap.max_recv_sge = 1;
-	qp_create.cap.max_inline_data = tp->max_inline_send;
+	qp_create.cap.max_inline_data = tp->ib_cm.max_inline;
 	qp_create.qp_context = (void *)hca;
 
 	tp->qp = ibv_create_qp(tp->pd, &qp_create);
 	if (!tp->qp) 
-                goto bail;
+		goto bail;
 
 	tp->ah = (ib_ah_handle_t*) dapl_os_alloc(sizeof(ib_ah_handle_t) * 0xffff);
 	tp->sid = (uint8_t*) dapl_os_alloc(sizeof(uint8_t) * 0xffff);
@@ -610,7 +647,7 @@ static int mcm_service_create(IN DAPL_HCA *hca)
 
 	/* save qp_num as part of ia_address, network order */
 	tp->addr.qpn = htonl(tp->qp->qp_num);
-        return 0;
+	return 0;
 bail:
 	dapl_log(DAPL_DBG_TYPE_ERR,
 		 " ucm_create_services: ERR %s\n", strerror(errno));
