@@ -35,8 +35,42 @@
 #include "dapl_osd.h"
 
 /*
- * CM proxy services, MCM on MIC to MPXYD via SCIF
- *
+ * MCM Provider Proxy services, MIC to MPXYD via SCIF or HST with SCIF
+ */
+
+/*
+ * MIX_IA_MODE
+ */
+
+int dapli_mix_mode(ib_hca_transport_t *tp, char *name)
+{
+	int ret, mfo_dev, mfo_mode;
+
+	mfo_mode = dapl_os_get_env_val("DAPL_MCM_MFO", 0); /* Force MIC Full Offload */
+
+	ret = scif_get_nodeIDs(NULL, 0, &tp->self.node);
+	if (ret < 0) {
+		dapl_log(1, " scif_get_nodeIDs() failed with error %s\n", strerror(errno));
+		return -1;
+	}
+	dapl_log(DAPL_DBG_TYPE_EXTENSION,
+		 " SCIF node_id: %d client req_id 0x%x, %s\n",
+		 (uint16_t)tp->self.node, dapl_os_getpid(), name);
+
+	if (tp->self.node == 0) {
+		tp->addr.ep_map = HOST_SOCK_DEV;  /* non-MIC mapping */
+		return 0;
+	}
+
+	/*  MIC node: "qib" and "hfi" devices requires full offload */
+	mfo_dev = !dapl_os_pstrcmp("qib", name) || !dapl_os_pstrcmp("hfi", name);
+	if (mfo_mode || mfo_dev) {
+		tp->addr.ep_map = MIC_FULL_DEV; /* MIC with full proxy offload, no direct verbs */
+	}
+	return 0;
+}
+
+/*
  * MIX_IA_OPEN
  */
 int dapli_mix_open(ib_hca_transport_t *tp, char *name, int port, int query_only)
@@ -52,24 +86,17 @@ int dapli_mix_open(ib_hca_transport_t *tp, char *name, int port, int query_only)
 	always_proxy = dapl_os_get_env_val("DAPL_MCM_ALWAYS_PROXY", 0);
 	scif_port_id = dapl_os_get_env_val("DAPL_MCM_PORT_ID", SCIF_OFED_PORT_8);
 
-	ret = scif_get_nodeIDs(NULL, 0, &tp->self.node);
-	if (ret < 0) {
-		dapl_log(1, " scif_get_nodeIDs() failed with error %s\n", strerror(errno));
-		return -1;
-	}
-	dapl_log(DAPL_DBG_TYPE_EXTENSION,
-		 " SCIF node_id: %d client req_id 0x%x\n",
-		 (uint16_t)tp->self.node, dapl_os_getpid());
-
-	if (tp->self.node == 0)
-		tp->addr.ep_map = HOST_SOCK_DEV;  /* non-MIC mapping */
-
-	if (query_only || (tp->self.node == 0 && !always_proxy)){
-		dapl_log(DAPL_DBG_TYPE_EXTENSION," Not running on MIC, no MPXY connect required\n");
+	if ((query_only && !MFO_EP(&tp->addr)) ||
+	    (tp->self.node == 0 && !always_proxy)) {
+		dapl_log(DAPL_DBG_TYPE_EXTENSION,
+			" %s, no MPXYD connect required\n",
+			query_only ? "Query only,":"Host node,");
 		tp->scif_ep = 0;
 		return 0;
 	}
-	dapl_log(DAPL_DBG_TYPE_EXTENSION," Running on MIC, MPXY connect required\n");
+	dapl_log(DAPL_DBG_TYPE_EXTENSION,
+		 " Running on MIC at %s ep_map, MPXY connect required\n",
+		 mcm_map_str(tp->addr.ep_map));
 
 	/* Create an endpoint for MPXYD to connect back */
 	listen_ep = scif_open();
@@ -144,6 +171,9 @@ int dapli_mix_open(ib_hca_transport_t *tp, char *name, int port, int query_only)
 	msg.port = port;
 	strcpy((char*)&msg.name, name);
 
+	if (getenv("DAPL_IB_MTU"))
+		msg.hdr.flags |= MIX_OP_MTU;
+
 	/* send any overridden attributes to proxy */
 	msg.dev_attr.ack_timer = tp->ib_cm.ack_timer;
 	msg.dev_attr.ack_retry = tp->ib_cm.ack_retry;
@@ -159,6 +189,9 @@ int dapli_mix_open(ib_hca_transport_t *tp, char *name, int port, int query_only)
 	msg.dev_attr.pkey_idx = tp->ib_cm.pkey_idx;
 	msg.dev_attr.pkey = tp->ib_cm.pkey;
 	msg.dev_attr.max_inline = tp->ib_cm.max_inline;
+	msg.dev_addr.ep_map = tp->addr.ep_map;
+
+	memcpy(&msg.dev_addr, &tp->addr, sizeof(dat_mcm_addr_t));
 
 	len = sizeof(dat_mix_open_t);
 	ret = scif_send(tp->scif_ep, &msg, len, SCIF_SEND_BLOCK);
@@ -199,19 +232,22 @@ int dapli_mix_open(ib_hca_transport_t *tp, char *name, int port, int query_only)
 	tp->ib_cm.hop_limit = msg.dev_attr.hop_limit;
 	tp->ib_cm.tclass = msg.dev_attr.tclass;
 	tp->ib_cm.sl = msg.dev_attr.sl;
-	tp->ib_cm.mtu = msg.dev_attr.mtu;
 	tp->ib_cm.rd_atom_in = msg.dev_attr.rd_atom_in;
 	tp->ib_cm.rd_atom_out = msg.dev_attr.rd_atom_out;
 	tp->ib_cm.pkey_idx = msg.dev_attr.pkey_idx;
 	tp->ib_cm.pkey = msg.dev_attr.pkey;
 	tp->ib_cm.max_inline = msg.dev_attr.max_inline;
-
 	tp->dev_id = msg.hdr.req_id;
 
+	if (MFO_EP(&tp->addr))
+		/* We do not use this var in MFO, but use it as a flag to signal success */
+		tp->ib_ctx = (struct ibv_context *)0xdeadbeef;
+
 	dapl_log(DAPL_DBG_TYPE_EXTENSION,
-		 " mix_open reply (msg %p, ln %d) EPs %d %d %d - dev_id %d\n",
+		 " mix_open reply (msg %p, ln %d) EPs %d %d %d - dev_id %d lid 0x%x\n",
 		 &msg, len, tp->scif_ep, tp->scif_ev_ep,
-		 tp->scif_tx_ep, tp->dev_id);
+		 tp->scif_tx_ep, tp->dev_id, ntohs(tp->addr.lid));
+
 	return 0;
 }
 
@@ -236,6 +272,99 @@ void dapli_mix_close(ib_hca_transport_t *tp)
 	}
 }
 
+/* MIX device ATTR */
+int dapli_mix_query_device(ib_hca_transport_t *tp, struct ibv_device_attr *dev_attr)
+{
+	dat_mix_device_attr_t msg;
+	scif_epd_t mix_ep = tp->scif_ep;
+	int ret, len;
+
+	if (!mix_ep)
+		return 0;
+
+	dapl_log(DAPL_DBG_TYPE_EXTENSION, " MIX_QUERY_DEVICE_ATTR tp = %p\n", tp);
+
+	/* get attr request */
+	msg.hdr.ver = DAT_MIX_VER;
+	msg.hdr.op = MIX_QUERY_DEVICE;
+	msg.hdr.status = 0;
+	msg.hdr.flags = MIX_OP_REQ;
+
+	len = sizeof(dat_mix_hdr_t);
+	ret = scif_send(mix_ep, &msg, len, SCIF_SEND_BLOCK);
+	if (ret != len) {
+		dapl_log(1, " ERR: %s msg %p send on %d, ret %d, exp %d, error %s\n",
+			 mix_op_str(msg.hdr.op), &msg, mix_ep, ret, len, strerror(errno));
+		return -1;
+	}
+	dapl_log(DAPL_DBG_TYPE_EXTENSION," Sent %s request on SCIF EP %d\n", mix_op_str(msg.hdr.op), mix_ep);
+
+	/* get device attr response */
+	len = sizeof(dat_mix_device_attr_t);
+	ret = scif_recv(mix_ep, &msg, len, SCIF_RECV_BLOCK);
+	if (ret != len) {
+		dapl_log(1, " ERR: rcv on new_ep %d, ret %d, exp %d, error %s\n", mix_ep, ret, len, strerror(errno));
+		return -1;
+	}
+	dapl_log(DAPL_DBG_TYPE_EXTENSION," Recv'd %s reply on SCIF EP %d for dev_id %d\n",
+		 mix_op_str(msg.hdr.op), mix_ep, msg.hdr.req_id);
+
+	if (msg.hdr.ver != DAT_MIX_VER || msg.hdr.op != MIX_QUERY_DEVICE ||
+	    msg.hdr.flags != MIX_OP_RSP || msg.hdr.status != MIX_SUCCESS) {
+		dapl_log(1, " ERR: MIX_QUERY_DEVICE ver %d, op %s, flgs %d, st %d dev_id %d\n",
+			     msg.hdr.ver, mix_op_str(msg.hdr.op),
+			     msg.hdr.flags, msg.hdr.status, msg.hdr.req_id);
+		if (msg.hdr.status != MIX_SUCCESS)
+			return msg.hdr.status;
+		else
+			return -1;
+	}
+
+	strncpy(dev_attr->fw_ver, msg.fw_ver, sizeof(dev_attr->fw_ver));
+	dev_attr->node_guid = msg.node_guid;
+	dev_attr->sys_image_guid = msg.sys_image_guid;
+	dev_attr->max_mr_size = msg.max_mr_size;
+	dev_attr->page_size_cap = msg.page_size_cap;
+	dev_attr->vendor_id = msg.vendor_id;
+	dev_attr->vendor_part_id = msg.vendor_part_id;
+	dev_attr->hw_ver = msg.hw_ver;
+	dev_attr->max_qp = msg.max_qp;
+	dev_attr->max_qp_wr = msg.max_qp_wr;
+	dev_attr->device_cap_flags = msg.device_cap_flags;
+	dev_attr->max_sge = msg.max_sge;
+	dev_attr->max_sge_rd = msg.max_sge_rd;
+	dev_attr->max_cq = msg.max_cq;
+	dev_attr->max_cqe = msg.max_cqe;
+	dev_attr->max_mr = msg.max_mr;
+	dev_attr->max_pd = msg.max_pd;
+	dev_attr->max_qp_rd_atom = msg.max_qp_rd_atom;
+	dev_attr->max_ee_rd_atom = msg.max_ee_rd_atom;
+	dev_attr->max_res_rd_atom = msg.max_ee_rd_atom;
+	dev_attr->max_qp_init_rd_atom = msg.max_qp_init_rd_atom;
+	dev_attr->max_ee_init_rd_atom = msg.max_ee_init_rd_atom;
+	dev_attr->atomic_cap = msg.atomic_cap;
+	dev_attr->max_ee = msg.max_ee;
+	dev_attr->max_rdd = msg.max_rdd;
+	dev_attr->max_mw = msg.max_mw;
+	dev_attr->max_raw_ipv6_qp = msg.max_raw_ipv6_qp;
+	dev_attr->max_raw_ethy_qp = msg.max_raw_ethy_qp;
+	dev_attr->max_mcast_grp = msg.max_mcast_grp;
+	dev_attr->max_mcast_qp_attach = msg.max_mcast_qp_attach;
+	dev_attr->max_total_mcast_qp_attach = msg.max_total_mcast_qp_attach;
+	dev_attr->max_ah = msg.max_ah;
+	dev_attr->max_fmr = msg.max_fmr;
+	dev_attr->max_map_per_fmr = msg.max_map_per_fmr;
+	dev_attr->max_srq = msg.max_srq;
+	dev_attr->max_srq_wr = msg.max_srq_wr;
+	dev_attr->max_srq_sge = msg.max_srq_sge;
+	dev_attr->max_pkeys = msg.max_pkeys;
+	dev_attr->local_ca_ack_delay = msg.local_ca_ack_delay;
+	dev_attr->phys_port_cnt = msg.phys_port_cnt;
+
+	dapl_log(DAPL_DBG_TYPE_EXTENSION," MIX_QUERY_DEVICE successful on SCIF EP %d\n", mix_ep);
+	return 0;
+}
+
 /* MIX_PROV_ATTR */
 int dapli_mix_get_attr(ib_hca_transport_t *tp, dat_mix_prov_attr_t *pr_attr)
 {
@@ -243,7 +372,10 @@ int dapli_mix_get_attr(ib_hca_transport_t *tp, dat_mix_prov_attr_t *pr_attr)
 	scif_epd_t mix_ep = tp->scif_ep;
 	int ret, len;
 
-	dapl_log(DAPL_DBG_TYPE_EXTENSION, " MIX_GET_ATTR tp = %p\n", tp);
+	if (!mix_ep)
+		return 0;
+
+	dapl_log(DAPL_DBG_TYPE_EXTENSION, " MIX_QUERY_PROV_ATTR tp = %p\n", tp);
 
 	/* get attr request */
 	msg.hdr.ver = DAT_MIX_VER;
@@ -283,6 +415,12 @@ int dapli_mix_get_attr(ib_hca_transport_t *tp, dat_mix_prov_attr_t *pr_attr)
 
 	memcpy(pr_attr, &msg.attr, sizeof(dat_mix_prov_attr_t));
 
+	/* update local TP CM attributes */
+	tp->retries = pr_attr->cm_retry;
+	tp->rep_time = pr_attr->cm_rep_time_ms;
+	tp->rtu_time = pr_attr->cm_rtu_time_ms;
+	tp->cm_timer = DAPL_MIN(tp->rep_time, tp->rtu_time);
+
 	dapl_log(DAPL_DBG_TYPE_EXTENSION," MIX_PROV_ATTR successful on SCIF EP %d\n", mix_ep);
 	return 0;
 }
@@ -390,9 +528,16 @@ int dapli_mix_mr_create(ib_hca_transport_t *tp, DAPL_LMR * lmr)
 	scif_epd_t mix_ep = tp->scif_ep;
 	int ret, len;
 
-	dapl_log(DAPL_DBG_TYPE_EXTENSION," lmr create %p, addr %p %p rmr_context %x mr->rkey %x\n",
-		 lmr, lmr->mr_handle->addr, lmr->param.registered_address,
-		 lmr->param.rmr_context, lmr->mr_handle->rkey );
+	dapl_log(DAPL_DBG_TYPE_EXTENSION," lmr create %p, addr %p rmr_context %x\n",
+		 lmr, lmr->param.registered_address, lmr->param.rmr_context);
+
+	if (MFO_EP(&tp->addr)) {
+		lmr->mr_handle = (ib_mr_handle_t) dapl_os_alloc (sizeof(struct ibv_mr));
+		if (NULL == lmr->mr_handle) {
+			dapl_log(1, " ERR: Could not allocat mr_hadle\n");
+			return -1;
+		}
+	}
 
 	/* request: */
 	msg.hdr.ver = DAT_MIX_VER;
@@ -404,7 +549,10 @@ int dapli_mix_mr_create(ib_hca_transport_t *tp, DAPL_LMR * lmr)
 	msg.mr_len = lmr->param.registered_size;
 	msg.sci_addr = lmr->sci_addr;
 	msg.sci_off = lmr->sci_off;
-	msg.ib_addr = (uint64_t) lmr->mr_handle->addr;
+	if (MFO_EP(&tp->addr))
+		msg.ib_addr = (uint64_t) lmr->param.registered_address;
+	else
+		msg.ib_addr = (uint64_t) lmr->mr_handle->addr;
 	msg.ib_rkey = lmr->param.rmr_context;
 	msg.ctx = (uint64_t)lmr;
 
@@ -476,6 +624,10 @@ int dapli_mix_mr_free(ib_hca_transport_t *tp, DAPL_LMR * lmr)
 			 msg.hdr.ver, msg.hdr.op, msg.hdr.flags, msg.hdr.status);
 		return -1;
 	}
+	if (MFO_EP(&tp->addr) && lmr->mr_handle) {
+		dapl_os_free(lmr->mr_handle, sizeof(struct ibv_mr));
+		lmr->mr_handle = IB_INVALID_HANDLE;
+	}
 	dapl_log(DAPL_DBG_TYPE_EXTENSION," removed lmr %p, id %d\n", lmr, lmr->mr_id);
 	return 0;
 }
@@ -705,6 +857,118 @@ int dapli_mix_cq_poll(ib_cq_handle_t m_cq, struct ibv_wc *wc)
 	return 0;
 }
 
+/*  MIX_PZ_CREATE */
+int dapli_mix_pz_create(DAPL_IA * ia_ptr, DAPL_PZ *m_pz)
+{
+	dat_mix_pz_t msg;
+	scif_epd_t mix_ep = ia_ptr->hca_ptr->ib_trans.scif_ep;
+	int ret, len;
+
+	m_pz->pd_handle = IB_INVALID_HANDLE;
+
+	/* request: QP_r local, QP_t shadowed */
+	msg.hdr.ver = DAT_MIX_VER;
+	msg.hdr.op = MIX_PZ_CREATE;
+	msg.hdr.status = 0;
+	msg.hdr.flags = MIX_OP_REQ;
+	msg.ctx = (uint64_t)m_pz;
+	msg.ib_pd = 0;
+
+	len = sizeof(dat_mix_pz_t);
+	ret = scif_send(mix_ep, &msg, len, SCIF_SEND_BLOCK);
+	if (ret != len) {
+		dapl_log(1, " ERR: %s snd on %d, ret %d, exp %d, err %s\n",
+			 mix_op_str(msg.hdr.op), mix_ep, ret, len,
+			 strerror(errno));
+		return -1;
+	}
+	dapl_log(DAPL_DBG_TYPE_EXTENSION," Sent %s request on SCIF EP\n",
+		 mix_op_str(msg.hdr.op));
+
+	/* wait for response */
+	msg.ctx = 0;
+	ret = scif_recv(mix_ep, &msg, len, SCIF_RECV_BLOCK);
+	if (ret != len) {
+		dapl_log(1, " ERR: rcv on ep %d, ret %d, exp %d, err %s\n",
+			    mix_ep, ret, len, strerror(errno));
+		return -1;
+	}
+	if (msg.hdr.ver != DAT_MIX_VER || msg.hdr.op != MIX_PZ_CREATE ||
+	    msg.hdr.flags != MIX_OP_RSP || msg.hdr.status != MIX_SUCCESS) {
+		dapl_log(1, " ERR: %s %p ver %d, op %d, flags %d, stat %d\n",
+			    mix_op_str(msg.hdr.op), m_pz, msg.hdr.ver,
+			    msg.hdr.op, msg.hdr.flags, msg.hdr.status);
+		return -1;
+	}
+
+	if (msg.ctx != (uint64_t)m_pz) {
+		dapl_log(1, " ERR: response ctx (0x%x) != sent one (0x%x)\n",
+				msg.ctx, (uint64_t)m_pz);
+		return -1;
+	}
+
+	/* save id from proxy PZ create */
+	m_pz->pd_handle = (ib_pd_handle_t)msg.ib_pd;
+
+	dapl_log(DAPL_DBG_TYPE_EXTENSION,
+		 " MIX_PZ_CREATE: pz %p, reply, proxy IB_PD %p\n",
+		 m_pz, m_pz->pd_handle);
+	return 0;
+}
+
+/* MIX_CQ_FREE, fits in header */
+int dapli_mix_pz_free(DAPL_PZ *m_pz)
+{
+	dat_mix_pz_t msg;
+	DAPL_IA * ia_ptr = m_pz->header.owner_ia;
+	scif_epd_t mix_ep = ia_ptr->hca_ptr->ib_trans.scif_ep;
+	int ret, len;
+
+	dapl_log(DAPL_DBG_TYPE_EXTENSION,
+		 " MIX_PZ_FREE: pz %p, send, proxy IB_PD %p\n",
+		 m_pz, m_pz->pd_handle);
+
+	/* request */
+	msg.hdr.ver = DAT_MIX_VER;
+	msg.hdr.op = MIX_PZ_FREE;
+	msg.hdr.status = 0;
+	msg.hdr.flags = MIX_OP_REQ;
+	msg.ctx = (uint64_t)m_pz;
+	msg.ib_pd = (uint64_t)m_pz->pd_handle;
+
+	len = sizeof(dat_mix_pz_t);
+	ret = scif_send(mix_ep, &msg, len, SCIF_SEND_BLOCK);
+	if (ret != len) {
+		dapl_log(1, " ERR: %s send on %d, ret %d, exp %d, error %s\n",
+			 mix_op_str(msg.hdr.op), mix_ep, ret, len, strerror(errno));
+	}
+
+	dapl_log(DAPL_DBG_TYPE_EXTENSION," Sent %s request on SCIF EP\n",
+		 mix_op_str(msg.hdr.op));
+
+	/* response */
+	ret = scif_recv(mix_ep, &msg, len, SCIF_RECV_BLOCK);
+	if (ret != len) {
+		dapl_log(1, " ERR: rcv on new_ep %d, ret %d, exp %d, error %s\n",
+			    mix_ep, ret, len, strerror(errno));
+		return -1;
+	}
+	if (msg.hdr.ver != DAT_MIX_VER || msg.hdr.op != MIX_PZ_FREE ||
+	    msg.hdr.flags != MIX_OP_RSP || msg.hdr.status != MIX_SUCCESS) {
+		dapl_log(1, " MIX_CQ_FREE ERR: ver %d, op %d, flags %d, or stat %d ln %d\n",
+			 msg.hdr.ver, msg.hdr.op, msg.hdr.flags, msg.hdr.status, ret);
+		return -1;
+	}
+
+	dapl_log(DAPL_DBG_TYPE_EXTENSION,
+		 " MIX_PZ_FREE: reply, proxy IB_PD 0x%x\n", msg.ib_pd);
+
+	m_pz->pd_handle = 0;
+
+	return 0;
+}
+
+
 /* SCIF DMA outbound writes and inbound msg receives; translate to scif_off via LMR */
 /* TODO: faster translation for post_send? */
 static inline int mix_proxy_data(ib_qp_handle_t m_qp, dat_mix_sr_t *msg, struct ibv_sge *sglist, int txlen, int mix_ep)
@@ -1152,9 +1416,11 @@ int dapli_mix_dto_event_in(ib_hca_transport_t *tp, scif_epd_t scif_ep, dat_mix_d
 			}
 			pmsg->wc[i].byte_len = cookie->val.dto.size;
 			dapl_log(DAPL_DBG_TYPE_EP,
-				 " mix_dto_event: MCM evd %p ep %p wr_id=%Lx ln=%d\n",
+				 " mix_dto_event: MCM evd %p ep %p wr_id=%Lx"
+				 " ln=%d op %d flgs %d\n",
 				 m_cq->evd, cookie->ep, pmsg->wc[i].wr_id,
-				 cookie->val.dto.size);
+				 cookie->val.dto.size, pmsg->wc[i].opcode,
+				 pmsg->wc[i].wc_flags);
 		}
 		mcm_const_ib_wc(&ib_wc, &pmsg->wc[i], 1);
 		dapl_os_lock(&m_cq->evd->header.lock);
@@ -1363,14 +1629,73 @@ int dapli_mix_recv(DAPL_HCA *hca, int scif_ep)
 	return ret;
 }
 
+int dapli_mix_query_port(ib_hca_transport_t *tp, unsigned long port_num, struct ibv_port_attr *port_attr)
+{
+	dat_mix_port_attr_t msg;
+	scif_epd_t mix_ep = tp->scif_ep;
+	int ret, len;
 
+	ret = scif_get_nodeIDs(NULL, 0, &tp->self.node);
+	if (ret < 0) {
+		dapl_log(1, " scif_get_nodeIDs() failed with error %s\n", strerror(errno));
+		return -1;
+	}
 
+	/* request: QP_r local, QP_t shadowed */
+	msg.hdr.ver = DAT_MIX_VER;
+	msg.hdr.op = MIX_QUERY_PORT;
+	msg.hdr.status = 0;
+	msg.hdr.flags = MIX_OP_REQ;
+	msg.hdr.req_id = port_num;
 
+	len = sizeof(dat_mix_hdr_t);
+	ret = scif_send(mix_ep, &msg, len, SCIF_SEND_BLOCK);
+	if (ret != len) {
+		dapl_log(1, " ERR: %s snd on %d, ret %d, exp %d, err %s\n",
+			 mix_op_str(msg.hdr.op), mix_ep, ret, len,
+			 strerror(errno));
+		return -1;
+	}
+	dapl_log(DAPL_DBG_TYPE_EXTENSION," Sent %s request on SCIF EP\n",
+		 mix_op_str(msg.hdr.op));
 
+	/* wait for response */
+	len = sizeof(dat_mix_port_attr_t);
+	ret = scif_recv(mix_ep, &msg, len, SCIF_RECV_BLOCK);
+	if (ret != len) {
+		dapl_log(1, " ERR: rcv on ep %d, ret %d, exp %d, err %s\n",
+			    mix_ep, ret, len, strerror(errno));
+		return -1;
+	}
+	if (msg.hdr.ver != DAT_MIX_VER || msg.hdr.op != MIX_QUERY_PORT ||
+	    msg.hdr.flags != MIX_OP_RSP || msg.hdr.status != MIX_SUCCESS) {
+		dapl_log(1, " ERR: %s ver %d, op %d, flags %d, stat %d\n",
+			    mix_op_str(msg.hdr.op), msg.hdr.ver,
+			    msg.hdr.op, msg.hdr.flags, msg.hdr.status);
+		return -1;
+	}
 
+	port_attr->gid_tbl_len = msg.gid_tbl_len;
+	port_attr->port_cap_flags = msg.port_cap_flags;
+	port_attr->max_msg_sz = msg.max_msg_sz;
+	port_attr->bad_pkey_cntr = msg.bad_pkey_cntr;
+	port_attr->qkey_viol_cntr = msg.qkey_viol_cntr;
+	port_attr->pkey_tbl_len = msg.pkey_tbl_len;
+	port_attr->lid = msg.lid;
+	port_attr->sm_lid = msg.sm_lid;
+	port_attr->lmc = msg.lmc;
+	port_attr->max_vl_num = msg.max_vl_num;
+	port_attr->sm_sl = msg.sm_sl;
+	port_attr->subnet_timeout = msg.subnet_timeout;
+	port_attr->init_type_reply = msg.init_type_reply;
+	port_attr->active_width = msg.active_width;
+	port_attr->active_speed = msg.active_speed;
+	port_attr->phys_state = msg.phys_state;
+	port_attr->link_layer = msg.link_layer;
+	port_attr->state = msg.state;
+	port_attr->max_mtu = msg.max_mtu;
+	port_attr->active_mtu = msg.active_mtu;
 
-
-
-
-
-
+	dapl_log(DAPL_DBG_TYPE_EXTENSION," MIX_QUERY_PORT successful on SCIF EP %d\n", mix_ep);
+	return 0;
+}
