@@ -110,14 +110,21 @@ int mcm_send_pi(struct dcm_ib_qp *m_qp,
 				sge.lkey = m_qp->wr_buf_rx_mr->lkey;
 			}
 			sge.addr = (uint64_t)(uintptr_t) wr_rx_ptr;
-			sge.length = (uint32_t) sizeof(struct mcm_wr_rx); /* 160 byte WR */
+			sge.length = (uint32_t)m_qp->wrc_rem.wr_sz;
+
+			if (m_qp->p2p_data && seg_len < m_qp->p2p_data) {
+				dapl_log(DAPL_DBG_TYPE_EP,
+					 " Sending p2p data, len %d\n", seg_len);
+				wr_flags |= M_PROXY_INLINE;
+				memcpy(wr_rx_ptr->inline_data, (void *)(l_addr + l_off), seg_len);
+			}
 
 			dapl_log(DAPL_DBG_TYPE_EP,
 				 " mcm_send_pi[%d]: seg_ln %d wr_idx %d, tl %d hd %d\n",
 				 i, seg_len, wr_idx, m_qp->wr_tl, m_qp->wr_hd);
 
 			/* build local m_wr_rx for remote PI */
-			memset((void*)wr_rx_ptr, 0, sizeof(struct mcm_wr_rx));
+			memset((void*)wr_rx_ptr, 0, MCM_WR_RX_NO_DATA);
 			wr_rx_ptr->org_id = (uint64_t) htonll((uint64_t)wr->wr_id);
 			wr_rx_ptr->flags = htonl(wr_flags);
 			wr_rx_ptr->w_idx = htonl(m_qp->wc_tl); /* snd back wc tail */
@@ -162,17 +169,13 @@ int mcm_send_pi(struct dcm_ib_qp *m_qp,
 				(uint64_t)(uintptr_t)
 				((struct mcm_wr_rx *) (m_qp->wrc_rem.wr_addr + (m_qp->wrc_rem.wr_sz * wr_idx)));
 
-			sge.addr = (uint64_t)(uintptr_t) &m_wr_rx;
-			sge.length = (uint32_t) sizeof(struct mcm_wr_rx); /* 160 byte WR */
-			sge.lkey = 0; /* inline doesn't need registered */
-
 			dapl_log(DAPL_DBG_TYPE_EP,
 				 " mcm_send_pi[%d]: WR_RX wr_id %Lx qn %x op %d flgs 0x%x"
 				 " imm %x laddr %p raddr %p rkey %x wr_ln %d ln %d\n",
 				 i, wr_imm.wr_id, m_qp->qp2->qp_num, wr_imm.opcode,
 				 wr_flags, ntohl(wr_imm.imm_data),
 				 l_addr + l_off, wr_imm.wr.rdma.remote_addr,
-				 wr_imm.wr.rdma.rkey, sizeof(struct mcm_wr_rx), l_len);
+				 wr_imm.wr.rdma.rkey, sge.length, l_len);
 			dapl_log(DAPL_DBG_TYPE_EP,
 				 " mcm_send_pi[%d]: WR wr_id %Lx qn %x op %d flgs %x"
 				 " imm %x raddr %p rkey %x sg_ln %d tl %d me %d hd %d\n",
@@ -186,7 +189,7 @@ int mcm_send_pi(struct dcm_ib_qp *m_qp,
 				dapl_log(DAPL_DBG_TYPE_ERR,
 					" mcm_send_pi ERR: m_wr %p idx %d laddr=%p ln=%d lkey=%x flgs %x"
 					" tl %d hd %d\n",
-					m_wr_rx, wr_idx, wr->sg_list[0].addr,
+					wr_rx_ptr, wr_idx, wr->sg_list[0].addr,
 					wr->sg_list[0].length, wr->sg_list[0].lkey,
 					wr_rx_ptr->flags, m_qp->wr_tl, m_qp->wr_hd);
 				dapl_log(DAPL_DBG_TYPE_ERR,
@@ -218,7 +221,7 @@ static inline void mcm_dto_req(struct dcm_ib_cq *m_cq, struct ibv_wc *wc)
 	cookie = (DAPL_COOKIE *)(uintptr_t)wc->wr_id;
 	m_qp = cookie->ep->qp_handle;
 
-	if (!m_qp->tp->scif_ep && MXS_EP(m_qp) &&
+	if (!m_qp->tp->scif_ep && MXF_EP(m_qp) &&
 	    (wc->opcode == (uint32_t)IBV_WR_RDMA_WRITE_WITH_IMM)) {
 		dapl_log(DAPL_DBG_TYPE_EP,
 			 " mcm_dto_req: RW_imm -> WR, wr_id %Lx\n", wc->wr_id);
@@ -441,7 +444,7 @@ int mcm_create_wc_q(struct dcm_ib_qp *m_qp, int entries)
 		m_qp->wrc.wc_addr, m_qp->wc_mr->addr, ALIGN_PAGE(m_qp->wrc.wc_len),
 		entries, m_qp->wc_mr->rkey, m_qp->wc_mr->lkey);
 
-	if (!m_qp->ep->header.owner_ia->hca_ptr->ib_trans.ib_cm.max_inline) {
+	if (!m_qp->tp->ib_cm.max_inline) {
 
 		if (posix_memalign((void **)&m_qp->wr_buf_rx,
 				   4096, entries * sizeof(mcm_wr_rx_t))) {
