@@ -63,6 +63,8 @@
 
 extern char *gid_str;
 
+#define SCM_BC_DIFF 2 /* cm_msg adjust, pre v8 */
+
 /* forward declarations */
 static DAT_RETURN
 dapli_socket_connect(DAPL_EP * ep_ptr,
@@ -561,6 +563,11 @@ static void dapli_socket_connected(dp_ib_cm_handle_t cm_ptr, int err)
 
 	/* send qp info and pdata to remote peer */
 	exp = sizeof(ib_cm_msg_t) - DCM_MAX_PDATA_SIZE;
+	if (cm_ptr->retry == SCM_CR_RETRY-1) {
+		exp += SCM_BC_DIFF; /* retry with pre-v8 req */
+		dapl_log(DAPL_DBG_TYPE_CM_WARN,
+			 " CONN_REQ: retry pre-v8 protocol; peer != v8\n");
+	}
 	iov[0].iov_base = (void *)&cm_ptr->msg;
 	iov[0].iov_len = exp;
 	if (cm_ptr->msg.p_size) {
@@ -583,16 +590,14 @@ static void dapli_socket_connected(dp_ib_cm_handle_t cm_ptr, int err)
 		goto bail;
 	}
 
- 	dapl_dbg_log(DAPL_DBG_TYPE_CM,
-		     " CONN_PENDING: sending SRC lid=0x%x,"
-		     " qpn=0x%x, psize=%d\n",
-		     ntohs(cm_ptr->msg.saddr.ib.lid),
-		     ntohl(cm_ptr->msg.saddr.ib.qpn), 
-		     ntohs(cm_ptr->msg.p_size));
-	dapl_dbg_log(DAPL_DBG_TYPE_CM,
-		     " CONN_PENDING: SRC GID %s\n",
-		     inet_ntop(AF_INET6, &cm_ptr->hca->ib_trans.gid,
-			       gid_str, sizeof(gid_str)));
+	dapl_log(DAPL_DBG_TYPE_CM,
+		 " CONN_REQ: (%d) SRC lid=0x%x,"
+		 " qpn=0x%x, psize=%d GID %s\n",
+		 exp, ntohs(cm_ptr->msg.saddr.ib.lid),
+		 ntohl(cm_ptr->msg.saddr.ib.qpn),
+		 ntohs(cm_ptr->msg.p_size),
+		 inet_ntop(AF_INET6, &cm_ptr->hca->ib_trans.gid,
+		           gid_str, sizeof(gid_str)));
 
 	DAPL_CNTR(((DAPL_IA *)dapl_llist_peek_head(&cm_ptr->hca->ia_list_head)), DCNT_IA_CM_REQ_TX);
 	return;
@@ -666,6 +671,7 @@ dapli_socket_connect(DAPL_EP * ep_ptr,
 	/* REQ: QP info in msg.saddr, IA address in msg.daddr, and pdata */
 	cm_ptr->hca = ia_ptr->hca_ptr;
 	cm_ptr->msg.op = ntohs(DCM_REQ);
+	cm_ptr->msg.mtu = ia_ptr->hca_ptr->ib_trans.ib_cm.mtu; /* local MTU to peer */
 	cm_ptr->msg.saddr.ib.qpn = htonl(ep_ptr->qp_handle->qp->qp_num);
 	cm_ptr->msg.saddr.ib.qp_type = ep_ptr->qp_handle->qp->qp_type;
 	cm_ptr->msg.saddr.ib.lid = ia_ptr->hca_ptr->ib_trans.lid;
@@ -730,9 +736,16 @@ static void dapli_socket_connect_rtu(dp_ib_cm_handle_t cm_ptr)
 	socklen_t sl;
 
 	/* read DST information into cm_ptr, overwrite SRC info */
-	dapl_dbg_log(DAPL_DBG_TYPE_EP, " connect_rtu: recv peer QP data\n");
-
 	len = recv(cm_ptr->socket, (char *)&cm_ptr->msg, exp, 0);
+
+	if (ntohs(cm_ptr->msg.ver) < DCM_VER_XPS) {
+		len += recv(cm_ptr->socket, (char *)&cm_ptr->msg+len, SCM_BC_DIFF, 0);
+		exp += SCM_BC_DIFF;
+	}
+
+	dapl_log(DAPL_DBG_TYPE_CM, " CONN_REP_IN: ver %d cm_sz %d, p_sz %d\n",
+		ntohs(cm_ptr->msg.ver), exp, ntohs(cm_ptr->msg.p_size));
+
 	if (len != exp || ntohs(cm_ptr->msg.ver) < DCM_VER_MIN) {
 		int err = dapl_socket_errno();
 		dapl_log(DAPL_DBG_TYPE_CM_WARN,
@@ -745,9 +758,10 @@ static void dapli_socket_connect_rtu(dp_ib_cm_handle_t cm_ptr)
 			 cm_ptr->retry);
 
 		/* Retry; corner case where server tcp stack resets under load */
-		if (err == ECONNRESET && --cm_ptr->retry) {
+		if ((err == ECONNRESET && --cm_ptr->retry) || (--cm_ptr->retry == SCM_CR_RETRY-1)) {
 			closesocket(cm_ptr->socket);
 			cm_ptr->socket = DAPL_INVALID_SOCKET;
+			dapl_log(DAPL_DBG_TYPE_CM_WARN, "CONN_REQ: retry %d\n", cm_ptr->retry);
 			dapli_socket_connect(cm_ptr->ep, (DAT_IA_ADDRESS_PTR)&cm_ptr->addr, 
 					     ntohs(((struct sockaddr_in *)&cm_ptr->addr)->sin_port) - 1000,
 					     ntohs(cm_ptr->msg.p_size), &cm_ptr->msg.p_data, cm_ptr->retry);
@@ -800,9 +814,10 @@ static void dapli_socket_connect_rtu(dp_ib_cm_handle_t cm_ptr)
 		len = recv(cm_ptr->socket, cm_ptr->msg.p_data, exp, 0);
 		if (len != exp) {
 			int err = dapl_socket_errno();
-			dapl_log(DAPL_DBG_TYPE_ERR,
-				 " CONN_RTU read pdata: ERR 0x%x %s, rcnt=%d -> %s\n",
-				 err, strerror(err), len,
+			dapl_log(DAPL_DBG_TYPE_CM_WARN,
+				 " CONN_RTU read pdata: ERR 0x%x %s"
+				 " rcv %d !=  exp %d -> %s\n",
+				 err, strerror(err), len, exp,
 				 inet_ntoa(((struct sockaddr_in *)
 					    ep_ptr->param.
 					    remote_ia_address_ptr)->sin_addr));
@@ -834,6 +849,11 @@ static void dapli_socket_connect_rtu(dp_ib_cm_handle_t cm_ptr)
 				DAPL_MIN(ep_ptr->param.ep_attr.max_rdma_read_out,
 					 cm_ptr->msg.rd_in);
 
+	/* Set QP MTU, if negotiated. 2K for compatibility */
+	ep_ptr->qp_handle->mtu = cm_ptr->msg.mtu ?
+				 DAPL_MIN(cm_ptr->msg.mtu, cm_ptr->hca->ib_trans.ib_cm.mtu):
+				 getenv("DAPL_IB_MTU") ? cm_ptr->hca->ib_trans.ib_cm.mtu : IBV_MTU_2048;
+
 	/* modify QP to RTR and then to RTS with remote info */
 	dapl_os_lock(&ep_ptr->header.lock);
 	if (dapls_modify_qp_state(ep_ptr->qp_handle->qp,
@@ -881,6 +901,7 @@ static void dapli_socket_connect_rtu(dp_ib_cm_handle_t cm_ptr)
 	dapl_os_unlock(&cm_ptr->lock);
 
 	cm_ptr->msg.op = ntohs(DCM_RTU);
+	cm_ptr->msg.mtu = ep_ptr->qp_handle->mtu; /* send negotiated MTU */
 	if (send(cm_ptr->socket, (char *)&cm_ptr->msg, 4, 0) == -1) {
 		int err = dapl_socket_errno();
 		dapl_log(DAPL_DBG_TYPE_ERR,
@@ -954,10 +975,11 @@ ud_bail:
 					     DCM_MAX_PDATA_SIZE, ep_ptr);
 	}
 	dapl_log(DAPL_DBG_TYPE_CM_EST,
-		 " SCM ACTIVE CONN: %x -> %s %x\n",
+		 " SCM ACTIVE CONN: %x -> %s %x mtu %d\n",
 		 ntohs(((struct sockaddr_in *) &cm_ptr->addr)->sin_port),
 		 inet_ntoa(((struct sockaddr_in *) &cm_ptr->msg.daddr.so)->sin_addr),
-		 ntohs(((struct sockaddr_in *) &cm_ptr->msg.daddr.so)->sin_port)-1000);
+		 ntohs(((struct sockaddr_in *) &cm_ptr->msg.daddr.so)->sin_port)-1000,
+		 ep_ptr->qp_handle->mtu);
 	return;
 
 bail:
@@ -1121,6 +1143,12 @@ static void dapli_socket_accept_data(ib_cm_srvc_handle_t acm_ptr)
 
 	/* read in DST QP info, IA address. check for private data */
 	len = recv(acm_ptr->socket, (char *)&acm_ptr->msg, exp, 0);
+
+	if (ntohs(acm_ptr->msg.ver) < DCM_VER_XPS) {
+		len += recv(acm_ptr->socket, (char *)&acm_ptr->msg+len, SCM_BC_DIFF, 0);
+		exp += SCM_BC_DIFF;
+	}
+
 	if (len != exp || ntohs(acm_ptr->msg.ver) < DCM_VER_MIN) {
 		int err = dapl_socket_errno();
 		dapl_log(DAPL_DBG_TYPE_ERR,
@@ -1251,6 +1279,14 @@ dapli_socket_accept_usr(DAPL_EP * ep_ptr,
 				DAPL_MIN(ep_ptr->param.ep_attr.max_rdma_read_out,
 					 cm_ptr->msg.rd_in);
 
+	if (ntohs(cm_ptr->msg.ver) <  DCM_VER_XPS)
+		exp += SCM_BC_DIFF;
+
+	/* Set QP MTU, if negotiated. 2K for compatibility */
+	ep_ptr->qp_handle->mtu = cm_ptr->msg.mtu ?
+				 DAPL_MIN(cm_ptr->msg.mtu, cm_ptr->hca->ib_trans.ib_cm.mtu):
+				 getenv("DAPL_IB_MTU") ? cm_ptr->hca->ib_trans.ib_cm.mtu : IBV_MTU_2048;
+
 	/* modify QP to RTR and then to RTS with remote info already read */
 	dapl_os_lock(&ep_ptr->header.lock);
 	if (dapls_modify_qp_state(ep_ptr->qp_handle->qp,
@@ -1290,6 +1326,7 @@ dapli_socket_accept_usr(DAPL_EP * ep_ptr,
 	local.ver = htons(DCM_VER);
 	local.op = htons(DCM_REP);
 	local.rd_in = ep_ptr->param.ep_attr.max_rdma_read_in;
+	local.mtu = ep_ptr->qp_handle->mtu; /* send negotiated MTU */
 	local.saddr.ib.qpn = htonl(ep_ptr->qp_handle->qp->qp_num);
 	local.saddr.ib.qp_type = ep_ptr->qp_handle->qp->qp_type;
 	local.saddr.ib.lid = ia_ptr->hca_ptr->ib_trans.lid;
@@ -1435,10 +1472,11 @@ ud_bail:
 		dapls_cr_callback(cm_ptr, event, NULL, 0, cm_ptr->sp);
 	}
 	dapl_log(DAPL_DBG_TYPE_CM_EST,
-		 " SCM PASSIVE CONN: %x <- %s %x\n",
+		 " SCM PASSIVE CONN: %x <- %s %x mtu %d\n",
 		 cm_ptr->sp->conn_qual,
 		 inet_ntoa(((struct sockaddr_in *) &cm_ptr->msg.daddr.so)->sin_addr),
-		 ntohs(((struct sockaddr_in *) &cm_ptr->msg.daddr.so)->sin_port));
+		 ntohs(((struct sockaddr_in *) &cm_ptr->msg.daddr.so)->sin_port),
+		 cm_ptr->ep->qp_handle->mtu);
 	return;
       
 bail:
