@@ -622,14 +622,15 @@ dp_ib_cm_handle_t ucm_cm_find(ib_hca_transport_t *tp, ib_cm_msg_t *msg)
 	lock = &tp->lock;
 
 	dapl_log(DAPL_DBG_TYPE_CM,
-		" ucm_recv: %s %d %x %x i %x c %x < %d %x %x i %x c %x\n",
+		" ucm_recv: %s %d %x %x i %x c %x < %d %x %x i %x c %x (%d,%d)\n",
 		dapl_cm_op_str(msg_op),
 		ntohl(msg->d_id), ntohs(msg->daddr.ib.lid),
 		UCM_PORT_NTOH(msg->dportx, msg->dport),
 		ntohl(msg->daddr.ib.qpn), ntohl(msg->dqpn),
 		ntohl(msg->s_id), ntohs(msg->saddr.ib.lid),
 		UCM_PORT_NTOH(msg->sportx, msg->sport),
-		ntohl(msg->saddr.ib.qpn), ntohl(msg->sqpn));
+		ntohl(msg->saddr.ib.qpn), ntohl(msg->sqpn),
+		tp->ib_cm.mtu, msg->mtu);
 
 retry_listenq:
 	dapl_os_lock(lock);
@@ -1467,12 +1468,13 @@ dapli_cm_connect(DAPL_EP *ep, dp_ib_cm_handle_t cm)
 {
 	dapl_log(DAPL_DBG_TYPE_EP, 
 		 " connect: lid %x i_qpn %x lport %x p_sz=%d -> "
-		 " lid %x c_qpn %x rport %x\n",
+		 " lid %x c_qpn %x rport %x l_mtu %d\n",
 		 htons(cm->msg.saddr.ib.lid), htonl(cm->msg.saddr.ib.qpn),
 		 UCM_PORT_NTOH(cm->msg.sportx,cm->msg.sport),
 		 htons(cm->msg.p_size),
 		 htons(cm->msg.daddr.ib.lid), htonl(cm->msg.dqpn),
-		 UCM_PORT_NTOH(cm->msg.dportx,cm->msg.dport));
+		 UCM_PORT_NTOH(cm->msg.dportx,cm->msg.dport),
+		 cm->hca->ib_trans.ib_cm.mtu);
 
 	dapl_os_lock(&cm->lock);
 	if (cm->state != DCM_INIT && cm->state != DCM_REP_PENDING) {
@@ -1513,6 +1515,8 @@ dapli_cm_connect(DAPL_EP *ep, dp_ib_cm_handle_t cm)
 
 	cm->state = DCM_REP_PENDING;
 	cm->msg.op = htons(DCM_REQ);
+	cm->msg.mtu = cm->hca->ib_trans.ib_cm.mtu; /* local MTU to peer */
+
 	if (ucm_send(&cm->hca->ib_trans, &cm->msg, 
 		     &cm->msg.p_data, ntohs(cm->msg.p_size))) {
 		dapl_os_unlock(&cm->lock);
@@ -1638,6 +1642,10 @@ static void ucm_connect_rtu(dp_ib_cm_handle_t cm, ib_cm_msg_t *msg)
 		cm->ep->param.ep_attr.max_rdma_read_out =
 				DAPL_MIN(cm->ep->param.ep_attr.max_rdma_read_out,
 					 cm->msg.rd_in);
+	/* Set QP MTU, if negotiated. 2K for compatibility */
+	ep->qp_handle->mtu = msg->mtu ?
+			     DAPL_MIN(msg->mtu, cm->hca->ib_trans.ib_cm.mtu):
+			     getenv("DAPL_IB_MTU") ? cm->hca->ib_trans.ib_cm.mtu : IBV_MTU_2048;
 
 	/* modify QP to RTR and then to RTS with remote info */
 	dapl_os_lock(&cm->ep->header.lock);
@@ -1671,6 +1679,7 @@ static void ucm_connect_rtu(dp_ib_cm_handle_t cm, ib_cm_msg_t *msg)
 	
 	/* Send RTU, no private data */
 	cm->msg.op = htons(DCM_RTU);
+	cm->msg.mtu = ep->qp_handle->mtu; /* send negotiated MTU */
 
 	dapl_os_lock(&cm->lock);
 	cm->state = DCM_CONNECTED;
@@ -1760,11 +1769,11 @@ ud_bail:
 	}
 
 	dapl_log(DAPL_DBG_TYPE_CM_EST,
-		 " UCM_ACTIVE_CONN %p %d [lid port qpn] %x %x %x -> %x %x %x xevent=%d\n",
+		 " UCM_ACTIVE_CONN %p %d [lid port qpn] %x %x %x -> %x %x %x mtu %d\n",
 		 cm->hca, cm->retries, ntohs(cm->msg.saddr.ib.lid),
 		 ntohs(cm->msg.sport), ntohl(cm->msg.saddr.ib.qpn),
 		 ntohs(cm->msg.daddr.ib.lid), ntohs(cm->msg.dport),
-		 ntohl(cm->msg.dqpn), sizeof(DAT_IB_EXTENSION_EVENT_DATA));
+		 ntohl(cm->msg.dqpn), ep->qp_handle->mtu);
 	return;
 bail:
 	if (ntohs(msg->op) != DCM_REJ_USER) {
@@ -1812,6 +1821,7 @@ static void ucm_accept(ib_cm_srvc_handle_t cm, ib_cm_msg_t *msg)
 	acm->msg.p_size = msg->p_size;
 	acm->msg.d_id = msg->s_id;
 	acm->msg.rd_in = msg->rd_in;
+	acm->msg.mtu = msg->mtu; /* save peer MTU */
 
 	/* CR saddr is CM daddr info, need EP for local saddr */
 	dapl_os_memcpy(&acm->msg.daddr, &msg->saddr, sizeof(union dcm_addr));
@@ -1832,14 +1842,15 @@ static void ucm_accept(ib_cm_srvc_handle_t cm, ib_cm_msg_t *msg)
 
 	dapl_log(DAPL_DBG_TYPE_CM,
 		 " accepting: op %s [id lid, port, cqp, iqp]:"
-		 " %d %x %x %x %x <- %d %x %x %x %x\n",
+		 " %d %x %x %x %x <- %d %x %x %x %x mtu %d\n",
 		 dapl_cm_op_str(ntohs(msg->op)),
 		 ntohl(acm->msg.s_id), ntohs(msg->daddr.ib.lid),
 		 UCM_PORT_NTOH(msg->dportx, msg->dport),
 		 ntohl(msg->dqpn), ntohl(msg->daddr.ib.qpn),
 		 ntohl(msg->s_id), ntohs(msg->saddr.ib.lid),
 		 UCM_PORT_NTOH(msg->sportx, msg->sport),
-		 ntohl(msg->sqpn), ntohl(msg->saddr.ib.qpn));
+		 ntohl(msg->sqpn), ntohl(msg->saddr.ib.qpn),
+		 acm->msg.mtu);
 
 #ifdef DAT_EXTENSIONS
 	if (acm->msg.daddr.ib.qp_type == IBV_QPT_UD) {
@@ -1958,13 +1969,13 @@ static void ucm_accept_rtu(dp_ib_cm_handle_t cm, ib_cm_msg_t *msg)
 	}
 
 	dapl_log(DAPL_DBG_TYPE_CM_EST,
-		 " UCM_PASSIVE_CONN %p %d [lid port qpn] %x %x %x <- %x %x %x\n",
+		 " UCM_PASSIVE_CONN %p %d [lid port qpn] %x %x %x <- %x %x %x mtu %d\n",
 		 cm->hca, cm->retries, ntohs(cm->msg.saddr.ib.lid),
 		 UCM_PORT_NTOH(cm->msg.sportx, cm->msg.sport),
 		 ntohl(cm->msg.saddr.ib.qpn),
 		 ntohs(cm->msg.daddr.ib.lid),
 		 UCM_PORT_NTOH(cm->msg.dportx, cm->msg.dport),
-		 ntohl(cm->msg.dqpn));
+		 ntohl(cm->msg.dqpn), cm->ep->qp_handle->mtu);
 	return;
 bail:
 	dapl_log(DAPL_DBG_TYPE_CM_WARN,
@@ -2090,11 +2101,11 @@ dapli_accept_usr(DAPL_EP *ep, DAPL_CR *cr, DAT_COUNT p_size, DAT_PVOID p_data)
 
 	dapl_dbg_log(DAPL_DBG_TYPE_CM,
 		     " ACCEPT_USR: s_id %d r_id %d lid=%x"
-		     " iqp=%x qp_type %d, psize=%d\n",
+		     " iqp=%x qp_type %d, psize=%d r_mtu %d l_mtu %d\n",
 		     ntohl(cm->msg.s_id), ntohl(cm->msg.d_id),
 		     ntohs(cm->msg.daddr.ib.lid),
 		     ntohl(cm->msg.daddr.ib.qpn), cm->msg.daddr.ib.qp_type, 
-		     p_size);
+		     p_size, cm->msg.mtu, cm->hca->ib_trans.ib_cm.mtu);
 
 #ifdef DAT_EXTENSIONS
 	if (cm->msg.daddr.ib.qp_type == IBV_QPT_UD &&
@@ -2110,6 +2121,10 @@ dapli_accept_usr(DAPL_EP *ep, DAPL_CR *cr, DAT_COUNT p_size, DAT_PVOID p_data)
 		ep->param.ep_attr.max_rdma_read_out =
 				DAPL_MIN(ep->param.ep_attr.max_rdma_read_out,
 					 cm->msg.rd_in);
+	/* Set QP MTU, if negotiated. 2K for compatibility */
+	ep->qp_handle->mtu = cm->msg.mtu ?
+			     DAPL_MIN(cm->msg.mtu, cm->hca->ib_trans.ib_cm.mtu):
+			     getenv("DAPL_IB_MTU") ? cm->hca->ib_trans.ib_cm.mtu : IBV_MTU_2048;
 
 	/* modify QP to RTR and then to RTS with remote info already read */
 	dapl_os_lock(&ep->header.lock);
@@ -2146,6 +2161,7 @@ dapli_accept_usr(DAPL_EP *ep, DAPL_CR *cr, DAT_COUNT p_size, DAT_PVOID p_data)
 	/* setup local QP info and type from EP, copy pdata, for reply */
 	cm->msg.op = htons(DCM_REP);
 	cm->msg.rd_in = ep->param.ep_attr.max_rdma_read_in;
+	cm->msg.mtu = ep->qp_handle->mtu; /* send negotiated MTU */
 	cm->msg.saddr.ib.qpn = htonl(ep->qp_handle->qp->qp_num);
 	cm->msg.saddr.ib.qp_type = ep->qp_handle->qp->qp_type;
 	cm->msg.saddr.ib.lid = cm->hca->ib_trans.addr.ib.lid; 
