@@ -57,6 +57,10 @@
 #define UCM_ARRAY_IDX(idx, abits) (idx >> abits)
 #define UCM_ENTRY_IDX(idx, abits) (idx & (abits - 1))
 
+/* Default UCM timers set for small cluster */
+#define UCM_DEF_NODES 32
+#define UCM_DEF_PPN   32
+#define UCM_DEF_THRESHOLD (UCM_DEF_NODES * UCM_DEF_PPN)
 
 /* DAPL CM objects MUST include list_entry, ref_count, event for EP linking */
 struct ib_cm_handle
@@ -125,6 +129,10 @@ typedef struct _ib_hca_transport
 	int			wait_time;
 	int			dreq_cnt;
 	int			drep_time;
+	int			nodes;
+	int			ppn;
+	int			ranks;
+	int 			threshold;
 	DAPL_OS_LOCK		slock;	
 	int			s_hd;
 	int			s_tl;
