@@ -49,6 +49,8 @@ int mcm_rep_ms = 4000;
 int mcm_rtu_ms = 2000;
 int mcm_dreq_ms = 1000;
 int mcm_proxy_in = 1;
+int mcm_mic0_mss = 1;  /* numa_node invalid, default mic0 == MSS */
+int mcm_mic1_mss = 0;  /* numa_node invalid, default mic1 == MXS */
 
 extern int mix_buffer_sg_po2;
 extern int mcm_rx_entries;
@@ -163,6 +165,12 @@ int mcm_init_cm_service(mcm_ib_dev_t *md)
 	md->cqe = mcm_depth;
 	md->signal = mcm_signal;
 
+	/* ibscif: MIC and HST different, save MIC addresses */
+	if (!strncmp(md->ibdev->name, "scif", 4)) {
+		md->m_lid = md->addr.lid;
+		memcpy(md->m_gid, md->addr.gid, 16);
+	}
+
 	/* Save addr information */
 	/* get lid for this hca-port, convert to network order */
 	if (ibv_query_port(md->ibctx, md->port, &port_attr)) {
@@ -187,12 +195,20 @@ int mcm_init_cm_service(mcm_ib_dev_t *md)
 		return -1;
 	}
 	/* EP mapping hint for MIC to HCA, set MSS if compat or PI disabled */
-	if (((md->numa_node != -1) &&
-	     (md->numa_node == md->mc->numa_node)) ||
-	      md->mc->ver == MIX_COMP || mcm_proxy_in == 0)
-		md->addr.ep_map = MIC_SSOCK_DEV;
-	else
-		md->addr.ep_map = MIC_XSOCK_DEV;
+	if (md->addr.ep_map != MIC_FULL_DEV) {
+		if (((md->numa_node != -1) &&
+			 (md->numa_node == md->mc->numa_node)) ||
+			  md->mc->ver == MIX_COMP || mcm_proxy_in == 0)
+			md->addr.ep_map = MIC_SSOCK_DEV;
+		else
+			md->addr.ep_map = MIC_XSOCK_DEV;
+
+		/* Invalid numa, check settings MSS for mic0, MXS for mic1 */
+		if (md->numa_node == -1 && md->mc->scif_id == 1 && mcm_mic0_mss)
+			md->addr.ep_map = MIC_SSOCK_DEV;
+		if (md->numa_node == -1 && md->mc->scif_id == 2 && mcm_mic1_mss)
+			md->addr.ep_map = MIC_SSOCK_DEV;
+	}
 
 	/* setup CM timers and queue sizes */
 	md->pd = ibv_alloc_pd(md->ibctx);
@@ -285,11 +301,11 @@ int mcm_init_cm_service(mcm_ib_dev_t *md)
 	/* save qp_num as part of ia_address, network order */
 	md->addr.qpn = htonl(md->qp->qp_num);
 
-	mlog(0, " IB LID 0x%x PORT %d GID %s QPN 0x%x: mic%d -> %s - %s, mic_ver %d %s\n",
-		ntohs(md->addr.lid), md->port,
+	mlog(0, " IB LID %x (%x) PORT %d GID %s QPN %x: mic%d -> %s - %s, mic_ver %d %s\n",
+		ntohs(md->addr.lid), md->m_lid, md->port,
 		inet_ntop(AF_INET6, md->addr.gid, gid_str, sizeof(gid_str)),
 		md->qp->qp_num, md->mc->scif_id - 1, md->ibdev->name,
-	        md->addr.ep_map == MIC_SSOCK_DEV ? "MSS":"MXS", md->mc->ver,
+		mcm_map_str(md->addr.ep_map), md->mc->ver,
 	        md->mc->ver == MIX_COMP ? "WARN: OLD MPSS -> RUN COMP MODE":"");
 
 	return 0;
@@ -330,7 +346,9 @@ int mcm_modify_qp(struct ibv_qp	*qp_handle,
 
 		qp_attr.dest_qp_num = ntohl(qpn);
 		qp_attr.rq_psn = 1;
-		qp_attr.path_mtu = m_qp->smd->md->dev_attr.mtu;
+		qp_attr.path_mtu = m_qp->mtu ?
+				   min(m_qp->mtu, m_qp->smd->md->dev_attr.mtu):
+				   m_qp->smd->md->dev_attr.mtu;
 		qp_attr.max_dest_rd_atomic = 16;
 		qp_attr.min_rnr_timer = m_qp->smd->md->dev_attr.rnr_timer;
 		qp_attr.ah_attr.dlid = ntohs(lid);
@@ -442,28 +460,25 @@ int mcm_modify_qp(struct ibv_qp	*qp_handle,
 	return ret;
 }
 
-/* move QP's to error state and destroy. Flush the proxy SR queue is exists */
+/* move QP's to error state. Flush the proxy SR queue is exists */
 void mcm_flush_qp(struct mcm_qp *m_qp)
 {
 	struct mcm_sr *m_sr;
 	struct dat_mix_wc wc;
 
-	if (m_qp->ib_qp1) {
+	mlog(2, " QP1 %p QP2 %p\n", m_qp->ib_qp1, m_qp->ib_qp2);
+
+	if (m_qp->ib_qp1)
 		mcm_modify_qp(m_qp->ib_qp1, IBV_QPS_ERR, 0, 0, NULL);
-		ibv_destroy_qp(m_qp->ib_qp1);
-		m_qp->ib_qp1 = NULL;
-	}
-	if (m_qp->ib_qp2) {
+
+	if (m_qp->ib_qp2)
 		mcm_modify_qp(m_qp->ib_qp2, IBV_QPS_ERR, 0, 0, NULL);
-		ibv_destroy_qp(m_qp->ib_qp2);
-		m_qp->ib_qp2 = NULL;
-	}
 
 	mpxy_lock(&m_qp->rxlock);
 	while (m_qp->sr_tl != m_qp->sr_hd) {
 		m_sr = (struct mcm_sr *)(m_qp->sr_buf + (m_qp->sr_sz * m_qp->sr_tl));
 		if (m_sr->wr_id) {
-			mlog(1, " QP %p SR[%d] %p wr_id %Lx dto_event -> IBV_WC_FLUSH_ERR\n",
+			mlog(2, " QP %p SR[%d] %p wr_id %Lx dto_event -> IBV_WC_FLUSH\n",
 				m_qp, m_sr->s_idx, m_sr, m_sr->wr_id);
 			wc.wr_id = m_sr->wr_id;
 			wc.imm_data = 0;
@@ -557,14 +572,23 @@ mcm_cm_t *m_cm_create(mcm_scif_dev_t *smd, mcm_qp_t *m_qp, dat_mcm_addr_t *r_add
 		/* MPXYD SRC IB info, QP2t = saddr2 all cases */
 		cm->msg.saddr2.qpn = htonl(m_qp->ib_qp2->qp_num);
 		cm->msg.saddr2.qp_type = m_qp->qp_attr2.qp_type;
-                cm->msg.saddr2.lid = smd->md->addr.lid;
-                cm->msg.saddr2.ep_map = smd->md->addr.ep_map;
-                memcpy(&cm->msg.saddr2.gid[0], &smd->md->addr.gid, 16);
+		cm->msg.saddr2.ep_map = smd->md->addr.ep_map;
 
                 /* MPXYD RCV IB info */
-                cm->msg.saddr1.lid = smd->md->addr.lid;
                 cm->msg.saddr1.ep_map = smd->md->addr.ep_map;
-                memcpy(&cm->msg.saddr1.gid[0], &smd->md->addr.gid, 16);
+
+		/* intra-node: QPt addr 2 - ibscif HST lid != MICx lid's */
+		if (smd->md->m_lid) {
+			cm->msg.saddr1.lid = smd->md->m_lid;
+			cm->msg.saddr2.lid = smd->md->addr.lid;
+			memcpy(&cm->msg.saddr1.gid[0], &smd->md->m_gid, 16);
+			memcpy(&cm->msg.saddr2.gid[0], &smd->md->addr.gid, 16);
+		} else {
+			cm->msg.saddr1.lid = smd->md->lid;
+			cm->msg.saddr2.lid = smd->md->lid;
+			memcpy(&cm->msg.saddr1.gid[0], &smd->md->addr.gid, 16);
+			memcpy(&cm->msg.saddr2.gid[0], &smd->md->addr.gid, 16);
+		}
 
                 /* MSS, QPr is on MIC, QP1r == saddr1 */
 	        if (MSS_EP(l_addr)) {
@@ -572,14 +596,14 @@ mcm_cm_t *m_cm_create(mcm_scif_dev_t *smd, mcm_qp_t *m_qp, dat_mcm_addr_t *r_add
 			cm->msg.saddr1.qp_type = cm->m_qp->qp_attr1.qp_type;
 
 			/* MSS_EP -> (MXS or unknown) WC queue for MXS peer PI service */
-			if (MXS_EP(r_addr) || UND_EP(r_addr)) {
+			if (MXF_EP(r_addr) || UND_EP(r_addr)) {
 				if (m_pi_create_wc_q(m_qp, mcm_rx_entries))
 					goto bail;
 			}
 	        }
 
         	/* MXS -> (MSS, HOST, or unknown), need a QPr on mpxyd, QP1r == saddr1 */
-                if (MXS_EP(l_addr) && !MXS_EP(r_addr)) {
+                if (MXF_EP(l_addr) && !MXF_EP(r_addr)) {
                 	/* note: pi_wr_q and pi_wc_q created via MXS create_qp */
                 	if (m_qp_create_pi(smd, m_qp))
                 		goto bail;
@@ -588,7 +612,7 @@ mcm_cm_t *m_cm_create(mcm_scif_dev_t *smd, mcm_qp_t *m_qp, dat_mcm_addr_t *r_add
                 }
 
                 /* MXS -> MXS, QPs and QPr is QP2 on mpxyd, saddr 1 == saddr2 */
-                if (MXS_EP(l_addr) && MXS_EP(r_addr))
+                if (MXF_EP(l_addr) && MXF_EP(r_addr))
                 	memcpy(&cm->msg.saddr1, &cm->msg.saddr2, sizeof(dat_mcm_addr_t));
 
 		mlog(8, " SRC: QPt qpn 0x%x lid 0x%x, QPr qpn 0x%x lid 0x%x"
@@ -780,6 +804,10 @@ static int mcm_send(mcm_ib_dev_t *md, dat_mcm_msg_t *msg, DAT_PVOID p_data, DAT_
 	int len, ret = -1;
 	uint16_t dlid = ntohs(msg->daddr1.lid);
 
+	/* intra-node, CM on host side, HST lid != MIC lid */
+	if (md->m_lid)
+		dlid = ntohs(md->addr.lid);
+
 	/* Get message from send queue, copy data, and send */
 	mpxy_lock(&md->txlock);
 	if ((smsg = mcm_get_smsg(md)) == NULL) {
@@ -816,7 +844,7 @@ static int mcm_send(mcm_ib_dev_t *md, dat_mcm_msg_t *msg, DAT_PVOID p_data, DAT_
         }
 
 	mlog(8," cm_send: op %s ln %d lid %x c_qpn %x rport %x, p_size %d\n",
-		mcm_op_str(ntohs(smsg->op)), sge.length, ntohs(smsg->daddr1.lid),
+		mcm_op_str(ntohs(smsg->op)), sge.length, dlid,
 		ntohl(smsg->dqpn), ntohs(smsg->dport), p_size);
 
 	/* empty slot, then create AH */
@@ -888,9 +916,17 @@ int mcm_cm_rej_out(mcm_ib_dev_t *md, dat_mcm_msg_t *msg, DAT_MCM_OP type, int sw
 		memcpy(&smsg.saddr2, &msg->saddr2, sizeof(dat_mcm_addr_t));
 	}
 
+	/* intra-node; addr1 == mic lid, addr2 == cm lid */
+	if (md->m_lid) {
+		smsg.saddr1.lid = md->m_lid;
+		memcpy(&smsg.saddr1.gid[0], &md->m_gid, 16);
+	}
+	smsg.saddr2.lid = md->addr.lid;
+	memcpy(&smsg.saddr2.gid[0], &md->addr.gid, 16);
+
 	mlog(2," sLID %x, sQPN %x sPORT %x -> dLID %x, dQPN %x dPORT %x\n",
-	     ntohs(smsg.saddr1.lid), ntohl(smsg.sqpn), ntohs(smsg.sport),
-	     ntohs(smsg.daddr1.lid), ntohl(smsg.dqpn), ntohs(smsg.dport));
+	     ntohs(smsg.saddr2.lid), ntohl(smsg.sqpn), ntohs(smsg.sport),
+	     ntohs(smsg.daddr2.lid), ntohl(smsg.dqpn), ntohs(smsg.dport));
 
 	if (type == MCM_REJ_USER)
 		MCNTR(md, MCM_CM_REJ_USER_OUT);
@@ -900,6 +936,29 @@ int mcm_cm_rej_out(mcm_ib_dev_t *md, dat_mcm_msg_t *msg, DAT_MCM_OP type, int sw
 	return (mcm_send(md, &smsg, NULL, 0));
 }
 
+int mcm_qp_busy(struct mcm_qp *m_qp)
+{
+	int busy = 0;
+
+	mpxy_lock(&m_qp->rxlock);
+	busy += (m_qp->pi_rw_cnt + m_qp->pi_rr_cnt);
+	busy += (m_qp->stall_cnt_rr + m_qp->post_cnt_wt);
+	mpxy_unlock(&m_qp->rxlock);
+	mlog(2," PI[%d]: rw_cnt %d rr_cnt %d stall %d post %d\n",
+	     busy, m_qp->pi_rw_cnt, m_qp->pi_rr_cnt,
+	     m_qp->stall_cnt_rr, m_qp->post_cnt_wt);
+
+	mpxy_lock(&m_qp->txlock);
+	busy += (m_qp->wr_pp + m_qp->wr_pp_rem);
+	busy += (m_qp->post_sig_cnt - m_qp->comp_cnt);
+	mpxy_unlock(&m_qp->txlock);
+	mlog(2," PO[%d]: wr_pp %d pp_rem %d pst_sig %d cmp_sig %d\n",
+	     busy, m_qp->wr_pp, m_qp->wr_pp_rem,
+	     m_qp->post_sig_cnt, m_qp->comp_cnt);
+
+	return busy;
+}
+
 void mcm_cm_disc(mcm_cm_t *cm)
 {
 	int finalize = 1;
@@ -909,7 +968,7 @@ void mcm_cm_disc(mcm_cm_t *cm)
 	switch (cm->state) {
 	case MCM_CONNECTED:
 		/* CONSUMER: move to err state to flush */
-		if (cm->m_qp)
+		if (cm->m_qp && !mcm_qp_busy(cm->m_qp))
 			mcm_flush_qp(cm->m_qp);
 
 		/* send DREQ, event after DREP or DREQ timeout */
@@ -950,7 +1009,7 @@ void mcm_cm_disc(mcm_cm_t *cm)
 	case MCM_DISC_RECV:
 		MCNTR(cm->md, MCM_CM_DREQ_IN);
 		/* CM_THREAD: move to err state to flush */
-		if (cm->m_qp)
+		if (cm->m_qp && !mcm_qp_busy(cm->m_qp))
 			mcm_flush_qp(cm->m_qp);
 
 		/* DREQ received, send DREP and schedule event, finalize */
@@ -964,6 +1023,8 @@ void mcm_cm_disc(mcm_cm_t *cm)
 		break;
 	case MCM_DISCONNECTED:
 		mlog(2," state = %s already disconnected\n",  mcm_state_str(cm->state) );
+		if (cm->m_qp) /* MIC client disc, force flush */
+			mcm_flush_qp(cm->m_qp);
 		mpxy_unlock(&cm->lock);
 		MCNTR(cm->md, MCM_CM_DREQ_DUP);
 		return;
@@ -1053,7 +1114,6 @@ static void mcm_process_recv(mcm_ib_dev_t *md, dat_mcm_msg_t *msg, mcm_cm_t *cm,
 	case MCM_RTU_PENDING: /* passive */
 		mlog(2, "RTU_PENDING: cm %p, my_id %d, cm_id %d\n",
 			cm, cm->entry.tid, cm->cm_id);
-		cm->state = MCM_CONNECTED;
 		mpxy_unlock(&cm->lock);
 		MCNTR(md, MCM_CM_RTU_IN);
 		mix_cm_rtu_in(cm, msg, len);
@@ -1433,6 +1493,7 @@ int mcm_cm_req_out(mcm_cm_t *m_cm)
 	m_cm->state = MCM_REP_PENDING;
 	m_cm->msg.op = htons(MCM_REQ);
 	m_cm->timer = mcm_time_us(); /* reset reply timer */
+	m_cm->msg.mtu = m_cm->smd->md->dev_attr.mtu; /* local MTU to peer */
 
 	if (mcm_send(m_cm->md, &m_cm->msg, &m_cm->msg.p_data, ntohs(m_cm->msg.p_size)))
 		return -1;
@@ -1450,15 +1511,16 @@ int mcm_cm_rtu_out(mcm_cm_t *m_cm)
 
 	MCNTR(m_cm->md, MCM_CM_RTU_OUT);
 
-	mlog(1, "[%d:%d] CONN_EST[%d]: %p 0x%x %x 0x%x %Lx %s -> 0x%x %x 0x%x %Lx %s\n",
+	mlog(1, "[%d:%d] CONN_EST[%d]: %p 0x%x %x 0x%x %Lx %s -> 0x%x %x 0x%x %Lx %s mtu %d\n",
 		m_cm->md->mc->scif_id, m_cm->smd->entry.tid,
 		m_cm->md->cntrs ? (uint32_t)((uint64_t *)m_cm->md->cntrs)[MCM_CM_RTU_OUT]:0,
 		m_cm, htons(m_cm->msg.saddr2.lid), htonl(m_cm->msg.saddr2.qpn),
 		htons(m_cm->msg.sport),	system_guid, mcm_map_str(m_cm->msg.saddr2.ep_map),
 		htons(m_cm->msg.daddr1.lid),
-		MXS_EP(&m_cm->msg.saddr1) && MXS_EP(&m_cm->msg.daddr1) ?
+		MXF_EP(&m_cm->msg.saddr1) && MXF_EP(&m_cm->msg.daddr1) ?
 				htonl(m_cm->msg.daddr2.qpn):htonl(m_cm->msg.daddr1.qpn),
-		htons(m_cm->msg.dport), ntohll(r_guid), mcm_map_str(m_cm->msg.daddr1.ep_map));
+		htons(m_cm->msg.dport), ntohll(r_guid), mcm_map_str(m_cm->msg.daddr1.ep_map),
+		m_cm->m_qp->mtu);
 
 	mpxy_lock(&m_cm->lock);
 	if (m_cm->state != MCM_REP_RCV) {
