@@ -186,8 +186,6 @@ void mix_scif_accept(scif_epd_t listen_ep)
 	mlog(8, " SCIF client: device open client_pid 0x%x - mlen %d - ep %d\n",
 		ntohl(msg.hdr.req_id), len, op_ep);
 
-	msg.hdr.flags = MIX_OP_RSP;
-
 	if (msg.hdr.ver < MIX_MIN || msg.hdr.ver > MIX_MAX || msg.hdr.op != MIX_IA_OPEN) {
 		mlog(0, " ERR: MIC client incompatible with MPXYD (exp %d,rcvd %d) or OP (exp %d,rcvd %d)\n",
 			DAT_MIX_VER, msg.hdr.ver, msg.hdr.op, MIX_IA_OPEN);
@@ -207,7 +205,6 @@ void mix_scif_accept(scif_epd_t listen_ep)
 	if (smd)
 		return;
 err:
-	mlog(0, " ERR: open_device -> closing SCIF client EPs %d %d %d \n", op_ep, tx_ep, ev_ep);
 
 out_close_tx_ep:
 	scif_close(tx_ep);
@@ -219,6 +216,110 @@ out_close_op_ep:
 	scif_close(op_ep);
 }
 
+static int mix_query_device(mcm_scif_dev_t *smd, dat_mix_device_attr_t *pmsg)
+{
+	int ret, len;
+	dat_mix_device_attr_t msg;
+	struct ibv_device_attr dev_attr;
+
+	len = sizeof(dat_mix_device_attr_t);
+	msg.hdr.ver = DAT_MIX_VER;
+	msg.hdr.op = MIX_QUERY_DEVICE;
+	msg.hdr.flags = MIX_OP_RSP;
+
+	ret = ibv_query_device(smd->md->ibctx, &dev_attr);
+	if (ret) {
+	        msg.hdr.status = MIX_EFAULT;
+	        goto done;
+	}
+	msg.hdr.status = MIX_SUCCESS;
+
+	strncpy(msg.fw_ver, dev_attr.fw_ver, sizeof(msg.fw_ver));
+	msg.node_guid = dev_attr.node_guid;
+	msg.sys_image_guid = dev_attr.sys_image_guid;
+	msg.max_mr_size = dev_attr.max_mr_size;
+	msg.page_size_cap = dev_attr.page_size_cap;
+	msg.vendor_id = dev_attr.vendor_id;
+	msg.vendor_part_id = dev_attr.vendor_part_id;
+	msg.hw_ver = dev_attr.hw_ver;
+	msg.max_qp = dev_attr.max_qp;
+	msg.max_qp_wr = dev_attr.max_qp_wr;
+	msg.device_cap_flags = dev_attr.device_cap_flags;
+	msg.max_sge = dev_attr.max_sge;
+	msg.max_sge_rd = dev_attr.max_sge_rd;
+	msg.max_cq = dev_attr.max_cq;
+	msg.max_cqe = dev_attr.max_cqe;
+	msg.max_mr = dev_attr.max_mr;
+	msg.max_pd = dev_attr.max_pd;
+	msg.max_qp_rd_atom = dev_attr.max_qp_rd_atom;
+	msg.max_ee_rd_atom = dev_attr.max_ee_rd_atom;
+	msg.max_ee_rd_atom = dev_attr.max_res_rd_atom;
+	msg.max_qp_init_rd_atom = dev_attr.max_qp_init_rd_atom;
+	msg.max_ee_init_rd_atom = dev_attr.max_ee_init_rd_atom;
+	msg.atomic_cap = dev_attr.atomic_cap;
+	msg.max_ee = dev_attr.max_ee;
+	msg.max_rdd = dev_attr.max_rdd;
+	msg.max_mw = dev_attr.max_mw;
+	msg.max_raw_ipv6_qp = dev_attr.max_raw_ipv6_qp;
+	msg.max_raw_ethy_qp = dev_attr.max_raw_ethy_qp;
+	msg.max_mcast_grp = dev_attr.max_mcast_grp;
+	msg.max_mcast_qp_attach = dev_attr.max_mcast_qp_attach;
+	msg.max_total_mcast_qp_attach = dev_attr.max_total_mcast_qp_attach;
+	msg.max_ah = dev_attr.max_ah;
+	msg.max_fmr = dev_attr.max_fmr;
+	msg.max_map_per_fmr = dev_attr.max_map_per_fmr;
+	msg.max_srq = dev_attr.max_srq;
+	msg.max_srq_wr = dev_attr.max_srq_wr;
+	msg.max_srq_sge = dev_attr.max_srq_sge;
+	msg.max_pkeys = dev_attr.max_pkeys;
+	msg.local_ca_ack_delay = dev_attr.local_ca_ack_delay;
+	msg.phys_port_cnt = dev_attr.phys_port_cnt;
+done:
+	return (scif_send_msg(smd->scif_op_ep, &msg, len));
+}
+
+static int mix_query_port(mcm_scif_dev_t *smd, dat_mix_port_attr_t *pmsg)
+{
+	int ret, len;
+	dat_mix_port_attr_t msg;
+	struct ibv_port_attr port_attr;
+
+	len = sizeof(dat_mix_port_attr_t);
+	msg.hdr.ver = DAT_MIX_VER;
+	msg.hdr.op = MIX_QUERY_PORT;
+	msg.hdr.flags = MIX_OP_RSP;
+
+	ret = ibv_query_port(smd->md->ibctx, pmsg->hdr.req_id, &port_attr);
+	if (ret) {
+	        msg.hdr.status = MIX_EFAULT;
+	        goto done;
+	}
+	msg.hdr.status = MIX_SUCCESS;
+
+	msg.gid_tbl_len = port_attr.gid_tbl_len;
+	msg.port_cap_flags = port_attr.port_cap_flags;
+	msg.max_msg_sz = port_attr.max_msg_sz;
+	msg.bad_pkey_cntr = port_attr.bad_pkey_cntr;
+	msg.qkey_viol_cntr = port_attr.qkey_viol_cntr;
+	msg.pkey_tbl_len = port_attr.pkey_tbl_len;
+	msg.lid = port_attr.lid;
+	msg.sm_lid = port_attr.sm_lid;
+	msg.lmc = port_attr.lmc;
+	msg.max_vl_num = port_attr.max_vl_num;
+	msg.sm_sl = port_attr.sm_sl;
+	msg.subnet_timeout = port_attr.subnet_timeout;
+	msg.init_type_reply = port_attr.init_type_reply;
+	msg.active_width = port_attr.active_width;
+	msg.active_speed = port_attr.active_speed;
+	msg.phys_state = port_attr.phys_state;
+	msg.link_layer = port_attr.link_layer;
+	msg.state = port_attr.state;
+	msg.max_mtu = port_attr.max_mtu;
+	msg.active_mtu = port_attr.active_mtu;
+done:
+	return (scif_send_msg(smd->scif_op_ep, &msg, len));
+}
+
 static int mix_prov_attr(mcm_scif_dev_t *smd, dat_mix_attr_t *pmsg)
 {
 	int len, ret;
@@ -422,6 +523,27 @@ void m_cq_free(struct mcm_cq *m_cq)
 	free(m_cq);
 }
 
+/* smd->cqlock/cqrlock held */
+void m_cq_flush(struct mcm_cq *m_cq)
+{
+	struct ibv_cq *ib_cq = NULL;
+	void *cq_ctx;
+	int ret, cnt=0;
+	struct ibv_wc wc;
+
+	mlog(8, " m_cq %p enter:\n", m_cq);
+	ret = ibv_get_cq_event(m_cq->ib_ch, &ib_cq, (void *)&cq_ctx);
+	if (ret == 0)
+		ibv_ack_cq_events(m_cq->ib_cq, 1);
+
+	do {
+		ret = ibv_poll_cq(m_cq->ib_cq, 1, &wc);
+		cnt += ret;
+	}
+	while (ret > 0);
+	mlog(8, " m_cq %p exit: %d events flushed\n", m_cq, cnt);
+}
+
 /* destroy proxy CQ, fits in header */
 static int mix_cq_destroy(mcm_scif_dev_t *smd, dat_mix_hdr_t *pmsg)
 {
@@ -555,6 +677,114 @@ resp:
 	return (scif_send_msg(smd->scif_op_ep, (void*)pmsg, len));
 }
 
+/* create new proxy-out PZ */
+static int mix_pz_create(mcm_scif_dev_t *smd, dat_mix_pz_t *pmsg)
+{
+	int len, ret;
+	mcm_pz_t *new_pz;
+
+	/* hdr already read, get operation data */
+	len = sizeof(dat_mix_pz_t) - sizeof(dat_mix_hdr_t);
+	ret = scif_recv(smd->scif_op_ep, ((char*)pmsg + sizeof(dat_mix_hdr_t)), len, SCIF_RECV_BLOCK);
+	if (ret != len) {
+		mlog(0, " ERR: ret %d, exp %d\n", ret, len);
+		return ret;
+	}
+
+	new_pz = malloc(sizeof(mcm_pz_t));
+	if (!new_pz) {
+		mlog(0, " ERR: Could not alloc mcm_pz_t\n");
+		goto err;
+	}
+
+	memset(new_pz, 0, sizeof(mcm_pz_t));
+
+	init_list(&new_pz->entry);
+	new_pz->smd = smd;
+	new_pz->pd_ctx = pmsg->ctx;
+
+	new_pz->ib_pd = ibv_alloc_pd(smd->md->ibctx);
+
+	if (!new_pz->ib_pd) {
+		mlog(0, " ERR: Could not alloc pd\n");
+		goto err;
+	}
+
+	mpxy_lock(&smd->pzlock);
+	insert_tail(&new_pz->entry, &smd->pzlist, new_pz);
+	mpxy_unlock(&smd->pzlock);
+	mlog(8, " new mcm_pz %p and ib_pd %p created\n", new_pz, new_pz->ib_pd);
+
+	pmsg->ib_pd = (uint64_t)new_pz;
+	pmsg->hdr.status = MIX_SUCCESS;
+	goto resp;
+err:
+	if (new_pz)
+		free(new_pz);
+
+	mlog(0, " ERR: %s\n", strerror(errno));
+	pmsg->hdr.status = MIX_EINVAL;
+resp:
+	/* send back response */
+	pmsg->hdr.flags = MIX_OP_RSP;
+	len = sizeof(dat_mix_pz_t);
+	return (scif_send_msg(smd->scif_op_ep, (void*)pmsg, len));
+}
+
+/* destroy proxy-out PZ */
+static int mix_pz_destroy(mcm_scif_dev_t *smd, dat_mix_pz_t *pmsg)
+{
+	mcm_pz_t *mcm_pz;
+	int len, ret;
+
+	/* hdr already read, get operation data */
+	len = sizeof(dat_mix_pz_t) - sizeof(dat_mix_hdr_t);
+	ret = scif_recv(smd->scif_op_ep, ((char*)pmsg + sizeof(dat_mix_hdr_t)), len, SCIF_RECV_BLOCK);
+	if (ret != len) {
+		mlog(0, " ERR: ret %d, exp %d\n", ret, len);
+		return ret;
+	}
+
+	mlog(8, " MIX_PZ_DESTROY: msg ctx 0x%lx, msg ib_pd 0x%lx\n", pmsg->ctx, pmsg->ib_pd);
+
+	/* find the PD in list */
+	mpxy_lock(&smd->pzlock);
+	mcm_pz = get_head_entry(&smd->pzlist);
+	while (mcm_pz) {
+		if ((uint64_t)mcm_pz == pmsg->ib_pd) {
+			mlog(8, " Matched mcm_pz %p (ib_pd %p)\n", mcm_pz, mcm_pz->ib_pd);
+			break;
+		}
+		mcm_pz = get_next_entry(&mcm_pz->entry, &smd->pzlist);
+	}
+	mpxy_unlock(&smd->pzlock);
+
+	if (!mcm_pz) {
+		mlog(0, " ERR: ib_pd 0x%lx was not found\n", pmsg->ib_pd);
+		goto err;
+	}
+
+	mpxy_lock(&smd->pzlock);
+	ret = ibv_dealloc_pd(mcm_pz->ib_pd);
+	remove_entry(&mcm_pz->entry);
+	free(mcm_pz);
+	mpxy_unlock(&smd->pzlock);
+
+	if (ret)
+		goto err;
+
+	pmsg->hdr.status = MIX_SUCCESS;
+	goto resp;
+err:
+	mlog(0, " ERR: %s\n", strerror(errno));
+	pmsg->hdr.status = MIX_EINVAL;
+resp:
+	/* send back response */
+	pmsg->hdr.flags = MIX_OP_RSP;
+	len = sizeof(dat_mix_pz_t);
+	return (scif_send_msg(smd->scif_op_ep, (void*)pmsg, len));
+}
+
 /* called with smd->qptlist lock held */
 void m_qp_free(struct mcm_qp *m_qp)
 {
@@ -567,8 +797,26 @@ void m_qp_free(struct mcm_qp *m_qp)
 		mpxy_unlock(&m_qp->smd->qprlock);
 	}
 
-	mlog(8, " m_qp %p m_cm %p cm_id %d\n",
-		m_qp, m_qp->cm, m_qp->cm ? m_qp->cm->entry.tid:0);
+	mlog(8, " m_qp %p m_cm %p cm_id %d cm_state %d\n",
+		m_qp, m_qp->cm, m_qp->cm ? m_qp->cm->entry.tid:0,
+		m_qp->cm ? m_qp->cm->state:0);
+
+	if (m_qp->cm)
+		m_qp->cm->state = MCM_DISCONNECTED;
+
+	mcm_flush_qp(m_qp); /* QP to error, flush consumer messages */
+
+	if (m_qp->m_cq_tx) { /* flush pending PO WRs on cq_tx */
+		mpxy_lock(&m_qp->smd->cqlock);
+		m_cq_flush(m_qp->m_cq_tx);
+		mpxy_unlock(&m_qp->smd->cqlock);
+	}
+
+	if (m_qp->m_cq_rx) { /* flush pending PI WRs on cq_rx */
+		mpxy_lock(&m_qp->smd->cqrlock);
+		m_cq_flush(m_qp->m_cq_rx);
+		mpxy_unlock(&m_qp->smd->cqrlock);
+	}
 
 	if (m_qp->cm) { /* unlink CM, serialized */
 		struct mcm_cm *cm = m_qp->cm;
@@ -580,7 +828,15 @@ void m_qp_free(struct mcm_qp *m_qp)
 		mpxy_unlock(&cm->lock);
 		mcm_dqconn_free(m_qp->smd, cm);
 	}
-	mcm_flush_qp(m_qp); /* move QP to error, flush & destroy */
+
+	if (m_qp->ib_qp1) {
+		ibv_destroy_qp(m_qp->ib_qp1);
+		m_qp->ib_qp1 = NULL;
+	}
+	if (m_qp->ib_qp2) {
+		ibv_destroy_qp(m_qp->ib_qp2);
+		m_qp->ib_qp2 = NULL;
+	}
 
 #ifdef MCM_PROFILE
 	if (mcm_profile)
@@ -592,6 +848,7 @@ void m_qp_free(struct mcm_qp *m_qp)
 	if (m_qp->m_cq_rx) {
 		mpxy_lock(&m_qp->smd->cqrlock);
 		m_cq_free(m_qp->m_cq_rx);
+		m_qp->m_cq_rx = NULL;
 		mpxy_unlock(&m_qp->smd->cqrlock);
 	}
 	mpxy_lock_destroy(&m_qp->txlock); /* proxy out */
@@ -737,7 +994,7 @@ static int m_qp_create(mcm_scif_dev_t *smd,
 		goto err;
 
 	/* plus proxy-in: create WR and WC pools, rx_cq */
-	if (smd->md->addr.ep_map == MIC_XSOCK_DEV) {
+	if (MXF_EP(&smd->md->addr)) {
 		if (m_pi_create_bpool(m_qp, attr->cap.max_recv_wr))
 			goto err;
 
@@ -799,7 +1056,7 @@ static int m_qp_create(mcm_scif_dev_t *smd,
 		m_qp->qp_attr2.max_inline_data = attr->cap.max_inline_data;
 
 	/* don't post recv msgs until RTU, if local MXS or local MSS to remote MXS */
-	if (smd->md->addr.ep_map == MIC_XSOCK_DEV)
+	if (MXF_EP(&smd->md->addr))
 		memcpy(&m_qp->qp_attr1, &m_qp->qp_attr2, sizeof(dat_mix_qp_attr_t));
 
 	/* TX proxy-out thread queue */
@@ -877,7 +1134,7 @@ static int mix_qp_create(mcm_scif_dev_t *smd, dat_mix_qp_t *pmsg)
 	/* return QPt, QPr info to MIC client, insert on QP list */
 	memcpy(&pmsg->qp_t, &new_mqp->qp_attr2, sizeof(dat_mix_qp_attr_t));
 
-	if (smd->md->addr.ep_map == MIC_XSOCK_DEV)
+	if (MXF_EP(&smd->md->addr))
 		memcpy(&pmsg->qp_r, &new_mqp->qp_attr1, sizeof(dat_mix_qp_attr_t));
 	else {
 		new_mqp->qp_attr1.qp_num = pmsg->qp_r.qp_num; /* QP1 == MIC QPr */
@@ -954,6 +1211,10 @@ static int mix_mr_create(mcm_scif_dev_t *smd, dat_mix_mr_t *pmsg)
 
 	pmsg->mr_id = m_mr->entry.tid;
 	smd->ref_cnt++;
+	if (MFO_EP(&smd->md->addr)) {
+		m_mr->mre.ib_lkey = m_mr->entry.tid;
+		m_mr->mre.ib_rkey = m_mr->entry.tid;
+	}
 	pmsg->hdr.status = MIX_SUCCESS;
 
 	mlog(8, " mr[%d] - len %d lmr_ctx %p, scif_addr %Lx, scif_off 0x%x, ib addr %Lx ib_rkey 0x%x\n",
@@ -1042,12 +1303,16 @@ void mix_dto_event(struct mcm_cq *m_cq, struct dat_mix_wc *wc, int nc)
 
 		if (msg.wc[i].status != IBV_WC_SUCCESS) {
 			if (msg.wc[i].status  != IBV_WC_WR_FLUSH_ERR) {
-				mlog(0, " ERROR (ep=%d): cq %p id %d ctx %p stat %d"
-					" [%d:%d] op 0x%x ln %d wr_id %p wc's %d verr 0x%x errno=%d,%s\n",
-					m_cq->smd->md->mc->scif_id, m_cq->smd->entry.tid,
-					m_cq->smd->scif_op_ep, m_cq, msg.cq_id, msg.cq_ctx,
-					msg.wc[i].status, msg.wc[i].opcode, msg.wc[i].byte_len,
-					msg.wc[i].wr_id, msg.wc_cnt, msg.wc[i].vendor_err,
+				mlog(0, " [%d:%d] ERROR (ep=%d): id %d stat %d"
+					" op %x flg %x ln %d wr_id %p wc's %d"
+					" verr 0x%x errno=%d,%s\n",
+					m_cq->smd->md->mc->scif_id,
+					m_cq->smd->entry.tid,
+					m_cq->smd->scif_op_ep, msg.cq_id,
+					msg.wc[i].status, msg.wc[i].opcode,
+					msg.wc[i].wc_flags, msg.wc[i].byte_len,
+					msg.wc[i].wr_id, msg.wc_cnt,
+					msg.wc[i].vendor_err,
 					errno, strerror(errno));
 			}
 		} else {
@@ -1083,8 +1348,9 @@ void mix_cm_event(mcm_cm_t *m_cm, uint32_t event)
 		mpxy_unlock(&m_cm->lock);
 	}
 
-	mlog(2, " MIX_CM_EVENT: cm %p cm_id %d, ctx %p, event 0x%x dev_id %d\n",
-	     m_cm, m_cm->entry.tid, msg.cm_ctx, event, m_cm->smd->entry.tid);
+	mlog(2, " MIX_CM_EVENT: (ep=%d) cm %p id %d ctx %p event 0x%x dev_id %d\n",
+		m_cm->smd->scif_ev_ep, m_cm, m_cm->entry.tid,
+		msg.cm_ctx, event, m_cm->smd->entry.tid);
 
 	len = sizeof(dat_mix_cm_event_t);
 	mpxy_lock(&m_cm->smd->evlock);
@@ -1155,10 +1421,16 @@ static int mix_cm_req_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 	mcm_hton_wrc((mcm_wrc_info_t *)m_cm->msg.p_proxy, &m_qp->wrc); /* PI WR/WC raddr,rkey info */
 	m_cm->msg.seg_sz = mix_buffer_sg_po2;
 
+	/* intra-node; set QPt to HST lid */
+	if (smd->md->m_lid) {
+		m_cm->msg.saddr2.lid = smd->md->addr.lid;
+		memcpy(&m_cm->msg.saddr2.gid[0], &smd->md->addr.gid, 16);
+	}
+
 	mlog(2," QP2 0x%x QP1 0x%x:"
 	       " CM sPORT 0x%x sQPN 0x%x sLID 0x%x - dPORT 0x%x dQPN 0x%x dLID 0x%x, psz %d %s\n",
 		m_cm->msg.saddr2.qpn, m_cm->msg.saddr1.qpn,
-		ntohs(m_cm->msg.sport), ntohl(m_cm->msg.sqpn), ntohs(m_cm->msg.saddr1.lid),
+		ntohs(m_cm->msg.sport), ntohl(m_cm->msg.sqpn), ntohs(m_cm->msg.saddr2.lid),
 		ntohs(m_cm->msg.dport), ntohl(m_cm->msg.dqpn), ntohs(m_cm->msg.daddr1.lid),
 		ntohs(m_cm->msg.p_size), mcm_map_str(m_cm->msg.daddr1.ep_map));
 
@@ -1222,8 +1494,10 @@ static int mix_cm_disc_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t s
 		return 0;
 	}
 
-	/* process DREQ */
+	/* process DREQ, flush QP */
 	mcm_cm_disc(m_cm);
+	if (m_cm->m_qp)
+		mcm_flush_qp(m_cm->m_qp); /* move QP to error, flush */
 	return 0;
 }
 
@@ -1261,6 +1535,7 @@ static int mix_cm_rtu_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 		 ntohs(m_cm->msg.daddr1.lid), ntohll(m_cm->msg.sys_guid));
 
 	/* send RTU on wire */
+	m_cm->msg.mtu = m_cm->m_qp->mtu; /* send negotiated MTU */
 	mcm_cm_rtu_out(m_cm);
 
 	return 0;
@@ -1360,6 +1635,16 @@ int mix_cm_rep_in(mcm_cm_t *m_cm, dat_mcm_msg_t *pkt, int pkt_len)
 	memcpy(&m_cm->msg.daddr1, &pkt->saddr1, sizeof(dat_mcm_addr_t));
 	memcpy(&m_cm->msg.daddr2, &pkt->saddr2, sizeof(dat_mcm_addr_t));
 	mcm_ntoh_wrc(&m_cm->m_qp->wrc_rem, (mcm_wrc_info_t *)m_cm->msg.p_proxy); /* peer RI WRC info */
+	if (m_cm->m_qp->wrc_rem.wr_sz > MCM_WR_RX_NO_DATA)
+		m_cm->m_qp->p2p_data = DAT_MCM_P2P_INLINE;
+	else
+		m_cm->m_qp->p2p_data = 0;
+
+	/* Set QP MTU, if negotiated. 2K for compatibility */
+	m_cm->m_qp->mtu = pkt->mtu ?
+			  min(pkt->mtu, m_cm->md->dev_attr.mtu):
+			  m_cm->md->mtu_env ? m_cm->md->mtu_env : IBV_MTU_2048;
+	m_cm->msg.mtu = m_cm->m_qp->mtu; /* forward negotiated MTU */
 
 	mlog(2, " WRC: m_qp %p - WR 0x%Lx rkey 0x%x ln %d, sz %d end %d"
 		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d\n",
@@ -1369,16 +1654,17 @@ int mix_cm_rep_in(mcm_cm_t *m_cm, dat_mcm_msg_t *pkt, int pkt_len)
 	     m_cm->m_qp->wrc.wc_rkey, m_cm->m_qp->wrc.wc_len,
 	     m_cm->m_qp->wrc.wc_sz, m_cm->m_qp->wrc.wc_end);
 
-	mlog(2, " WRC_rem: m_qp %p - WR 0x%Lx rkey 0x%x ln %d, sz %d end %d"
-		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d\n",
-	     m_cm->m_qp, m_cm->m_qp->wrc_rem.wr_addr, m_cm->m_qp->wrc_rem.wr_rkey,
+	mlog(2, " WRC_rem: WR 0x%Lx rkey 0x%x ln %d, sz %d end %d"
+		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d p2p %d\n",
+	     m_cm->m_qp->wrc_rem.wr_addr, m_cm->m_qp->wrc_rem.wr_rkey,
 	     m_cm->m_qp->wrc_rem.wr_len, m_cm->m_qp->wrc_rem.wr_sz,
 	     m_cm->m_qp->wrc_rem.wr_end, m_cm->m_qp->wrc_rem.wc_addr,
 	     m_cm->m_qp->wrc_rem.wc_rkey, m_cm->m_qp->wrc_rem.wc_len,
-	     m_cm->m_qp->wrc_rem.wc_sz, m_cm->m_qp->wrc_rem.wc_end);
+	     m_cm->m_qp->wrc_rem.wc_sz, m_cm->m_qp->wrc_rem.wc_end,
+	     m_cm->m_qp->p2p_data);
 
 	/* MXS <- MSS or HOST, fabric: TX: QP2->QP1 direct, RX: QP1<-QP2 proxy */
-	if ((MXS_EP(&m_cm->md->addr) && !MXS_EP(&m_cm->msg.daddr1)) &&
+	if ((MXF_EP(&m_cm->md->addr) && !MXF_EP(&m_cm->msg.daddr1)) &&
 	    system_guid != m_cm->msg.sys_guid) {
 		mlog(2, " MXS <- %s remote \n", mcm_map_str(m_cm->msg.daddr1.ep_map));
 
@@ -1409,7 +1695,7 @@ int mix_cm_rep_in(mcm_cm_t *m_cm, dat_mcm_msg_t *pkt, int pkt_len)
 		mpxy_unlock(&m_cm->smd->qprlock);
 
 	/* MXS <- MXS, proxy-in both sides, fabric, 1 QP only  */
-	} else if ((MXS_EP(&m_cm->md->addr) && MXS_EP(&m_cm->msg.daddr1)) &&
+	} else if ((MXF_EP(&m_cm->md->addr) && MXF_EP(&m_cm->msg.daddr1)) &&
 		   system_guid != m_cm->msg.sys_guid) {
 		mlog(2, " MXS <- MXS remote \n");
 
@@ -1430,7 +1716,7 @@ int mix_cm_rep_in(mcm_cm_t *m_cm, dat_mcm_msg_t *pkt, int pkt_len)
 		m_qp_destroy_pi(m_cm->m_qp);
 
 	/* MXS <- MXS, proxy-in both sides, inside system, no QP's, SCIF services only */
-	} else if ((MXS_EP(&m_cm->md->addr) && MXS_EP(&m_cm->msg.daddr1)) &&
+	} else if ((MXF_EP(&m_cm->md->addr) && MXF_EP(&m_cm->msg.daddr1)) &&
 		   system_guid == m_cm->msg.sys_guid) {
 		mlog(0, " MXS <- MXS local NOT SUPPORTED \n");
 		qp = NULL;
@@ -1439,10 +1725,10 @@ int mix_cm_rep_in(mcm_cm_t *m_cm, dat_mcm_msg_t *pkt, int pkt_len)
 	} else {
 		mlog(2, " MSS <- %s remote \n", mcm_map_str(m_cm->msg.daddr1.ep_map));
 
-		if (MXS_EP(&m_cm->msg.daddr1) && m_pi_prep_rcv_q(m_cm->m_qp))
+		if (MXF_EP(&m_cm->msg.daddr1) && m_pi_prep_rcv_q(m_cm->m_qp))
 				goto err;
 
-		if (!MXS_EP(&m_cm->msg.daddr1))
+		if (!MXF_EP(&m_cm->msg.daddr1))
 			m_pi_destroy_wc_q(m_cm->m_qp); /* created if ep_map was unknown */
 
 		qp = m_cm->m_qp->ib_qp2;
@@ -1516,6 +1802,7 @@ int mix_cm_req_in(mcm_cm_t *cm, dat_mcm_msg_t *pkt, int pkt_len)
 	acm->msg.p_size = pkt->p_size;
 	acm->msg.d_id = pkt->s_id;
 	acm->msg.rd_in = pkt->rd_in;
+	acm->msg.mtu = pkt->mtu;
 #ifdef MPXYD_LOCAL_SUPPORT
 	acm->msg.sys_guid = pkt->sys_guid; /* remote system guid */;
 #else
@@ -1527,13 +1814,14 @@ int mix_cm_req_in(mcm_cm_t *cm, dat_mcm_msg_t *pkt, int pkt_len)
 	memcpy(&acm->msg.daddr1, &pkt->saddr1, sizeof(dat_mcm_addr_t));
 	memcpy(&acm->msg.daddr2, &pkt->saddr2, sizeof(dat_mcm_addr_t));
 
-	mlog(2, " [%d:%d] cm %p ep %d sPORT %x %s <- dPORT %x lid=%x psz=%d %s %s %Lx (msg %p %d)\n",
+	mlog(2, " [%d:%d] cm %p ep %d: %x %s <- %x lid=%x psz=%d %s %s %Lx (%p %d) lmtu %d rmtu %d\n",
 		 cm->md->mc->scif_id, cm->smd->entry.tid, acm, acm->smd->scif_ev_ep,
 		 ntohs(acm->msg.sport), mcm_map_str(acm->md->addr.ep_map),
 		 ntohs(acm->msg.dport), ntohs(acm->msg.daddr1.lid), htons(acm->msg.p_size),
 		 mcm_map_str(acm->msg.daddr2.ep_map),
 		 acm->md->addr.lid == acm->msg.daddr1.lid ? "platform":"fabric",
-		 ntohll(acm->msg.sys_guid), &msg, sizeof(dat_mcm_msg_t));
+		 ntohll(acm->msg.sys_guid), &msg, sizeof(dat_mcm_msg_t),
+		 cm->md->dev_attr.mtu, pkt->mtu);
 
 	if (pkt->p_size)
 		memcpy(acm->msg.p_data, pkt->p_data, ntohs(pkt->p_size));
@@ -1568,25 +1856,35 @@ int mix_cm_rtu_in(mcm_cm_t *m_cm, dat_mcm_msg_t *pkt, int pkt_len)
 	dat_mix_cm_t msg;
 	int len;
 
-	mlog(1, "[%d:%d] CONN_EST[%d]: %p 0x%x %x 0x%x %Lx %s <- 0x%x %x 0x%x %Lx %s\n",
+	mlog(1, "[%d:%d] CONN_EST[%d]: %p 0x%x %x 0x%x %Lx %s <- 0x%x %x 0x%x %Lx %s mtu %d\n",
 		m_cm->md->mc->scif_id, m_cm->smd->entry.tid,
 		m_cm->md->cntrs ? (uint32_t)((uint64_t *)m_cm->md->cntrs)[MCM_CM_RTU_IN]:0,
 		m_cm, htons(pkt->daddr1.lid),
-		MXS_EP(&m_cm->msg.daddr1) && MXS_EP(&m_cm->msg.saddr1) ?
+		MXF_EP(&m_cm->msg.daddr1) && MXF_EP(&m_cm->msg.saddr1) ?
 			htonl(m_cm->msg.daddr2.qpn):htonl(m_cm->msg.daddr1.qpn),
 		htons(pkt->dport), system_guid, mcm_map_str(pkt->daddr1.ep_map),
 		htons(pkt->saddr2.lid), htonl(pkt->saddr2.qpn),
-		htons(pkt->sport), ntohll(pkt->sys_guid), mcm_map_str(pkt->saddr2.ep_map));
+		htons(pkt->sport), ntohll(pkt->sys_guid), mcm_map_str(pkt->saddr2.ep_map),
+		m_cm->m_qp->mtu);
 
-	/* MXS_EP <- HST_EP, host sends WC on RTU, save WRC info */
-	if (MXS_EP(&pkt->daddr1) && HST_EP(&pkt->saddr2)) {
+	/* MXF_EP <- HST_EP, host sends WC on RTU, save WRC info */
+	if (MXF_EP(&pkt->daddr1) && HST_EP(&pkt->saddr2)) {
 		mcm_ntoh_wrc(&m_cm->m_qp->wrc_rem, (mcm_wrc_info_t *)pkt->p_proxy);
-		mlog(2, " WRC_rem: m_qp %p - addr 0x%Lx rkey 0x%x len %d, sz %d end %d\n",
+		if (m_cm->m_qp->wrc_rem.wr_sz > MCM_WR_RX_NO_DATA)
+			m_cm->m_qp->p2p_data = DAT_MCM_P2P_INLINE;
+		else
+			m_cm->m_qp->p2p_data = 0;
+
+		mlog(2, " WRC_rem: WC addr 0x%Lx rkey 0x%x len %d, sz %d end %d p2p %d\n",
 		     m_cm->m_qp, m_cm->m_qp->wrc_rem.wc_addr, m_cm->m_qp->wrc_rem.wc_rkey,
 		     m_cm->m_qp->wrc_rem.wc_len, m_cm->m_qp->wrc_rem.wc_sz,
-		     m_cm->m_qp->wrc_rem.wc_end);
+		     m_cm->m_qp->wrc_rem.wc_end, m_cm->m_qp->p2p_data);
 	}
 
+	mpxy_lock(&m_cm->lock);
+	m_cm->state = MCM_CONNECTED;
+	mpxy_unlock(&m_cm->lock);
+
 	/* Forward, as is, conn_reply message to MIC client, with remote QP info */
 	msg.hdr.ver = m_cm->md->mc->ver;
 	msg.hdr.flags = MIX_OP_REQ;
@@ -1664,6 +1962,10 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 	m_cm->ref_cnt++; /* Passive: QP ref */
 	m_cm->m_qp->cm = m_cm;
 	mcm_ntoh_wrc(&m_cm->m_qp->wrc_rem, (mcm_wrc_info_t *)m_cm->msg.p_proxy); /* save peer PI WRC info */
+	if (m_cm->m_qp->wrc_rem.wr_sz > MCM_WR_RX_NO_DATA)
+		m_cm->m_qp->p2p_data = DAT_MCM_P2P_INLINE;
+	else
+		m_cm->m_qp->p2p_data = 0;
 
 	mlog(2, " WRC: m_qp %p - WR 0x%Lx rkey 0x%x ln %d, sz %d end %d"
 		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d sg_po2 %d\n",
@@ -1674,16 +1976,17 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 	     m_cm->m_qp->wrc.wc_sz, m_cm->m_qp->wrc.wc_end,
 	     m_cm->msg.seg_sz);
 
-	mlog(2, " WRC_rem: m_qp %p - WR 0x%Lx rkey 0x%x ln %d, sz %d end %d"
-		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d\n",
-	     m_cm->m_qp, m_cm->m_qp->wrc_rem.wr_addr, m_cm->m_qp->wrc_rem.wr_rkey,
+	mlog(2, " WRC_rem: WR 0x%Lx rkey 0x%x ln %d, sz %d end %d"
+		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d p2p %d\n",
+	     m_cm->m_qp->wrc_rem.wr_addr, m_cm->m_qp->wrc_rem.wr_rkey,
 	     m_cm->m_qp->wrc_rem.wr_len, m_cm->m_qp->wrc_rem.wr_sz,
 	     m_cm->m_qp->wrc_rem.wr_end, m_cm->m_qp->wrc_rem.wc_addr,
 	     m_cm->m_qp->wrc_rem.wc_rkey, m_cm->m_qp->wrc_rem.wc_len,
-	     m_cm->m_qp->wrc_rem.wc_sz, m_cm->m_qp->wrc_rem.wc_end);
+	     m_cm->m_qp->wrc_rem.wc_sz, m_cm->m_qp->wrc_rem.wc_end,
+	     m_cm->m_qp->p2p_data);
 
 	/* MXS -> MSS or HOST, remote: need QPr1, saddr1 on mpxyd */
-	if ((MXS_EP(&m_cm->md->addr) && !MXS_EP(&m_cm->msg.daddr1)) &&
+	if ((MXF_EP(&m_cm->md->addr) && !MXF_EP(&m_cm->msg.daddr1)) &&
 	    (system_guid != m_cm->msg.sys_guid) ) {
 		mlog(2, " MXS -> %s remote \n", mcm_map_str(m_cm->msg.daddr1.ep_map));
 
@@ -1695,13 +1998,13 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 
 		/* KR to KL or XEON, QP1<-QP2 and QP2->QP1 */
 		/* update the src information in CM msg */
-		m_cm->msg.saddr1.ep_map = MIC_XSOCK_DEV;
+		m_cm->msg.saddr1.ep_map = smd->md->addr.ep_map;
 		m_cm->msg.saddr1.qpn = htonl(m_cm->m_qp->ib_qp1->qp_num);
 		m_cm->msg.saddr1.qp_type = m_cm->m_qp->qp_attr1.qp_type;
 	        m_cm->msg.saddr1.lid = m_cm->smd->md->addr.lid;
 		memcpy(&m_cm->msg.saddr1.gid[0], &m_cm->smd->md->addr.gid, 16);
 
-		m_cm->msg.saddr2.ep_map = MIC_XSOCK_DEV;
+		m_cm->msg.saddr2.ep_map = smd->md->addr.ep_map;
 		m_cm->msg.saddr2.qpn = htonl(m_cm->m_qp->ib_qp2->qp_num);
 		m_cm->msg.saddr2.qp_type = m_cm->m_qp->qp_attr2.qp_type;
 	        m_cm->msg.saddr2.lid = m_cm->smd->md->addr.lid;
@@ -1725,7 +2028,7 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 		mpxy_unlock(&smd->qprlock);
 
 	/* MXS -> MXS, proxy-in both sides, remote, 1 QP - already setup */
-	} else if ((MXS_EP(&m_cm->md->addr) && MXS_EP(&m_cm->msg.daddr1)) &&
+	} else if ((MXF_EP(&m_cm->md->addr) && MXF_EP(&m_cm->msg.daddr1)) &&
 		   (system_guid != m_cm->msg.sys_guid)) {
 		mlog(2, " MXS -> MXS remote \n");
 
@@ -1733,8 +2036,8 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 			goto err;
 
 		/* update the QPt src information in CM msg */
-		m_cm->msg.saddr1.ep_map = MIC_XSOCK_DEV;
-		m_cm->msg.saddr2.ep_map = MIC_XSOCK_DEV;
+		m_cm->msg.saddr1.ep_map = smd->md->addr.ep_map;
+		m_cm->msg.saddr2.ep_map = smd->md->addr.ep_map;
 		m_cm->msg.saddr2.qpn = htonl(m_cm->m_qp->ib_qp2->qp_num);
 		m_cm->msg.saddr2.qp_type = m_cm->m_qp->qp_attr2.qp_type;
 		m_cm->msg.saddr2.lid = m_cm->smd->md->addr.lid;
@@ -1752,7 +2055,7 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 		mpxy_unlock(&smd->qprlock);
 
 	/* MXS -> MXS, proxy-in both sides, local, no QP's, SCIF services only */
-	} else if ((MXS_EP(&m_cm->md->addr) && MXS_EP(&m_cm->msg.daddr1)) &&
+	} else if ((MXF_EP(&m_cm->md->addr) && MXF_EP(&m_cm->msg.daddr1)) &&
 		   (system_guid == m_cm->msg.sys_guid)) {
 		mlog(2, " MXS -> MXS local - MODE NOT SUPPORTED, running MXS -> MXS remote mode\n");
 
@@ -1760,8 +2063,8 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 		dgid = (union ibv_gid *)m_cm->msg.daddr2.gid;
 		dqpn = m_cm->msg.daddr2.qpn;
 		dlid = m_cm->msg.daddr2.lid;
-		m_cm->msg.saddr1.ep_map = MIC_XSOCK_DEV;
-		m_cm->msg.saddr2.ep_map = MIC_XSOCK_DEV;
+		m_cm->msg.saddr1.ep_map = smd->md->addr.ep_map;
+		m_cm->msg.saddr2.ep_map = smd->md->addr.ep_map;
 
 	/* MSS -> MSS,MXS,HOST - fabric, TX: QP2->QP1 on mpxyd and RX: QP1->QP2 on MIC */
 	} else {
@@ -1769,8 +2072,12 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 
 		/* KL to KL, QP1->QP2 and QP1<-QP2 */
 		/* update the QPt src information in CM msg, QPr updated on MIC */
-		m_cm->msg.saddr1.ep_map = MIC_SSOCK_DEV;
-		m_cm->msg.saddr2.ep_map = MIC_SSOCK_DEV;
+		m_cm->msg.saddr1.ep_map = smd->md->addr.ep_map;
+		if (smd->md->m_lid) {
+			m_cm->msg.saddr1.lid = smd->md->m_lid;
+			memcpy(&m_cm->msg.saddr1.gid[0], &m_cm->smd->md->m_gid, 16);
+		}
+		m_cm->msg.saddr2.ep_map = smd->md->addr.ep_map;
 		m_cm->msg.saddr2.qpn = htonl(m_cm->m_qp->ib_qp2->qp_num);
 		m_cm->msg.saddr2.qp_type = m_cm->m_qp->qp_attr2.qp_type;
 		m_cm->msg.saddr2.lid = m_cm->smd->md->addr.lid;
@@ -1781,7 +2088,7 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 		dqpn = m_cm->msg.daddr1.qpn;
 		dlid = m_cm->msg.daddr1.lid;
 
-		if (MXS_EP(&m_cm->msg.daddr1)) {
+		if (MXF_EP(&m_cm->msg.daddr1)) {
 			if (m_pi_create_wc_q(m_cm->m_qp, mcm_rx_entries))
 				goto err;
 
@@ -1800,6 +2107,11 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 	m_cm->msg.sys_guid = rand();
 #endif
 
+	/* Set QP MTU, if negotiated. 2K for compatibility */
+	m_cm->m_qp->mtu = m_cm->msg.mtu ?
+			  min(m_cm->msg.mtu, m_cm->md->dev_attr.mtu):
+			  m_cm->md->mtu_env ? m_cm->md->mtu_env : IBV_MTU_2048;
+
 	if (qp) {
 		if (mcm_modify_qp(qp, IBV_QPS_RTR, dqpn, dlid, dgid))
 			goto err;
@@ -1815,8 +2127,9 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 			goto err;
 	}
 
-	/* send RTU on wire, monitor for retries */
+	/* send REP on wire, monitor for retries */
 	m_cm->state = MCM_RTU_PENDING;
+	m_cm->msg.mtu = m_cm->m_qp->mtu; /* send negotiated MTU */
 	mpxy_unlock(&m_cm->lock);
 	mcm_cm_rep_out(m_cm);
 	return 0;
@@ -1927,22 +2240,20 @@ static int mix_proxy_out(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, mcm_qp_t *m_qp
 	mpxy_lock(&m_qp->txlock);
 	if (((m_qp->wr_hd + 1) & m_qp->wr_end) == m_qp->wr_tl) { /* full */
 		ret = ENOMEM;
+		mlog(0, " ERR: WR full hd %d tl %d\n", m_qp->wr_hd, m_qp->wr_tl);
 		goto bail;
 	}
 	m_qp->wr_hd = (m_qp->wr_hd + 1) & m_qp->wr_end; /* move hd */
 	m_wr = (struct mcm_wr *)(m_qp->wr_buf + (m_qp->wr_sz * m_qp->wr_hd));
 
-	mlog(4, " inline, m_wr %p m_sge %p len %d hd %d tl %d\n",
-		m_wr, m_wr->sg, len, m_qp->wr_hd, m_qp->wr_tl);
-
 	/* IB rdma write WR */
 	const_ib_rw(&m_wr->wr, &pmsg->wr, m_wr->sg);
 	m_wr->wr.sg_list = m_wr->sg;
 	m_wr->wr.num_sge = len ? 1:0;
 
-	mlog(4, " INLINE m_wr (%p)raddr %p rkey 0x%llx, ib_wr raddr %p rkey 0x%llx \n",
-		&pmsg->wr.wr.rdma.remote_addr, pmsg->wr.wr.rdma.remote_addr, pmsg->wr.wr.rdma.rkey,
-		&m_wr->wr.wr.rdma.remote_addr, m_wr->wr.wr.rdma.remote_addr, m_wr->wr.wr.rdma.rkey);
+	mlog(4, " INLINE m_wr[%d] %p raddr %p rkey 0x%x, ib_wr raddr %p rkey 0x%x %d bytes\n",
+		m_qp->wr_hd, m_wr, pmsg->wr.wr.rdma.remote_addr, pmsg->wr.wr.rdma.rkey,
+		m_wr->wr.wr.rdma.remote_addr, m_wr->wr.wr.rdma.rkey, len);
 
 	/* M_WR */
 	m_wr->org_id = pmsg->wr.wr_id;
@@ -2031,7 +2342,7 @@ retry_mr:
 			" pcnt %d sg_rate %d hd %d tl %d sz %d m_idx %x\n",
 			m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid,
 			m_qp->r_entry.tid,
-			(MXS_EP(&m_qp->cm->msg.daddr1)) ? "po_pi":"po_direct",
+			(MXF_EP(&m_qp->cm->msg.daddr1)) ? "po_pi":"po_direct",
 			m_qp, m_wr, m_wr->wr.wr_id, m_wr->wr.send_flags,
 			m_qp->post_cnt,	mcm_rw_signal, m_qp->wr_hd, m_qp->wr_tl,
 			m_wr->wr.sg_list->length, m_wr->m_idx);
@@ -2048,7 +2359,6 @@ retry_mr:
 			goto bail;
 		}
 	}
-	mlog(4, " inline data rcv'ed %d bytes\n", len);
 
 	if ((smd->md->indata) && (len <= mcm_ib_inline))
 		m_wr->wr.send_flags |= IBV_SEND_INLINE;
@@ -2219,6 +2529,12 @@ int mix_scif_recv(mcm_scif_dev_t *smd, scif_epd_t scif_ep)
 	case MIX_QP_FREE:
 		ret = mix_qp_destroy(smd, phdr);
 		break;
+	case MIX_PZ_CREATE:
+		ret = mix_pz_create(smd, (dat_mix_pz_t *)phdr);
+		break;
+	case MIX_PZ_FREE:
+		ret = mix_pz_destroy(smd, (dat_mix_pz_t *)phdr);
+		break;
 	case MIX_CQ_CREATE:
 		ret = mix_cq_create(smd, (dat_mix_cq_t *)phdr);
 		break;
@@ -2243,6 +2559,12 @@ int mix_scif_recv(mcm_scif_dev_t *smd, scif_epd_t scif_ep)
 	case MIX_PROV_ATTR:
 		ret = mix_prov_attr(smd, (dat_mix_attr_t *)phdr);
 		break;
+	case MIX_QUERY_DEVICE:
+		ret = mix_query_device(smd, (dat_mix_device_attr_t *)phdr);
+		break;
+	case MIX_QUERY_PORT:
+		ret = mix_query_port(smd, (dat_mix_port_attr_t *)phdr);
+		break;
 	case MIX_CM_REQ:
 		ret = mix_cm_req_out(smd, (dat_mix_cm_t *)phdr, scif_ep);
 		break;
