@@ -416,7 +416,7 @@ int m_pi_prep_rcv_q(struct mcm_qp *m_qp)
 	int i;
 
 	/* MXS -> MSS or HST, PI service will be on QP1 */
-	if (MXS_EP(&m_qp->smd->md->addr) &&
+	if (MXF_EP(&m_qp->smd->md->addr) &&
 	   (MSS_EP(&m_qp->cm->msg.daddr1) || HST_EP(&m_qp->cm->msg.daddr1)))
 	        ib_qp = m_qp->ib_qp1;
 	else
@@ -476,11 +476,10 @@ static int m_pi_send_wc(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, int status
 	int wc_idx, ret;
 
 	mlog(0x10,"[%d:%d:%d] WC_rem: wr_rx[%d] %p wc_hd %d flgs %x WR_r tl %d-%d"
-		  " wt %d hd %d wr_id %Lx org_id %Lx\n",
+		  " wt %d hd %d oid %Lx st %d\n",
 		m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid, m_qp->r_entry.tid,
 		wr_rx->w_idx, wr_rx, m_qp->wc_hd_rem, wr_rx->flags, m_qp->wr_tl_r,
-		wr_rx->w_idx, m_qp->wr_tl_r_wt, m_qp->wr_hd_r, wr_rx->wr.wr_id,
-		wr_rx->org_id);
+		wr_rx->w_idx, m_qp->wr_tl_r_wt, m_qp->wr_hd_r, wr_rx->org_id, status);
 
 	/* local WR and remote WR are serialized, should never reach tail of remote WR */
 	if (((m_qp->wc_hd_rem + 1) & m_qp->wrc.wc_end) == m_qp->wc_tl_rem) {
@@ -529,13 +528,13 @@ static int m_pi_send_wc(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, int status
 		wr.send_flags |= IBV_SEND_INLINE;
 
 	mlog(4, " WC: RW_imm post: wr_id[%d] %Lx sglist %p sge %d op %d flgs %x"
-		" idata %x WR_rem = raddr %p rkey %x ln %d op %x\n",
+		" idata %x WR_rem = raddr %p rkey %x io_ln %d op %x\n",
 		wr_rx->w_idx, wr.wr_id, wr.sg_list, wr.num_sge, wr.opcode,
 		wr.send_flags, ntohl(wr.imm_data), wr.wr.rdma.remote_addr,
-		wr.wr.rdma.rkey, sge.length, wr_rx->wr.opcode);
+		wr.wr.rdma.rkey, wr_rx->sg[0].length, wr_rx->wr.opcode);
 
 	/* MXS -> MSS or HST, PI service will be on QP1 */
-	if (MXS_EP(&m_qp->smd->md->addr) &&
+	if (MXF_EP(&m_qp->smd->md->addr) &&
 	   (MSS_EP(&m_qp->cm->msg.daddr1) || HST_EP(&m_qp->cm->msg.daddr1)))
 	        ib_qp = m_qp->ib_qp1;
 	else
@@ -562,7 +561,7 @@ static int m_pi_send_wc(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, int status
 }
 
 /* called with rxlock, process all RR's up to signal marker at wr_last */
-static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, struct ibv_wc *wc)
+static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig)
 {
 	mcm_scif_dev_t *smd = m_qp->smd;
 	struct mcm_wr_rx *wr_rx;
@@ -576,9 +575,20 @@ static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, str
 	while (m_qp->pi_rr_cnt) { /* RR's pending */
 		wr_rx = (struct mcm_wr_rx *)(m_qp->wrc.wr_addr + (m_qp->wrc.wr_sz * wr_idx));
 
-		if (!(wr_rx->flags & M_READ_POSTED)) {
-			/* reached RR signaled marker, or head pointer */
-			if (wr_idx == wr_sig->w_idx || wr_idx == m_qp->wr_hd_r)
+		/* SCIF sync required on IB RW, multiple SCIF writes are not ordered */
+		if (m_qp->post_cnt_wt &&
+		    (wr_rx->flags & (M_SEND_FS|M_SEND_LS)) &&
+		    (!(wr_rx->flags & (M_READ_WRITE_TO_DONE|M_READ_WRITE_TO)))) {
+			break;
+		}
+
+		if (!(wr_rx->flags & M_READ_DONE)) {
+			/* reached head pointer */
+			if (wr_idx == m_qp->wr_hd_r)
+				break;
+
+			/* maintain order */
+			if (wr_rx->flags & M_READ_POSTED)
 				break;
 
 			wr_idx = (wr_idx + 1) & m_qp->wrc.wr_end; /* next WR */
@@ -589,9 +599,8 @@ static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, str
 		if (wr_rx == wr_sig)
 			mcm_qp_prof_ts(m_qp, MCM_QP_IB_RR, wr_rx->time, wr_rx->qcnt, wr_cnt);
 #endif
-		mlog(4, " WR_rx[%d-%d] %p m_qp %p wc %p wc->op %x wr_rx->wr.op %x\n",
-			wr_rx->w_idx, wr_sig->w_idx, wr_rx, m_qp, wc,
-			wc->opcode, wr_rx->wr.opcode);
+		mlog(4, " WR_rx[%d-%d] %p m_qp %p wr_rx->wr.op %x\n",
+			wr_rx->w_idx, wr_sig->w_idx, wr_rx, m_qp, wr_rx->wr.opcode);
 
 		m_qp->pi_rr_cnt--; /* rdma read complete */
 		MCNTR(smd->md, MCM_QP_READ_DONE);
@@ -622,11 +631,27 @@ static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, str
 			sg_len = wr_rx->sg[2].length;
 			r_off = m_pi_mr_trans(smd, wr_rx->wr.wr.rdma.remote_addr,
 					      wr_rx->wr.wr.rdma.rkey, sg_len);
-			if (!r_off)
+			if (!r_off && sg_len)
 				goto bail;
 
-			mlog(4, " RDMA_WRITE op: wr_rx[%d] %p -> scif r_off %Lx len %d\n",
-				 wr_rx->w_idx, wr_rx, r_off, sg_len, 0);
+			if (!sg_len) {  /* 0 byte rdma write, no scif */
+				mlog(1, " RDMA_WRITE op: wr_rx[%d] %p,"
+					 " raddr %p rkey %x 0 bytes\n",
+					 wr_rx->w_idx, wr_rx,
+					 wr_rx->wr.wr.rdma.remote_addr,
+					 wr_rx->wr.wr.rdma.rkey);
+
+				m_qp->post_cnt_wt++;
+				wr_rx->flags &= ~M_READ_DONE;
+				wr_rx->flags |= M_READ_WRITE_TO;
+				wr_rx->wr.wr_id = wr_rx->org_id; /* mark done */
+
+				if (wr_idx == m_qp->wr_hd_r)
+					break;
+
+				wr_idx = (wr_idx + 1) & m_qp->wrc.wr_end; /* next WR */
+				continue;
+			}
 		}
 
 		/* sg[0] entry == proxy-out buffer, src for IB RR */
@@ -650,14 +675,24 @@ static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, str
 			wr_rx->time = mcm_ts_us();
 			wr_rx->qcnt = m_qp->post_cnt_wt;
 #endif
-			if (w_len < 256)
-				wt_flag = SCIF_RMA_USECPU;
+			if (wr_rx->flags & M_SEND_LS)
+				wt_flag |= SCIF_RMA_ORDERED;
 
 			ret = scif_writeto(smd->scif_tx_ep, l_off, w_len, r_off, wt_flag);
 
 			if (ret) {
-				mlog(0, " ERR: scif_sendto, ret %d err: %d %s\n",
+				mlog(0," [%d:%d:%d] ERR: scif_sendto, ret %d err: %d %s\n",
+					smd->md->mc->scif_id, smd->entry.tid, m_qp->r_entry.tid,
 					ret, errno, strerror(errno));
+				mlog(0," PI: wc %d rr %d stall %d wt %d\n",
+				     m_qp->pi_rw_cnt, m_qp->pi_rr_cnt,
+				     m_qp->stall_cnt_rr, m_qp->post_cnt_wt);
+				mlog(0," PO: wr %d wr_rem %d pst_sig %d cmp_sig %d\n",
+				     m_qp->wr_pp, m_qp->wr_pp_rem,
+				     m_qp->post_sig_cnt, m_qp->comp_cnt);
+				mlog(0, " WR_rx[%d] %p l_o %Lx r_o %Lx rb 0x%x-0x%x ln %d id %Lx tl %d hd %d\n",
+					wr_rx->w_idx, wr_rx, l_off, r_off, l_start, l_end, w_len, wr_rx->org_id,
+					m_qp->wr_tl_r, m_qp->wr_hd_r);
 				goto bail;
 			}
 			MCNTR(smd->md, MCM_SCIF_WRITE_TO);
@@ -688,23 +723,33 @@ static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, str
 			goto bail;
 		}
 		MCNTR(smd->md, MCM_SCIF_SIGNAL);
-		wr_rx->flags &= ~M_READ_POSTED; /* reset READ_POSTED */
-		wr_rx->flags |= M_READ_DONE;
+		wr_rx->flags &= ~M_READ_DONE;
 		wr_rx->flags |= M_READ_WRITE_TO;
 		m_qp->post_cnt_wt++;
 
-		/* reached RR signaled marker, or head */
-		if (wr_idx == wr_sig->w_idx || wr_idx == m_qp->wr_hd_r)
+		/* reached head */
+		if (wr_idx == m_qp->wr_hd_r)
 			break;
 
 		wr_idx = (wr_idx + 1) & m_qp->wrc.wr_end; /* next WR */
 	}
 	return;
 bail:
-	/* report error via WC back to proxy-out */
-	mlog(0, " ERR: writeto: wr_rx[%d] %p -> raddr %Lx rkey %x (scif r_off %Lx) len %d\n",
-		wr_rx->w_idx, wr_rx, wr_rx->wr.wr.rdma.remote_addr,
-		wr_rx->wr.wr.rdma.rkey, r_off, sg_len);
+	/* report error via WC back to proxy-out, all pending WRs */
+	wr_idx = m_qp->wr_tl_r_wt;
+	do {
+		wr_rx = (struct mcm_wr_rx *)(m_qp->wrc.wr_addr + (m_qp->wrc.wr_sz * wr_idx));
+
+		mlog(0, " ERR: wr_rx[%d] %p -> IB raddr %Lx %x"
+			" SCIF r_o %Lx, ln %d fl %x wt_pnd %d\n",
+			wr_rx->w_idx, wr_rx, wr_rx->wr.wr.rdma.remote_addr,
+			wr_rx->wr.wr.rdma.rkey, r_off, sg_len, wr_rx->flags,
+			m_qp->post_cnt_wt);
+
+		m_pi_send_wc(m_qp, wr_rx, IBV_WC_REM_ACCESS_ERR);
+		wr_idx = (wr_idx + 1) & m_qp->wrc.wr_end; /* next WR */
+
+	} while (wr_idx != m_qp->wr_hd_r);
 
 	return;
 }
@@ -722,25 +767,23 @@ void m_pi_req_event(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, struct ibv_wc
 {
 	mlog(4, " WR_rx[%d] %p %s complete po-addr=%p ln=%d, key=%x ctx=%Lx\n",
 		wr_rx->w_idx, wr_rx,
-		wr_rx->wr.opcode == IBV_WR_RDMA_READ ? "RR":"RW_IMM WC",
+		type == WRID_RX_RR ? "RR":"RW_IMM WC",
 		wr_rx->sg[0].addr, wr_rx->sg[0].length,
 		wr_rx->sg[0].lkey, wr_rx->context);
 
 	if (wc->status && (wc->status != IBV_WC_WR_FLUSH_ERR)) {
-		char *sbuf = (char*)wr_rx->sg[1].addr;
-
-		mlog(0," WR ERR: st %d, vn %x pst %d cmp %d qstate 0x%x\n",
-			wc->status, wc->vendor_err, m_qp->post_cnt,
-			m_qp->comp_cnt, m_qp->ib_qp2->state);
-		mlog(0, " WR ERR: wr_rx %p laddr %p=0x%x - %p=0x%x, len=%d, lkey=%x\n",
-			wr_rx,  sbuf, sbuf[0], &sbuf[wr_rx->sg[1].length],
-			sbuf[wr_rx->sg[1].length], wr_rx->sg[1].length, wr_rx->sg[1].lkey);
-		mlog(0, " WR ERR: wr_id %Lx sglist %p sge %d op %d flgs"
-			" %d idata 0x%x raddr %p rkey %x saddr %p key %x ln %d\n",
-		     wr_rx->org_id, wr_rx->sg, wr_rx->wr.num_sge,
-		     wr_rx->wr.opcode, wr_rx->wr.send_flags, wr_rx->wr.imm_data,
-		     wr_rx->wr.wr.rdma.remote_addr, wr_rx->wr.wr.rdma.rkey,
-		     wr_rx->sg[0].addr, wr_rx->sg[0]. lkey,wr_rx->sg[0].length);
+		mlog(0," WR ERR: %s st %d, vn %x rr %d wt %d WC[%d] %p %x %d\n",
+			type == WRID_RX_RR ? "RR":"RW_IMM WC",
+			wc->status, wc->vendor_err, m_qp->post_cnt_rr,
+			m_qp->post_cnt_wt, m_qp->wc_hd_rem,
+			m_qp->wrc_rem.wc_addr, m_qp->wrc_rem.wc_rkey,
+			m_qp->wrc_rem.wc_sz);
+		mlog(0, " WR ERR: wr_rx[%d] oid %Lx sge %d op %d flgs %d"
+			" imm 0x%x SRC %p %x ln %d, DST %p %x ln %d\n",
+			wr_rx->w_idx, wr_rx->org_id, wr_rx->wr.num_sge,
+		        wr_rx->wr.opcode, wr_rx->wr.send_flags, wr_rx->wr.imm_data,
+		        wr_rx->sg[0].addr, wr_rx->sg[0].lkey, wr_rx->sg[0].length,
+		        wr_rx->sg[1].addr, wr_rx->sg[1].lkey, wr_rx->sg[1].length);
 
 		/* send WC with ERR to RW initiator, hold rxlock */
 		mpxy_lock(&m_qp->rxlock);
@@ -752,7 +795,9 @@ void m_pi_req_event(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, struct ibv_wc
 	/* RR complete, ready for SCIF_writeto to complete RW or SR */
 	if (type == WRID_RX_RR) {
 		mpxy_lock(&m_qp->rxlock);
-		m_pi_post_writeto(m_qp, wr_rx, wc);
+		wr_rx->flags &= ~M_READ_POSTED; /* reset READ_POSTED */
+		wr_rx->flags |= M_READ_DONE;
+		m_pi_post_writeto(m_qp, wr_rx);
 		mpxy_unlock(&m_qp->rxlock);
 		write(m_qp->smd->md->mc->rx_pipe[1], "w", sizeof "w"); /* signal rx_thread */
 		sched_yield();
@@ -787,6 +832,20 @@ static void m_pi_post_read(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx)
 	/* shared proxy-in buffer, device level serialization */
 	mpxy_lock(&smd->rblock);
 
+	/* Stall if data request arrives before RTU */
+	if (m_qp->cm && m_qp->cm->state != MCM_CONNECTED && !m_qp->post_cnt_rr) {
+		mlog(1, " [%d:%d:%d] WARN: WR[%d] %p PAUSED !CONN: po-addr=%p"
+			" ln=%d, key=%x ctx=%Lx\n",
+			m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid,
+			m_qp->r_entry.tid, wr_rx->w_idx, wr_rx,
+			wr_rx->sg[0].addr, wr_rx->sg[0].length,
+			wr_rx->sg[0].lkey, wr_rx->context);
+		wr_rx->flags |= M_READ_PAUSED;
+		m_qp->stall_cnt_rr++;
+		mpxy_unlock(&smd->rblock);
+		return;
+	}
+
 	/* slice out proxy buffer for this segment */
 	l_start = ALIGN_64(smd->m_hd_r);
 	if ((l_start + l_len) > smd->m_len_r)
@@ -901,7 +960,7 @@ static void m_pi_post_read(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx)
 	mpxy_unlock(&smd->rblock);
 
 	/* MXS -> MSS or HST, PI service will be on QP1 */
-	if (MXS_EP(&m_qp->smd->md->addr) &&
+	if (MXF_EP(&m_qp->smd->md->addr) &&
 	   (MSS_EP(&m_qp->cm->msg.daddr1) || HST_EP(&m_qp->cm->msg.daddr1)))
 	        ib_qp = m_qp->ib_qp1;
 	else
@@ -911,25 +970,35 @@ static void m_pi_post_read(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx)
 	wr_rx->time = mcm_ts_us();
 	wr_rx->qcnt = m_qp->pi_rr_cnt;
 #endif
-	wr_rx->flags |= M_READ_POSTED;
-	errno = 0;
-	ret = ibv_post_send(ib_qp, &ib_wr, &bad_wr);
-	if (ret)
-		goto bail;
 
-	m_qp->pi_rr_cnt++;
-	m_qp->post_cnt_rr++;
-	MCNTR(smd->md, MCM_QP_READ);
+	if (wr_rx->flags & M_PROXY_INLINE) {
+		mlog(0x10, "wr_rx flag PROXY_INLINE is set. data len %d\n", wr_rx->sg[0].length);
+		memcpy((void *)rbuf, wr_rx->inline_data, wr_rx->sg[0].length);
+		m_qp->pi_rr_cnt++;
+		m_qp->post_cnt_rr++;
+		wr_rx->flags |= M_READ_DONE;
+		m_pi_post_writeto(m_qp, wr_rx);
+	}
+	else {
+		errno = 0;
+		wr_rx->flags |= M_READ_POSTED;
+		ret = ibv_post_send(ib_qp, &ib_wr, &bad_wr);
+		if (ret)
+			goto bail;
 
-	mlog(0x10, "[%d:%d:%d] WR[%d] %p RR(%d,%d,%d): wr_id %Lx qn %x flgs %x,%x ln %d "
-		   "r_addr,key %Lx %x to l_addr,key %Lx %x tl %d hd %d, m_idx %x\n",
-		smd->md->mc->scif_id, smd->entry.tid, m_qp->r_entry.tid,
-		wr_rx->w_idx, wr_rx, m_qp->post_cnt_rr, m_qp->stall_cnt_rr,
-		m_qp->pi_rr_cnt, ib_wr.wr_id, ib_qp->qp_num, ib_wr.send_flags,
-		wr_rx->flags, l_len, ib_wr.wr.rdma.remote_addr,
-		ib_wr.wr.rdma.rkey, ib_wr.sg_list->addr, ib_wr.sg_list->lkey,
-		m_qp->wr_tl_r, m_qp->wr_hd_r, wr_rx->m_idx);
+		m_qp->pi_rr_cnt++;
+		m_qp->post_cnt_rr++;
+		MCNTR(smd->md, MCM_QP_READ);
 
+		mlog(0x10, "[%d:%d:%d] WR[%d] %p RR(%d,%d,%d): wr_id %Lx qn %x flgs %x,%x ln %d "
+			   "r_addr,key %Lx %x to l_addr,key %Lx %x tl %d hd %d, m_idx %x\n",
+			smd->md->mc->scif_id, smd->entry.tid, m_qp->r_entry.tid,
+			wr_rx->w_idx, wr_rx, m_qp->post_cnt_rr, m_qp->stall_cnt_rr,
+			m_qp->pi_rr_cnt, ib_wr.wr_id, ib_qp->qp_num, ib_wr.send_flags,
+			wr_rx->flags, l_len, ib_wr.wr.rdma.remote_addr,
+			ib_wr.wr.rdma.rkey, ib_wr.sg_list->addr, ib_wr.sg_list->lkey,
+			m_qp->wr_tl_r, m_qp->wr_hd_r, wr_rx->m_idx);
+	}
 	write(smd->md->mc->tx_pipe[1], "w", sizeof "w");
 	return;
 bail:
@@ -957,11 +1026,13 @@ buf_err:
 		ib_wr.send_flags, l_len, ib_wr.wr.rdma.remote_addr,
 		ib_wr.wr.rdma.rkey, ib_wr.sg_list->addr, ib_wr.sg_list->lkey,
 		m_qp->wr_tl_r, m_qp->wr_tl_r_wt, m_qp->wr_hd_r);
+
+	m_pi_send_wc(m_qp, wr_rx, IBV_WC_REM_ACCESS_ERR); /* report error */
 }
 
 void m_pi_rcv_event(struct mcm_qp *m_qp, wrc_idata_t *wrc)
 {
-	mlog(8," WRC id %x, type %x, flags %x\n", wrc->id, wrc->type, wrc->flags);
+	mlog(8," WRC id %d, type %x, flags %x\n", wrc->id, wrc->type, wrc->flags);
 	if (wrc->type == M_WR_TYPE) {
 		struct mcm_wr_rx *wr_rx;
 
@@ -974,9 +1045,10 @@ void m_pi_rcv_event(struct mcm_qp *m_qp, wrc_idata_t *wrc)
 		mcm_ntoh_wr_rx(wr_rx); /* received in network order, convert */
 		wr_rx->context = (uint64_t)(uintptr_t)m_qp;  /* local side QP context */
 
-		mlog(8," WR_rx[%d] %p org_id %Lx wc_tl_rem %d flgs %x wr.wr_id %Lx imm 0x%x\n",
+		mlog(8," WR_rx[%d] %p id %Lx wc_tl_rem %d flgs %x wr.wr_id %Lx imm 0x%x raddr %Lx,%x,%d\n",
 			wrc->id, wr_rx, wr_rx->org_id, wr_rx->w_idx, wr_rx->flags,
-			wr_rx->wr.wr_id, wr_rx->wr.imm_data);
+			wr_rx->wr.wr_id, wr_rx->wr.imm_data,
+			wr_rx->sg[0].addr, wr_rx->sg[0].lkey, wr_rx->sg[0].length);
 
 		mpxy_lock(&m_qp->rxlock);
 		m_qp->wc_tl_rem = wr_rx->w_idx; /* remote WC tail update in WR */
@@ -1089,7 +1161,7 @@ retry:
 			r_wr.wr_id = (uint64_t)(uintptr_t) m_qp;
 
 			/* MXS -> MSS or HST, PI service will be on QP1 */
-			if (MXS_EP(&m_qp->smd->md->addr) &&
+			if (MXF_EP(&m_qp->smd->md->addr) &&
 			   (MSS_EP(&m_qp->cm->msg.daddr1) || HST_EP(&m_qp->cm->msg.daddr1)))
 			        ib_qp = m_qp->ib_qp1;
 			else
@@ -1100,7 +1172,7 @@ retry:
 				err = ibv_post_recv(ib_qp, &r_wr, &r_err);
 				if (err) {
 					mlog(0,"ERR: qp %p (%s) qpn %x ibv_post_recv ret = %d %s\n",
-						m_qp, (MXS_EP(&m_qp->smd->md->addr) &&
+						m_qp, (MXF_EP(&m_qp->smd->md->addr) &&
 						MSS_EP(&m_qp->cm->msg.daddr1)) ? "QP1":"QP2",
 						m_qp->ib_qp2 ?
 						m_qp->ib_qp2->qp_num:m_qp->ib_qp1->qp_num,
@@ -1116,6 +1188,17 @@ retry:
 	goto retry;
 }
 
+/* Process scif_writeto DMAs waiting for previous WTs completions, order */
+void m_pi_pending_wt(struct mcm_qp *m_qp)
+{
+	struct mcm_wr_rx *wr_rx;
+
+	mpxy_lock(&m_qp->rxlock);
+	wr_rx = (struct mcm_wr_rx *)(m_qp->wrc.wr_addr + (m_qp->wrc.wr_sz * m_qp->wr_hd_r));
+	m_pi_post_writeto(m_qp, wr_rx);
+	mpxy_unlock(&m_qp->rxlock);
+}
+
 /*
  * Pending Proxy-in services for RDMA Writes from remote peer
  *
@@ -1129,18 +1212,10 @@ void m_pi_pending_wr(struct mcm_qp *m_qp, int *data)
 	struct mcm_wr_rx *wr_rx;
 	int wr_idx, wr_max, wr_cnt;
 
-	if (m_qp->cm && m_qp->cm->state != MCM_CONNECTED) {
-		if (m_qp->post_cnt_wt) {
-			mlog(8," !CONN: qp %p cm %p %s tl_r %d wt_tl_r %d hd_r %d pp %d st %d data %d\n",
-				m_qp, m_qp->cm, m_qp->cm ? mcm_state_str(m_qp->cm->state):"",
-				m_qp->wr_tl_r, m_qp->wr_tl_r_wt,
-				m_qp->wr_hd_r, m_qp->post_cnt_wt,
-				m_qp->stall_cnt_rr, *data);
-		}
-		return;
-	}
-
 	mpxy_lock(&m_qp->rxlock);
+	if (m_qp->cm && m_qp->cm->state != MCM_CONNECTED && !m_qp->post_cnt_rr)
+		goto done;
+
 	wr_max = mcm_wr_max;
 	wr_idx = m_qp->wr_tl_r_wt; /* last write_to marker */
 	wr_cnt = 0;
