@@ -103,7 +103,7 @@ int m_po_create_bpool(struct mcm_qp *m_qp, int entries)
 				 m_qp, entries * sizeof(mcm_wr_rx_t), entries);
 			goto err;
 		}
-		memset(m_qp->wr_buf, 0, entries * sizeof(mcm_wr_rx_t));
+		memset(m_qp->wr_buf_rx, 0, entries * sizeof(mcm_wr_rx_t));
 
 		m_qp->wr_buf_rx_mr = ibv_reg_mr(m_qp->smd->md->pd, (void*)m_qp->wr_buf_rx,
 						entries * sizeof(mcm_wr_rx_t),
@@ -274,9 +274,15 @@ static int m_po_send_pi(struct mcm_qp *m_qp, struct mcm_wr *m_wr, int wr_idx)
 		sge.lkey = m_qp->wr_buf_rx_mr->lkey;
 	}
 	sge.addr = (uint64_t)(uintptr_t) wr_rx_ptr;
-	sge.length = (uint32_t) sizeof(struct mcm_wr_rx);
+	sge.length = (uint32_t) m_qp->wrc_rem.wr_sz;
 
 	/* proxy m_wr over to remote m_wr_rem slot, remote will initiate RR and send back WC */
+	if (m_qp->p2p_data && m_wr->sg[0].length < m_qp->p2p_data) {
+		mlog(0x4, " Sending the proxy data ( len %d ) inside the WR.\n", m_wr->sg[0].length);
+		m_wr->flags |= M_PROXY_INLINE;
+		memcpy (wr_rx_ptr->inline_data, (void *)m_wr->sg[0].addr, m_wr->sg[0].length);
+	}
+
 	m_wr->flags |= M_SEND_PI;
 	mcm_hton_wr_rx(wr_rx_ptr, m_wr, m_qp->wc_tl); /* build rx_wr for wire transfer, send it */
 
@@ -313,7 +319,7 @@ static int m_po_send_pi(struct mcm_qp *m_qp, struct mcm_wr *m_wr, int wr_idx)
 
 
 	/* MXS -> MSS or HST, PI service will be on QP1 */
-	if (MXS_EP(&m_qp->smd->md->addr) &&
+	if (MXF_EP(&m_qp->smd->md->addr) &&
 	   (MSS_EP(&m_qp->cm->msg.daddr1) || HST_EP(&m_qp->cm->msg.daddr1)))
 	        ib_qp = m_qp->ib_qp1;
 	else
@@ -336,10 +342,11 @@ static int m_po_send_pi(struct mcm_qp *m_qp, struct mcm_wr *m_wr, int wr_idx)
 	ret = ibv_post_send(ib_qp, &wr, &bad_wr);
 	if (ret) {
 		mlog(0, " ERR: m_wr %p idx %d laddr=%p ln=%d lkey=%x flgs %x"
-			" tl %d hd %d pp %d sig %d\n",
+			" tl %d hd %d pp %d sig %d ret %d %s\n",
 			m_wr, wr_idx, sge.addr, sge.length, sge.lkey,
 			m_wr->flags, m_qp->wr_tl, m_qp->wr_hd,
-			m_qp->wr_pp_rem, m_qp->post_sig_cnt);
+			m_qp->wr_pp_rem, m_qp->post_sig_cnt,
+			ret, strerror(errno));
 		mlog(0, " ERR: wr_id %Lx %p sglist %p sge %d op %d flgs %x"
 			" idata 0x%x raddr %p rkey %x \n",
 			m_wr->wr.wr_id, m_wr->wr.sg_list,
@@ -443,15 +450,15 @@ void m_po_pending_wr(struct mcm_qp *m_qp, int *data)
 					m_wr->wr.send_flags &= ~IBV_SEND_SIGNALED;
 			}
 
-			if (!(MXS_EP(&m_qp->cm->msg.daddr1)) &&
+			if (!(MXF_EP(&m_qp->cm->msg.daddr1)) &&
 			     (m_wr->wr.send_flags & IBV_SEND_SIGNALED)) {
 				m_qp->post_sig_cnt++;
 
-				mlog(0x10, "[%d:%d:%d] %s_RW_post_sig: qp %p wr %p wr_id %p flgs 0x%x,"
+				mlog(0x10, "[%d:%d:%d] %s_RW_post_sig: WR[%d] qp %p wr_id %p flgs 0x%x,"
 					" pcnt %d sg_rate %d hd %d tl %d sz %d m_idx %x\n",
 					m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid, m_qp->r_entry.tid,
-					(MXS_EP(&m_qp->cm->msg.daddr1)) ? "po_pi":"po_direct",
-					m_qp, m_wr, m_wr->wr.wr_id, m_wr->wr.send_flags,
+					(MXF_EP(&m_qp->cm->msg.daddr1)) ? "po_pi":"po_direct",
+					m_wr->w_idx, m_qp, m_wr->wr.wr_id, m_wr->wr.send_flags,
 					m_qp->post_cnt,	mcm_rw_signal, m_qp->wr_hd, m_qp->wr_tl,
 					m_wr->wr.sg_list->length, m_wr->m_idx);
 #if MCM_PROFILE
@@ -463,7 +470,7 @@ void m_po_pending_wr(struct mcm_qp *m_qp, int *data)
 			m_wr->wr.wr_id = WRID_SET(m_wr, WRID_TX_RW);
 
 			errno = 0;
-			if (MXS_EP(&m_qp->cm->msg.daddr1)) /* remote PI */
+			if (MXF_EP(&m_qp->cm->msg.daddr1)) /* remote PI */
 				ret = m_po_send_pi(m_qp, m_wr, wr_idx);
 			else
 				ret = ibv_post_send(m_qp->ib_qp2, &m_wr->wr, &bad_wr);
@@ -475,7 +482,28 @@ void m_po_pending_wr(struct mcm_qp *m_qp, int *data)
 				wc.byte_len = m_wr->sg->length;
 				wc.status = ret ? IBV_WC_GENERAL_ERR : IBV_WC_SUCCESS;
 				wc.opcode = m_wr->wr.opcode == IBV_WR_SEND ? IBV_WC_SEND:IBV_WC_RDMA_WRITE;
+				if (m_wr->wr.opcode == IBV_WR_RDMA_WRITE_WITH_IMM)
+					wc.wc_flags = IBV_WC_WITH_IMM;
+				else
+					wc.wc_flags = 0;
 				wc.vendor_err = ret;
+				if (ret) {
+					mlog(0, "[%d:%d:%d] ERR %s_RW_post: WR[%d] wr_id %p flgs 0x%x,"
+						" pcnt %d sg_rate %d hd %d tl %d sz %d m_idx %x\n",
+						m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid,
+						m_qp->r_entry.tid,
+						(MXF_EP(&m_qp->cm->msg.daddr1)) ? "po_pi":"po_direct",
+						m_wr->w_idx, m_wr->wr.wr_id, m_wr->wr.send_flags,
+						m_qp->post_cnt,	mcm_rw_signal, m_qp->wr_hd, m_qp->wr_tl,
+						m_wr->wr.sg_list->length, m_wr->m_idx);
+					mlog(0, "[%d:%d:%d] ERR wr_id %Lx next %p sglist %p sge %d op %d flgs"
+						" %d idata 0x%x raddr %p rkey %x \n",
+						m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid,
+						m_qp->r_entry.tid, m_wr->wr.wr_id, m_wr->wr.next,
+						m_wr->wr.sg_list, m_wr->wr.num_sge, m_wr->wr.opcode,
+						m_wr->wr.send_flags, m_wr->wr.imm_data,
+						m_wr->wr.wr.rdma.remote_addr, m_wr->wr.wr.rdma.rkey);
+				}
 				mix_dto_event(m_qp->ib_qp2->send_cq->cq_context, &wc, 1);
 			}
 
@@ -554,7 +582,7 @@ int m_po_proxy_data(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, struct mcm_qp *m_qp
 	off_t l_off, r_off;
 	uint64_t total_offset;
 	int  l_start, l_end, l_len, cacheln_off, seg_len;
-	struct mcm_wr *m_wr;
+	struct mcm_wr *m_wr = NULL;
 	struct ibv_sge *m_sge;
 
 	mlog(4, " q_id %d, q_ctx %p, len %d, wr_id %p, sge %d, op %x flgs %x wr_idx %d\n",
@@ -598,7 +626,7 @@ int m_po_proxy_data(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, struct mcm_qp *m_qp
 		}
 		write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
 		mpxy_unlock(&m_qp->txlock);
-		sched_yield();
+		sleep_usec(1000);
 		mpxy_lock(&m_qp->txlock);
 	}
 	if (retries) {
@@ -789,13 +817,12 @@ retry_mr:
 						mpxy_unlock(&smd->tblock);
 						goto bail;
 					}
-					mlog(0x10, "[%d:%d:%d] %s_RF_post_sig: qp %p wr %p wr_id %p flgs 0x%x,"
-						" pcnt %d sg_rate %d hd %d tl %d sz %d m_idx %x\n",
-						m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid,
-						m_qp->r_entry.tid,
-						(MXS_EP(&m_qp->cm->msg.daddr1)) ? "po_pi":"po_direct",
-						m_qp, m_wr, pmsg->wr.wr_id, m_wr->wr.send_flags,
-						m_qp->post_cnt,	mcm_rw_signal, m_qp->wr_hd, m_qp->wr_tl,
+					mlog(0x10, "[%d:%d:%d] %s_RF_post_sig: WR[%d] qp %p wr_id %p flgs 0x%x,"
+						" sg_rate %d hd %d tl %d sz %d m_idx %x\n",
+						m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid, m_qp->r_entry.tid,
+						(MXF_EP(&m_qp->cm->msg.daddr1)) ? "po_pi":"po_direct",
+						m_wr->w_idx, m_qp, pmsg->wr.wr_id, m_wr->wr.send_flags,
+						mcm_rw_signal, m_qp->wr_hd, m_qp->wr_tl,
 						m_wr->wr.sg_list->length, m_wr->m_idx);
 				}
 
@@ -899,10 +926,26 @@ bail:
 		struct dat_mix_wc wc;
 
 		wc.wr_id = pmsg->wr.wr_id;
-		wc.byte_len = 0;
+		wc.byte_len = pmsg->sge[0].length;
 		wc.status = wc_err;
 		wc.opcode = pmsg->wr.opcode == IBV_WR_SEND ? IBV_WC_SEND:IBV_WC_RDMA_WRITE;
+		if (pmsg->wr.opcode == IBV_WR_RDMA_WRITE_WITH_IMM)
+			wc.wc_flags = IBV_WC_WITH_IMM;
+		else
+			wc.wc_flags = 0;
 		wc.vendor_err = ret;
+
+		mlog(0, "[%d:%d:%d] ERR %s_RF_post: WR[%d] qp %p wr_id %p, "
+			" post %d hd %d tl %d sz %d \n",
+			m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid, m_qp->r_entry.tid,
+			(MXF_EP(&m_qp->cm->msg.daddr1)) ? "po_pi":"po_direct",
+			m_wr ? m_wr->w_idx:0, m_qp, m_wr, pmsg->wr.wr_id,
+			m_qp->post_cnt,	m_qp->wr_hd, m_qp->wr_tl, wc.byte_len);
+		mlog(0, "[%d:%d:%d] ERR m_wr: raddr %Lx rkey 0x%x, ib_wr: raddr %Lx rkey 0x%x\n",
+			m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid, m_qp->r_entry.tid,
+			pmsg->wr.wr.rdma.remote_addr, pmsg->wr.wr.rdma.rkey,
+			m_wr ? m_wr->wr.wr.rdma.remote_addr:0, m_wr ? m_wr->wr.wr.rdma.rkey:0);
+
 		mix_dto_event(m_qp->ib_qp2->send_cq->cq_context, &wc, 1);
 	}
 
@@ -929,15 +972,18 @@ void m_po_wc_event(struct mcm_qp *m_qp, struct mcm_wc_rx *wc_rx, int wc_idx)
 		     wc_rx->wr_idx, m_qp->wr_end);
 		return;
 	}
-	m_cq = m_qp->m_cq_rx;
+	m_cq = m_qp->m_cq_tx;
 
 	if (wc_rx->wc.status == IBV_WC_SUCCESS) {
 		mlog(8," WC_RX: SUCCESS m_wr %p idx %d=%d flags 0x%x \n",
 			m_wr, m_wr->w_idx, wc_rx->wr_idx, m_wr->flags);
 		if (m_wr->flags & M_SEND_CN_SIG) {
+			m_wr->flags &= ~M_SEND_CN_SIG;
 			wc_ev.wr_id = m_wr->org_id;
 			wc_ev.status = IBV_WC_SUCCESS;
 			wc_ev.byte_len = wc_rx->wc.byte_len;
+			wc_ev.opcode = wc_rx->wc.opcode;
+			wc_ev.wc_flags = wc_rx->wc.wc_flags;
 			event++;
 		}
 	}
@@ -1053,6 +1099,7 @@ retry:
 		       i+1, ret, wc[i].wr_id, wc[i].status, wc[i].opcode,
 		       wc[i].vendor_err, wc[i].byte_len, wr_type);
 
+		/* Proxy_in -> */
 		if ((wr_type == WRID_RX_RR) || (wr_type == WRID_RX_RW_IMM)) {
 			m_wr_rx = (struct mcm_wr_rx *)WRID_ADDR(wc[i].wr_id);
 			assert(m_wr_rx);
@@ -1070,19 +1117,26 @@ retry:
 			       wc[i].vendor_err, wc[i].byte_len, wr_type);
 			continue;
 		}
+		/* Proxy_out ->  */
 		m_wr = (struct mcm_wr *)WRID_ADDR(wc[i].wr_id);
 		m_qp = (struct mcm_qp *)m_wr->context;
-		if (MSS_EP(&m_qp->cm->msg.daddr1))
+		if (m_qp->cm && !MXF_EP(&m_qp->cm->msg.daddr1))
 			m_qp->comp_cnt++;
 		MCNTR(m_qp->smd->md, MCM_QP_WRITE_DONE);
 
-		mlog(8," wr_id[%d of %d] m_wr %p m_qp %p\n", i, ret, m_wr, m_qp);
+		mlog(8," wr_id[%d of %d] m_wr %p m_qp %p\n", i+1, ret, m_wr, m_qp);
 
 		if (wc[i].status == IBV_WC_SUCCESS) {
 			if (m_wr->flags & M_SEND_CN_SIG) {
+				m_wr->flags &= ~M_SEND_CN_SIG;
 				wc_ev[num].wr_id = m_wr->org_id;
 				wc_ev[num].status = IBV_WC_SUCCESS;
 				wc_ev[num].byte_len = wc[i].byte_len;
+				wc_ev[num].opcode = m_wr->wr.opcode == IBV_WR_SEND ? IBV_WC_SEND:IBV_WC_RDMA_WRITE;
+				if (m_wr->wr.opcode == IBV_WR_RDMA_WRITE_WITH_IMM)
+					wc_ev[num].wc_flags = IBV_WC_WITH_IMM;
+				else
+					wc_ev[num].wc_flags = 0;
 				num++;
 			}
 		}
@@ -1112,7 +1166,7 @@ retry:
 		}
 
 		/* Can move PO buffer tail if no peer PI service */
-		if (!MXS_EP(&m_qp->cm->msg.daddr1)) {
+		if (!MXF_EP(&m_qp->cm->msg.daddr1)) {
 #if MCM_PROFILE
 			mcm_qp_prof_ts( m_qp, MCM_QP_IB_RW,
 					m_wr->wr.wr.atomic.swap,
