@@ -146,8 +146,8 @@ static int init_scif()
 		scif_close(scif_listen_ep);
 		return -1;
 	}
-	mlog(1," MPXYD: Listening on reserved SCIF OFED port %d, listen_EP %d, backlog %d\n",
-		(uint16_t)scif_id.port, scif_sport, scif_listen_qlen);
+	mlog(1," MPXYD: Listening on reserved SCIF OFED port %d, backlog %d\n",
+		(uint16_t)scif_id.port, scif_listen_qlen);
 
 	return 0;
 }
@@ -221,11 +221,23 @@ static struct ibv_context *open_ib_device(struct mcm_ib_dev *md, char *name, int
 			}
 			else {
 				char val[64];
+				struct ibv_device_attr device_attr;
 
+				if (ibv_query_device(ibctx, &device_attr)) {
+					mlog(0,"ERR ibv_device, %s\n", strerror(errno));
+					ibv_close_device(ibctx);
+					ibctx = NULL;
+					goto bail;
+				}
+
+				md->dev_attr.rd_atom_in = device_attr.max_qp_rd_atom;
+				md->dev_attr.rd_atom_out = device_attr.max_qp_init_rd_atom;
 				md->ibdev = iblist[i];
 				if (!rd_dev_file(md->ibdev->ibdev_path,
 				    "device/numa_node", val, sizeof val))
 					md->numa_node = atoi(val);
+				else if (!strncmp(name, "scif", 4))
+					md->numa_node = md->mc->numa_node; /* intra-node, MSS */
 				else
 					mlog(0," ERR ibdev %s numa_node at "
 					     "%s/device/numa_node unreadable\n",
@@ -444,6 +456,7 @@ void mpxy_destroy_smd(mcm_scif_dev_t *smd)
 	mpxy_lock_destroy(&smd->cqlock);
 	mpxy_lock_destroy(&smd->cqrlock);
 	mpxy_lock_destroy(&smd->mrlock);
+	mpxy_lock_destroy(&smd->pzlock);
 	mpxy_lock_destroy(&smd->evlock);
 	mpxy_lock_destroy(&smd->tblock);
 	mpxy_lock_destroy(&smd->rblock);
@@ -616,6 +629,7 @@ static mcm_scif_dev_t *mcm_create_smd(mcm_ib_dev_t *md, scif_epd_t op_ep, scif_e
 	mpxy_lock_init(&smd->cqlock, NULL);  /* cq list */
 	mpxy_lock_init(&smd->cqrlock, NULL);  /* cq rx list */
 	mpxy_lock_init(&smd->mrlock, NULL);  /* mr list */
+	mpxy_lock_init(&smd->pzlock, NULL);  /* pz list */
 	mpxy_lock_init(&smd->evlock, NULL);  /* DTO event, multi-threads */
 	mpxy_lock_init(&smd->tblock, NULL);  /* tx proxy buffer, shared across all QP's */
 	mpxy_lock_init(&smd->rblock, NULL);  /* rx proxy buffer, shared across all QP's */
@@ -628,6 +642,7 @@ static mcm_scif_dev_t *mcm_create_smd(mcm_ib_dev_t *md, scif_epd_t op_ep, scif_e
 	init_list(&smd->cqlist);
 	init_list(&smd->cqrlist);
 	init_list(&smd->mrlist);
+	init_list(&smd->pzlist);
 
 	return smd;
 err:
@@ -657,8 +672,8 @@ mcm_scif_dev_t *mix_open_device(dat_mix_open_t *msg, scif_epd_t op_ep, scif_epd_
 	mcm_scif_dev_t *smd = NULL;
 	int i, ret;
 
-	mlog(8, " Open IB device - %s, IB port %d, scif_node %d EPs %d %d %d op_msg %p\n",
-		msg->name, msg->port, node, op_ep, tx_ep, ev_ep, msg);
+	mlog(8, " IB device - %s, IB port %d, scif_node %d EPs %d %d %d op_msg %p lid %x\n",
+		msg->name, msg->port, node, op_ep, tx_ep, ev_ep, msg, ntohs(msg->dev_addr.lid));
 
 	mc = &mcm_client_list[node];
 
@@ -748,7 +763,9 @@ mcm_scif_dev_t *mix_open_device(dat_mix_open_t *msg, scif_epd_t op_ep, scif_epd_
 	memset(md->cntrs, 0, sizeof(uint64_t) * MCM_ALL_COUNTERS);
 	md->mc = mc;
 	md->port = msg->port;
+	memcpy(&md->addr, &msg->dev_addr, sizeof(dat_mcm_addr_t));
 	md->ibctx = open_ib_device(md, msg->name, msg->port);
+	md->addr.ep_map = msg->dev_addr.ep_map;
 
 	if ((!md->ibctx) || mcm_init_cm_service(md)) {
 		mcm_destroy_md(md);
@@ -769,15 +786,31 @@ found:
 
 	msg->hdr.req_id = smd->entry.tid;
 	msg->hdr.status = MIX_SUCCESS;
+	msg->dev_attr.rd_atom_in = md->dev_attr.rd_atom_in;
+	msg->dev_attr.rd_atom_out = md->dev_attr.rd_atom_out;
+	if (!(mcm_ib_inline_data(md->ibctx))  || !mcm_ib_inline)
+		msg->dev_attr.max_inline = 0;
+
 	memcpy(&md->dev_attr, &msg->dev_attr, sizeof(dat_mix_dev_attr_t));
-	memcpy(&msg->dev_addr, &md->addr, sizeof(dat_mcm_addr_t));
+	memcpy(&msg->dev_addr, &md->addr, sizeof(dat_mcm_addr_t)); /* proxy CM lid */
+
+	/* intra-node: restore MIC lid, gid */
+	if (md->m_lid) {
+		msg->dev_addr.lid = md->m_lid;
+		memcpy(msg->dev_addr.gid, md->m_gid, 16);
+	}
+
+	/* MTU changed via DAPL_IB_MTU */
+	if (msg->hdr.flags & MIX_OP_MTU)
+		md->mtu_env = md->dev_attr.mtu;
 err:
 	if (!smd) {
-		mlog(0, " ERR: mix_open_device failed for %s - %d\n", msg->name, msg->port);
+		mlog(1, " WARN: open failed for %s - %d\n", msg->name, msg->port);
 		msg->hdr.status = MIX_ENODEV;
 	}
 
 	/* send back response */
+	msg->hdr.flags = MIX_OP_RSP;
 	ret = scif_send_msg(op_ep, (void*)msg, sizeof(dat_mix_open_t));
 	if (ret) {
 		mlog(0, " ERR: scif_send dev_id %d op_ep %d, closing device %p\n",
@@ -789,9 +822,10 @@ err:
 		goto bail;
 	}
 
-	mlog(1, " MIC client: open mdev[%d] %p smd %p mic%d[%d] -> %s[%d] port %d - %s\n",
+	mlog(1, " MIC client: mdev[%d] %p->%p mic%d[%d] -> %s[%d] port %d lid %x %s mtu %d (%d)\n",
 		md->smd_list.tid, md, smd, mc->scif_id-1, mc->numa_node, msg->name,
-		md->numa_node, msg->port, md->addr.ep_map == MIC_SSOCK_DEV ? "MSS":"MXS");
+		md->numa_node, msg->port, ntohs(msg->dev_addr.lid), mcm_map_str(md->addr.ep_map),
+		md->dev_attr.mtu, md->mtu_env);
 bail:
 	mpxy_unlock(&mc->oplock);
 	mpxy_unlock(&mc->cmlock);
@@ -1159,6 +1193,7 @@ void mpxy_rx_thread(void *mic_client)
 				m_qp = get_head_entry(&smd->qprlist);
 				while (m_qp) {
 					m_pi_pending_wr(m_qp, &data); /* RR's and scif_sendto */
+					m_pi_pending_wt(m_qp); /* WT's pending */
 					m_qp = get_next_entry(&m_qp->r_entry, &smd->qprlist);
 				}
 				mpxy_unlock(&smd->qprlock);
@@ -1218,7 +1253,7 @@ static void mpxy_server(void)
 		set.revents = 0;
 		mlog(0x8, "Server sleep\n");
 		poll(&set, 1, -1); /* sleep */
-		mlog(0x8, "Server wake, cpu_id=%d\n");
+		mlog(0x8, "Server wake, cpu_id=%d\n", cpu_id);
 		/* process listens */
 		if (mcm_poll(scif_listen_ep, POLLIN) == POLLIN)
 			mix_scif_accept(scif_listen_ep);
@@ -1317,8 +1352,8 @@ int main(int argc, char **argv)
 	logfile = mpxy_open_log();
 	mpxy_log_options();
 
-	mlog(0, "CCL Proxy - SCIF/IB DAPL RDMA Proxy Service, Mix Version %d (Build-%u) v2\n",
-		DAT_MIX_VER, PACKAGE_DATE);
+	mlog(0, "CCL Proxy - SCIF/IB DAPL RDMA Proxy Service %s (%u)\n",
+		 PACKAGE_VERSION, PACKAGE_DATE);
 
 	if (init_scif()) {
 		mlog(0, "ERROR - unable to open/init SCIF device\n");
@@ -1417,8 +1452,7 @@ void mcm_dat_dev_log(struct mcm_scif_dev *smd)
 
 	/* show PO mbuf_wc busy slots */
 	idx = smd->m_buf_tl;
-	while ((smd->m_buf_tl != smd->m_buf_hd) &&
-	       (smd->m_buf_hd - smd->m_buf_tl)) {
+	while (idx != smd->m_buf_hd) {
 		if ((smd->m_buf_wc[idx].m_idx && !smd->m_buf_wc[idx].done) || 1) {
 			struct mcm_wr *m_wr = NULL;
 			struct mcm_qp *m_qp = NULL;
@@ -1440,8 +1474,6 @@ void mcm_dat_dev_log(struct mcm_scif_dev *smd)
 				smd->m_buf_wc[idx].tl, smd->m_buf_wc[idx].hd);
 		}
 		idx = (idx + 1) & smd->m_buf_end;
-		if (idx == (smd->m_buf_hd+2))
-			break;
 	}
 
 	/* show PI mbuf_wc busy slots, start from tail */
@@ -1481,7 +1513,7 @@ void mcm_ib_dev_log(struct mcm_ib_dev *md)
 		md->mc->scif_id, md, ntohs(md->addr.lid), md->port,
 		inet_ntop(AF_INET6, md->addr.gid, gid_str, sizeof(gid_str)),
 		md->qp->qp_num, md->mc->scif_id - 1, md->ibdev->name,
-		md->addr.ep_map == MIC_SSOCK_DEV ? "MSS":"MXS", md->mc->ver);
+		mcm_map_str(md->addr.ep_map), md->mc->ver);
 }
 
 static int check_io_run;
