@@ -58,7 +58,7 @@
 #define min(a, b) ((a < b) ? (a) : (b))
 #define max(a, b) ((a > b) ? (a) : (b))
 
-#define MCM_IB_INLINE		160
+#define MCM_IB_INLINE		(sizeof(mcm_wr_rx_t))
 #define MIX_MAX_MSG_SIZE	(8*1024*1024)
 
 #define MIX_MIN  4 		/* oldest version supported */
@@ -117,6 +117,8 @@ typedef struct mcm_ib_dev {
 	uint64_t		*ports;	/* SCIF device open clients, cm_id*/
 	struct dat_mcm_addr	addr;
 	uint16_t		lid;
+	uint16_t		m_lid; /* intra-node, ibscif HST lid != MICx lids */
+	uint8_t			m_gid[16];
 	struct dat_mix_dev_attr	dev_attr; /* provided with mix_open */
 	int			s_hd;
 	int			s_tl;
@@ -130,6 +132,7 @@ typedef struct mcm_ib_dev {
 	int 			numa_node;
 	int			indata;
 	void			*cntrs;
+	uint8_t			mtu_env;
 
 } mcm_ib_dev_t;
 
@@ -209,6 +212,7 @@ typedef struct mcm_qp {
 	int			comp_cnt;
 	char			*wr_buf_rx;	/* mcm_wr_rx_t entries, for devices without inline data  */
 	struct ibv_mr		*wr_buf_rx_mr;
+	int			p2p_data; 	/* Max number of bytes to pass from proxy to proxy in the WR */
 	/* Proxy-in: WR management, remote view from TX side */
 	mcm_wrc_info_t		wrc_rem;	/* WR and WC buffers: remote, in CM req and reply */
 	int			wr_pp_rem;	/* work request pending */
@@ -241,6 +245,7 @@ typedef struct mcm_qp {
 	int 			sr_len;		/* SR WR buffer pool len */
 	int			sr_sz;		/* SR WR entry size */
 	int			post_sr;
+	uint8_t			mtu;		/* negotiated QP MTU */
 #ifdef MCM_PROFILE
 	mcm_qp_prof_t		ts;
 	uint32_t		last_wr_sig;
@@ -265,6 +270,14 @@ typedef struct mcm_cq {
 
 } mcm_cq_t;
 
+/*  DAPL MCM PZ object */
+typedef struct mcm_pz {
+	LLIST_ENTRY		entry;
+	struct mcm_scif_dev	*smd;
+	struct ibv_pd		*ib_pd;
+	uint64_t		pd_ctx; /* MIC client */
+} mcm_pz_t;
+
 /*  DAPL MCM MR object, id in entry */
 typedef struct mcm_mr {
 	LLIST_ENTRY		entry;
@@ -308,6 +321,7 @@ typedef struct mcm_scif_dev {
 	LLIST_ENTRY		cqlist;		/* client cq create list */
 	LLIST_ENTRY		cqrlist;	/* mpxyd cq list for proxy in service */
 	LLIST_ENTRY		mrlist;		/* mr list */
+	LLIST_ENTRY		pzlist;		/* pz list */
 	mpxy_lock_t		clock;		/* LOCKS: cm lock */
 	mpxy_lock_t		llock;		/* listen lock */
 	mpxy_lock_t		plock;		/* port space lock */
@@ -316,6 +330,7 @@ typedef struct mcm_scif_dev {
 	mpxy_lock_t		cqlock;		/* cq tx lock */
 	mpxy_lock_t		cqrlock;	/* cq rx lock */
 	mpxy_lock_t		mrlock;		/* mr lock */
+	mpxy_lock_t		pzlock;		/* pz lock */
 	mpxy_lock_t		evlock;		/* DTO event lock, multi-threads on ev_ep */
 	int			destroy;	/* destroying device, all resources */
 	int			ref_cnt;	/* child references */
@@ -600,6 +615,7 @@ int m_pi_create_sr_q(struct mcm_qp *m_qp, int entries);
 int m_pi_create_bpool(struct mcm_qp *m_qp, int max_recv_wr);
 void m_qp_destroy_pi(struct mcm_qp *m_qp);
 int m_qp_create_pi(mcm_scif_dev_t *smd, struct mcm_qp *m_qp);
+void m_pi_pending_wt(struct mcm_qp *m_qp);
 void m_pi_pending_wr(struct mcm_qp *m_qp, int *data);
 void m_pi_pending_wc(struct mcm_qp *m_qp, int *events);
 void m_pi_req_event(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, struct ibv_wc *wc, int type);
@@ -807,7 +823,7 @@ static inline void const_ib_rr(struct ibv_send_wr *iwr, struct dat_mix_wr *mwr,
 static inline void mcm_pr_addrs(int lvl, struct dat_mcm_msg *msg, int state, int in)
 {
 	if (in) {
-		if (MXS_EP(&msg->daddr1) && MXS_EP(&msg->saddr1)) {
+		if (MXF_EP(&msg->daddr1) && MXF_EP(&msg->saddr1)) {
 			mlog(lvl, " QPr_t addr2: %s 0x%x %x 0x%x %s <- QPt_r addr2: 0x%x %x 0x%x %s\n",
 				mcm_state_str(state), htons(msg->daddr2.lid),
 				htonl(msg->daddr2.qpn), htons(msg->dport),
@@ -829,7 +845,7 @@ static inline void mcm_pr_addrs(int lvl, struct dat_mcm_msg *msg, int state, int
 				htons(msg->sport), mcm_map_str(msg->saddr1.ep_map));
 		}
 	} else {
-		if (MXS_EP(&msg->saddr1) && MXS_EP(&msg->daddr1)) {
+		if (MXF_EP(&msg->saddr1) && MXF_EP(&msg->daddr1)) {
 			mlog(lvl, " QPr_t addr2: %s 0x%x %x 0x%x %s -> QPt_r addr2: 0x%x %x 0x%x %s\n",
 				mcm_state_str(state), htons(msg->saddr2.lid),
 				htonl(msg->saddr2.qpn), htons(msg->sport),
