@@ -99,13 +99,16 @@
 #define MAX_RDMA_RD    4
 #define MAX_PROCS      1000
 
+#define min(a, b) ((a < b) ? (a) : (b))
+#define max(a, b) ((a > b) ? (a) : (b))
+
 /* Header files needed for DAT/uDAPL */
 #include "dat2/udat.h"
 #include "dat2/dat_ib_extensions.h"
 
 /* definitions */
 #define SERVER_CONN_QUAL  45248
-#define DTO_TIMEOUT       (1000*1000*5)
+#define DTO_TIMEOUT       DAT_TIMEOUT_INFINITE
 #define CNO_TIMEOUT       (1000*1000*1)
 #define DTO_FLUSH_TIMEOUT (1000*1000*2)
 #define CONN_TIMEOUT      (1000*1000*100)
@@ -194,15 +197,19 @@ struct dt_time {
 struct dt_time ts;
 
 /* defaults */
+static int all_data_sizes = 0;
+static int increment = 0;
 static int failed = 0;
 static int uni_direction = 0;
 static int align_data=1;
 static int rdma_read = 0;
 static int write_only = 0;
+static int write_only_pp = 0;
 static int write_immed = 0;
 static int performance_times = 0;
 static int connected = 0;
-static int burst = 100;
+static int burst = 1000;
+static int msg_burst = 100;
 static int signal_rate = 10;
 static int server = 1;
 static int verbose = 0;
@@ -213,6 +220,7 @@ static int conn_poll_count = 0;
 static int rdma_rd_poll_count[MAX_RDMA_RD] = { 0 };
 static int delay = 0;
 static int buf_len = RDMA_BUFFER_SIZE;
+static int buf_len_p2;
 static int use_cno = 0;
 static int recv_msg_index = 0;
 static int burst_msg_posted = 0;
@@ -220,6 +228,7 @@ static int burst_msg_index = 0;
 static int ucm = 0;
 static int rq_cnt, sq_cnt;
 static DAT_SOCK_ADDR6 remote;
+static int data_check = 0;
 
 /* forward prototypes */
 const char *DT_RetToStr(DAT_RETURN ret_value);
@@ -243,6 +252,7 @@ DAT_RETURN create_events(void);
 DAT_RETURN destroy_events(void);
 DAT_RETURN do_rdma_write_imm_with_msg(void);
 DAT_RETURN do_rdma_write_with_msg(void);
+DAT_RETURN do_rdma_write_ping_pong(int p2, int bytes);
 DAT_RETURN do_rdma_read_with_msg(void);
 DAT_RETURN do_ping_pong_msg(void);
 
@@ -257,25 +267,25 @@ void flush_evds(void)
 	DAT_EVENT event;
 
 	/* Flush async error queue */
-	printf("%d ERR: Checking ASYNC EVD...\n", getpid());
+	LOGPRINTF("%d: Checking ASYNC EVD...\n", getpid());
 	while (dat_evd_dequeue(h_async_evd, &event) == DAT_SUCCESS) {
-		printf(" ASYNC EVD ENTRY: handle=%p reason=%d\n",
+		LOGPRINTF("%d ERR: ASYNC EVD ENTRY: handle=%p reason=%d\n", getpid(),
 			event.event_data.asynch_error_event_data.dat_handle,
 			event.event_data.asynch_error_event_data.reason);
 	}
 	/* Flush receive queue */
-	printf("%d ERR: Checking RECEIVE EVD...\n", getpid());
+	LOGPRINTF("%d: Checking RECEIVE EVD...\n", getpid());
 	while (dat_evd_dequeue(h_dto_rcv_evd, &event) == DAT_SUCCESS) {
-		printf(" RCV EVD ENTRY: op=%d stat=%d ln=%d ck="F64x"\n",
+		LOGPRINTF(" RCV EVD ENTRY: op=%d stat=%d ln=%d ck="F64x"\n",
 			event.event_data.dto_completion_event_data.operation,
 			event.event_data.dto_completion_event_data.status,
 			event.event_data.dto_completion_event_data.transfered_length,
 			event.event_data.dto_completion_event_data.user_cookie.as_64);
 	}
 	/* Flush request queue */
-	printf("%d ERR: Checking REQUEST EVD...\n", getpid());
+	LOGPRINTF("%d: Checking REQUEST EVD...\n", getpid());
 	while (dat_evd_dequeue(h_dto_req_evd, &event) == DAT_SUCCESS) {
-		printf(" REQ EVD ENTRY: op=%d stat=%d ln=%d ck="F64x"\n",
+		LOGPRINTF(" REQ EVD ENTRY: op=%d stat=%d ln=%d ck="F64x"\n",
 			event.event_data.dto_completion_event_data.operation,
 			event.event_data.dto_completion_event_data.status,
 			event.event_data.dto_completion_event_data.transfered_length,
@@ -557,10 +567,12 @@ int main(int argc, char **argv)
 	DAT_PROVIDER_ATTR pr_attr;
 
 	/* parse arguments */
-	while ((c = getopt(argc, argv, "auwtscvpb:d:B:h:P:S:")) != -1) {
+	while ((c = getopt(argc, argv, "UDauwWtscvpb:d:B:h:P:S:i:")) != -1) {
 		switch (c) {
+		case 'i':
+			increment = atoi(optarg);
 		case 'a':
-			align_data = 1;
+			all_data_sizes = 1;
 			fflush(stdout);
 			break;
 		case 'u':
@@ -571,6 +583,17 @@ int main(int argc, char **argv)
 			write_only = 1;
 			fflush(stdout);
 			break;
+		case 'D':
+			data_check = 1;
+			printf("%d Running DATA CHECK mode\n", getpid());
+			/* fall through */
+		case 'W':
+			write_only_pp = 1;
+			uni_direction = 1;
+			signal_rate = 1;
+			burst = 1000;
+			fflush(stdout);
+			break;
 		case 't':
 			performance_times = 1;
 			fflush(stdout);
@@ -613,12 +636,24 @@ int main(int argc, char **argv)
 		case 'S':
 			signal_rate = atoi(optarg);
 			break;
+		case 'U':
+			/* fall through */
 		default:
 			print_usage();
 			exit(-12);
 		}
 	}
 
+	if (all_data_sizes && !write_only_pp) {
+		printf("\n\t -a option only valid with -W option\n\n");
+		exit(-12);
+	}
+
+	if (data_check && strstr(provider, "scif")) {
+		printf("\n\t -D option is not valid with scif provider\n\n");
+		exit(-12);
+	}
+
 #if defined(_WIN32) || defined(_WIN64)
 	{
 		WSADATA wsaData;
@@ -637,8 +672,13 @@ int main(int argc, char **argv)
 	if (signal_rate > burst)
 		signal_rate = burst;
 
-	rq_cnt = MSG_BUF_COUNT + (burst);
-	sq_cnt = MSG_BUF_COUNT + MAX_RDMA_RD + signal_rate;
+	if (write_only || write_only_pp) {
+		rq_cnt = MSG_BUF_COUNT * 2;
+		sq_cnt = MSG_BUF_COUNT + MAX_RDMA_RD + signal_rate;
+	} else {
+		rq_cnt = MSG_BUF_COUNT + msg_burst;
+		sq_cnt = MSG_BUF_COUNT + MAX_RDMA_RD + msg_burst;
+	}
 
 	if (!server) {
 		printf("%d Running as client - waiting for server input\n",
@@ -655,23 +695,40 @@ int main(int argc, char **argv)
 	}
 	fflush(stdout);
 
+	if (write_only_pp) {
+		/* rdma write pingpong, default == 1 byte */
+		if (!all_data_sizes) {
+			if (!data_check)
+				buf_len = 1;
+		} else if (!increment) { /* power of 2 */
+			buf_len_p2 = 1;
+			i = 0;
+			while (buf_len_p2 < buf_len) {
+				buf_len_p2 <<= 1;
+				i++;
+			}
+			buf_len_p2 = i;
+		}
+	}
+
 	if (align_data) {
 		/* allocate send and receive buffers */
-		if (posix_memalign((void**)&rbuf, 4096, buf_len * (burst+1)) ||
-		    posix_memalign((void**)&sbuf, 4096, buf_len * (burst+1))) {
+		if (posix_memalign((void**)&rbuf, 4096, max(4096, buf_len * rq_cnt)) ||
+		    posix_memalign((void**)&sbuf, 4096, max(4096, buf_len * rq_cnt))) {
 			perror("malloc");
 			exit(1);
 		}
 	} else {
 		/* allocate send and receive buffers */
-		if (((rbuf = malloc(buf_len * (burst+1))) == NULL) ||
-		    ((sbuf = malloc(buf_len * (burst+1))) == NULL)) {
+		if (((rbuf = malloc(max(64, buf_len * rq_cnt))) == NULL) ||
+		    ((sbuf = malloc(max(64, buf_len * rq_cnt))) == NULL)) {
 			perror("malloc");
 			exit(1);
 		}
 	}
-	LOGPRINTF("%d Allocated RDMA buffers (r:%p,s:%p) len %d \n",
-		  getpid(), rbuf, sbuf, buf_len);
+	init_data();
+	LOGPRINTF("%d Allocated RDMA buffers (r:%p=%d,s:%p=%d) len %d \n",
+		  getpid(), rbuf, *rbuf, sbuf, *sbuf, buf_len);
 
 	if (posix_memalign((void**)&p_rmr_rcv, 4096, 4096) ||
 	    posix_memalign((void**)&p_rmr_snd, 4096, 4096)) {
@@ -687,6 +744,7 @@ int main(int argc, char **argv)
 	ret = dat_ia_open(provider, 8, &h_async_evd, &h_ia);
 	stop = get_time();
 	ts.open += ((stop - start) * 1.0e6);
+	ts.total += ts.open;
 	if (ret != DAT_SUCCESS) {
 		fprintf(stderr, "%d: Error Adaptor open: %s\n",
 			getpid(), DT_RetToStr(ret));
@@ -727,6 +785,7 @@ int main(int argc, char **argv)
 	ret = dat_pz_create(h_ia, &h_pz);
 	stop = get_time();
 	ts.pzc += ((stop - start) * 1.0e6);
+	ts.total += ts.pzc;
 	if (ret != DAT_SUCCESS) {
 		fprintf(stderr, "%d Error creating Protection Zone: %s\n",
 			getpid(), DT_RetToStr(ret));
@@ -760,8 +819,8 @@ int main(int argc, char **argv)
 	ep_attr.max_rdma_size = 0x10000;
 	ep_attr.qos = 0;
 	ep_attr.recv_completion_flags = 0;
-	ep_attr.max_recv_dtos = MSG_BUF_COUNT + (burst * 3);
-	ep_attr.max_request_dtos = MSG_BUF_COUNT + (burst * 3) + MAX_RDMA_RD;
+	ep_attr.max_recv_dtos = rq_cnt;
+	ep_attr.max_request_dtos = sq_cnt;
 	ep_attr.max_recv_iov = MSG_IOV_COUNT;
 	ep_attr.max_request_iov = MSG_IOV_COUNT;
 	ep_attr.max_rdma_read_in = MAX_RDMA_RD;
@@ -841,20 +900,76 @@ int main(int argc, char **argv)
 #endif
 
 	/*********** RDMA write data *************/
-	if ((write_immed) && (write_only))
+	if (write_only_pp) {
+		int max, inc;
+
+		if (all_data_sizes) {
+			if (increment) {
+				i = 1;
+				inc = increment;
+				max = buf_len/inc;
+			} else {
+				i = 0;
+				inc = 0;
+				max = buf_len_p2;
+			}
+		} else {
+			if (data_check) {
+				i = buf_len;
+				max = buf_len;
+				inc = 1;
+			}
+			else
+			{
+				i = buf_len;
+				max = buf_len;
+				inc = buf_len;
+			}
+		}
+		printf("\n %d RDMA WRITE PINGPONG %s\n\n", getpid(),
+			data_check ? "with DATA CHECK":"");
+
+		for (; i <= max; i++) {
+			if (all_data_sizes) {
+				int l_len = (i*inc) ? (i*inc) : 1 << i;
+
+				if ( l_len > 4 && do_rdma_write_ping_pong(i, l_len - 1)) {
+					fprintf(stderr, "%d Error do_rdma_write_ping_pong\n", getpid());
+					goto cleanup;
+				}
+			}
+
+			if (do_rdma_write_ping_pong(i, i*inc)) {
+				fprintf(stderr, "%d Error do_rdma_write_ping_pong\n", getpid());
+				goto cleanup;
+			}
+
+			if (all_data_sizes) {
+				int l_len = (i*inc) ? (i*inc) : 1 << i;
+
+				if ( l_len > 1 && l_len < buf_len && do_rdma_write_ping_pong(i, l_len + 1)) {
+					fprintf(stderr, "%d Error do_rdma_write_ping_pong\n", getpid());
+					goto cleanup;
+				}
+			}
+		}
+	}
+	else if (write_immed && write_only) {
 		ret = do_rdma_write_imm_with_msg();
-	else
+	}
+	else {
 		ret = do_rdma_write_with_msg();
+	}
 
 	if (ret != DAT_SUCCESS) {
 		fprintf(stderr, "%d Error do_rdma_write_%swith_msg: %s\n",
-			getpid(), write_immed ? "imm_":"", DT_RetToStr(ret));
+			getpid(), write_immed && write_only ? "imm_":"",
+			DT_RetToStr(ret));
 		goto cleanup;
 	} else
-		LOGPRINTF("%d do_rdma_write_%swith_msg complete\n",
-			  getpid(), write_immed ? "imm_":"");
+		LOGPRINTF("%d rdma_write test complete\n", getpid());
 
-	if (write_only || !rdma_read)
+	if (write_only_pp || write_only || !rdma_read)
 		goto complete;
 
 	/*********** RDMA read data *************/
@@ -878,7 +993,6 @@ int main(int argc, char **argv)
 	}
 
 cleanup:
-	flush_evds();
 	failed++;
 complete:
 
@@ -903,6 +1017,8 @@ complete:
 			h_ep = DAT_HANDLE_NULL;
 		}
 	}
+	if (connected)
+		flush_evds();
 
 	/* free EVDs */
 	LOGPRINTF("%d destroy events\n", getpid());
@@ -927,6 +1043,7 @@ complete:
 	ret = dat_pz_free(h_pz);
 	stop = get_time();
 	ts.pzf += ((stop - start) * 1.0e6);
+	ts.total += ts.pzf;
 	if (ret != DAT_SUCCESS) {
 		fprintf(stderr, "%d Error freeing PZ: %s\n",
 			getpid(), DT_RetToStr(ret));
@@ -941,6 +1058,7 @@ complete:
 	ret = dat_ia_close(h_ia, DAT_CLOSE_ABRUPT_FLAG);
 	stop = get_time();
 	ts.close += ((stop - start) * 1.0e6);
+	ts.total += ts.close;
 	if (ret != DAT_SUCCESS) {
 		fprintf(stderr, "%d: Error Adaptor close: %s\n",
 			getpid(), DT_RetToStr(ret));
@@ -951,9 +1069,16 @@ complete:
 	free(rbuf);
 	free(sbuf);
 
-	if (ts.rtt)
-		printf("%d: Message RTT: Total=%6.2lf usec, %d bursts, itime=%6.2lf usec, pc=%d\n",
-			getpid(), ts.rtt, burst, ts.rtt / burst, poll_count);
+	if (ts.rtt && !all_data_sizes) {
+		printf( "%d: %s PingPong: (%d x %d) Total %6.2lf us:"
+			" latency %3.2lf us, BW %4.2lf MB/s\n",
+			getpid(), write_only_pp ? "RDMA write":"Message",
+			write_only_pp ? burst : msg_burst, buf_len, ts.rtt,
+			write_only_pp ? ts.rtt/burst/2:ts.rtt/msg_burst/2,
+			write_only_pp ? (double)(1/(ts.rtt/burst/2/buf_len)):
+					(double)(1/(ts.rtt/msg_burst/2/buf_len)));
+	}
+
 	if (ts.rdma_wr && (!server || (server && !uni_direction))) {
 		int msgs = uni_direction ? burst : burst * 2;
 
@@ -1193,6 +1318,7 @@ DAT_RETURN connect_ep(char *hostname,
 		printf("%d Server is waiting for client connection to send"
 			" server info\n",
 			getpid());
+		fflush(stdout);
 		if (send_server_params(ser_sa)) {
 			printf("%d Failed to send server params\n", getpid());
 			return -1;
@@ -1408,6 +1534,7 @@ no_resolution:
 
 	printf("\n%d CONNECTED!\n\n", getpid());
 	connected = 1;
+	fflush(stdout);
 
 #if CONNECT_ONLY
 	return 0;
@@ -1416,16 +1543,16 @@ no_resolution:
 	 *  Setup our remote memory and tell the other side about it
 	 */
 	p_rmr_snd->virtual_address = htonll((DAT_VADDR) (uintptr_t) rbuf);
-	p_rmr_snd->segment_length = htonl(RDMA_BUFFER_SIZE);
+	p_rmr_snd->segment_length = htonl(buf_len);
 	p_rmr_snd->rmr_context = htonl(rmr_context_recv);
 
 	printf("%d Send RMR msg to remote: r_key_ctx=0x%x,va=%p,len=0x%x\n",
-	       getpid(), rmr_context_recv, rbuf, RDMA_BUFFER_SIZE);
+	       getpid(), rmr_context_recv, rbuf, buf_len);
 
 	ret = send_msg(p_rmr_snd,
 		       sizeof(DAT_RMR_TRIPLET),
 		       lmr_context_send_msg,
-		       cookie, DAT_COMPLETION_SUPPRESS_FLAG);
+		       cookie, DAT_COMPLETION_DEFAULT_FLAG);
 
 	if (ret != DAT_SUCCESS) {
 		fprintf(stderr, "%d Error send_msg: %s\n",
@@ -1571,15 +1698,17 @@ void disconnect_ep(void)
 DAT_RETURN do_rdma_write_with_msg(void)
 {
 	DAT_EVENT event;
+	DAT_DTO_COMPLETION_EVENT_DATA *dto_event;
 	DAT_LMR_TRIPLET l_iov[MSG_IOV_COUNT];
 	DAT_RMR_TRIPLET r_iov;
 	DAT_DTO_COOKIE cookie;
 	DAT_RETURN ret;
+	DAT_COMPLETION_FLAGS flags;
 	int i;
 
 	printf("\n %d RDMA WRITE DATA with SEND MSG\n\n", getpid());
 
-	cookie.as_64 = 0x5555;
+	dto_event = &event.event_data.dto_completion_event_data;
 
 	if (recv_msg_index >= MSG_BUF_COUNT)
 		return (DAT_ABORT);
@@ -1592,6 +1721,9 @@ DAT_RETURN do_rdma_write_with_msg(void)
 	else
 		strcpy((char *)sbuf, "client RDMA write data...");
 
+	if  (uni_direction && server)
+		goto rmsg;
+
 	for (i = 0; i < MSG_IOV_COUNT; i++) {
 		l_iov[i].lmr_context = lmr_context_send;
 		l_iov[i].segment_length = buf_len / MSG_IOV_COUNT;
@@ -1605,26 +1737,50 @@ DAT_RETURN do_rdma_write_with_msg(void)
 
 	start = get_time();
 	for (i = 0; i < burst; i++) {
-		cookie.as_64 = 0x9999;
+		if (!((i+1) % signal_rate))
+			flags =  DAT_COMPLETION_DEFAULT_FLAG;
+		else
+			flags = DAT_COMPLETION_SUPPRESS_FLAG;
+
+		cookie.as_64 = i;
+		LOGPRINTF("%d rdma_write # %d %s\n",
+			  getpid(), i + 1, flags ? "SUPPRESS":"SIGNAL");
 		ret = dat_ep_post_rdma_write(h_ep,	// ep_handle
 					     MSG_IOV_COUNT,	// num_segments
 					     l_iov,	// LMR
 					     cookie,	// user_cookie
 					     &r_iov,	// RMR
-					     DAT_COMPLETION_SUPPRESS_FLAG);
+					     flags);
 		if (ret != DAT_SUCCESS) {
 			fprintf(stderr,
 				"%d: ERROR: dat_ep_post_rdma_write() %s\n",
 				getpid(), DT_RetToStr(ret));
 			return (DAT_ABORT);
 		}
+		if (flags == DAT_COMPLETION_DEFAULT_FLAG) {
+			if (collect_event(h_dto_req_evd,
+					  &event,
+					  DTO_TIMEOUT,
+					  &rdma_wr_poll_count) != DAT_SUCCESS) {
+				printf("%d %s RDMA write buffer contains: %s\n",
+				       getpid(), server ? "SERVER:" : "CLIENT:", rbuf);
+				return (DAT_ABORT);
+			}
+			if (dto_event->status ||
+			    dto_event->user_cookie.as_64 != i) {
+				fprintf(stderr,	"ERROR rdma_write: cookie="
+						" "F64x " exp 0x%x st 0x%x\n",
+					dto_event->user_cookie.as_64, i,
+					dto_event->status);
+				return (DAT_ABORT);
+			}
+		}
 		LOGPRINTF("%d rdma_write # %d completed\n", getpid(), i + 1);
 	}
 
-	/*
-	 *  Send RMR information a 2nd time to indicate completion
-	 *  NOTE: already swapped to network order in connect_ep
-	 */
+	if (server)
+		goto rmsg;
+smsg:
 	printf("%d Sending RDMA WRITE completion message\n", getpid());
 
 	ret = send_msg(p_rmr_snd,
@@ -1639,14 +1795,18 @@ DAT_RETURN do_rdma_write_with_msg(void)
 	} else {
 		LOGPRINTF("%d send_msg completed\n", getpid());
 	}
-
+	if (server)
+		goto acked;
+rmsg:
 	/* inbound recv event, send completion's suppressed */
 	if (collect_event(h_dto_rcv_evd,
 			  &event,
 			  DTO_TIMEOUT,
-			  &rdma_wr_poll_count) != DAT_SUCCESS)
+			  &rdma_wr_poll_count) != DAT_SUCCESS) {
+		printf("%d %s RDMA write buffer contains: %s\n",
+			getpid(), server ? "SERVER:" : "CLIENT:", rbuf);
 		return (DAT_ABORT);
-
+	}
 	stop = get_time();
 	ts.rdma_wr = ((stop - start) * 1.0e6);
 
@@ -1663,15 +1823,18 @@ DAT_RETURN do_rdma_write_with_msg(void)
 	    || (event.event_data.dto_completion_event_data.user_cookie.as_64 !=
 		recv_msg_index)) {
 		fprintf(stderr,
-			"unexpected event data for receive: len=%d cookie=" F64x
-			" exp %d/%d\n",
-			(int)event.event_data.dto_completion_event_data.
-			transfered_length,
+			"unexpected event data for receive: st=%d len=%d"
+			"cookie=" F64x " exp %d/%d\n",
+			event.event_data.dto_completion_event_data.status,
+			(int)event.event_data.dto_completion_event_data.transfered_length,
 			event.event_data.dto_completion_event_data.user_cookie.
 			as_64, (int)sizeof(DAT_RMR_TRIPLET), recv_msg_index);
 
 		return (DAT_ABORT);
 	}
+	if (server)
+		goto smsg;
+acked:
 
 	/* swap received RMR msg: network order to host order */
 	r_iov = p_rmr_rcv[recv_msg_index];
@@ -1787,73 +1950,80 @@ DAT_RETURN do_rdma_write_imm_with_msg(void)
 		}
 	}
 
+	if (uni_direction && !server)
+		goto smsg;
 done:
-	if ((!uni_direction) || (uni_direction && server)) {
-		/* Wait to RECEIVE the LAST message, immediate data expected */
-		LOGPRINTF("%d Waiting for final inbound RW_imm from peer\n", getpid());
-		if (collect_event(h_dto_rcv_evd,
-				  &event,
-				  DTO_TIMEOUT,
-				  &rdma_wr_poll_count) != DAT_SUCCESS)
-			return (DAT_ABORT);
+	/* Wait to RECEIVE the LAST message, immediate data expected */
+	LOGPRINTF("%d Waiting for final inbound RW_imm from peer\n", getpid());
+	if (collect_event(h_dto_rcv_evd,
+			  &event,
+			  DTO_TIMEOUT,
+			  &rdma_wr_poll_count) != DAT_SUCCESS)
+		return (DAT_ABORT);
 
-		if (event.event_number != (int)DAT_IB_DTO_EVENT ||
-		    ext_event->type != DAT_IB_RDMA_WRITE_IMMED_DATA ||
-		    ext_event->val.immed.data != 0x7777) {
-			printf("unexpected event 0x%x type 0x%x or idata 0x%x"
-			       ", waiting for RW-IMMED #0x%x\n",
-			       event.event_number, ext_event->type,
-			       ext_event->val.immed.data, DAT_IB_DTO_EVENT);
-			return (DAT_ABORT);
-		}
-		recv_msg_index++;
+	if (event.event_number != (int)DAT_IB_DTO_EVENT ||
+	    ext_event->type != DAT_IB_RDMA_WRITE_IMMED_DATA ||
+	    ext_event->val.immed.data != 0x7777) {
+		printf("unexpected event 0x%x type 0x%x or idata 0x%x"
+		       ", waiting for RW-IMMED #0x%x\n",
+		       event.event_number, ext_event->type,
+		       ext_event->val.immed.data, DAT_IB_DTO_EVENT);
+		return (DAT_ABORT);
+	}
+	recv_msg_index++;
 
-		/* Send last message received ACK message back */
-		cookie.as_64 = 0x9999;
-		ret = send_msg(p_rmr_snd,
-			       sizeof(DAT_RMR_TRIPLET),
-			       lmr_context_send_msg,
-			       cookie, DAT_COMPLETION_SUPPRESS_FLAG);
-
-		if (ret != DAT_SUCCESS) {
-			fprintf(stderr, "%d Error send_msg: %s\n",
-				getpid(), DT_RetToStr(ret));
-			return (ret);
-		} else {
-			LOGPRINTF("%d send_msg completed\n", getpid());
-		}
+	if (server)
+		goto rmsg;
+smsg:
+	printf("%d sending LAST msg ACK to remote\n", getpid());
+	/* Send last message received ACK message back */
+	cookie.as_64 = 0x9999;
+	ret = send_msg(p_rmr_snd,
+		       sizeof(DAT_RMR_TRIPLET),
+		       lmr_context_send_msg,
+		       cookie, DAT_COMPLETION_SUPPRESS_FLAG);
 
+	if (ret != DAT_SUCCESS) {
+		fprintf(stderr, "%d Error send_msg: %s\n",
+			getpid(), DT_RetToStr(ret));
+		return (ret);
+	} else {
+		LOGPRINTF("%d send_msg completed\n", getpid());
 	}
 
-	if (!uni_direction || !server) {
-		/* Wait for my LAST message ACK from remote side */
-		printf("%d waiting for LAST msg ACK from remote\n", getpid());
-		if (collect_event(h_dto_rcv_evd,
-				  &event,
-				  DTO_TIMEOUT,
-				  &rdma_wr_poll_count) != DAT_SUCCESS)
-			return (DAT_ABORT);
+	if (server)
+		goto acked;
+rmsg:
+	/* Wait for my LAST message ACK from remote side */
+	printf("%d waiting for LAST msg ACK from remote\n", getpid());
+	if (collect_event(h_dto_rcv_evd,
+			  &event,
+			  DTO_TIMEOUT,
+			  &rdma_wr_poll_count) != DAT_SUCCESS)
+		return (DAT_ABORT);
 
-		printf("%d LAST rdma write ACK message arrived!\n", getpid());
-		if (event.event_number != DAT_DTO_COMPLETION_EVENT) {
-			fprintf(stderr, "%d Error unexpected DTO event : %s\n",
-				getpid(), DT_EventToStr(event.event_number));
-			return (DAT_ABORT);
-		}
-
-		if ((dto_event->transfered_length != sizeof(DAT_RMR_TRIPLET))
-		    || (dto_event->user_cookie.as_64 != recv_msg_index)) {
-			fprintf(stderr,
-				"unexpected event data for receive: len=%d "
-				"cookie=" F64x" exp %d/%d\n",
-				(int)dto_event->transfered_length,
-				dto_event->user_cookie.as_64,
-				(int)sizeof(DAT_RMR_TRIPLET), recv_msg_index);
-			return (DAT_ABORT);
-		}
-		printf("%d LAST RDMA_WRITE ACK from remote \n", getpid());
+	printf("%d LAST rdma write ACK message arrived!\n", getpid());
+	if (event.event_number != DAT_DTO_COMPLETION_EVENT) {
+		fprintf(stderr, "%d Error unexpected DTO event : %s\n",
+			getpid(), DT_EventToStr(event.event_number));
+		return (DAT_ABORT);
 	}
 
+	if ((dto_event->transfered_length != sizeof(DAT_RMR_TRIPLET))
+	    || (dto_event->user_cookie.as_64 != recv_msg_index)) {
+		fprintf(stderr,
+			"unexpected event data for receive: len=%d "
+			"cookie=" F64x" exp %d/%d\n",
+			(int)dto_event->transfered_length,
+			dto_event->user_cookie.as_64,
+			(int)sizeof(DAT_RMR_TRIPLET), recv_msg_index);
+		return (DAT_ABORT);
+	}
+	printf("%d LAST RDMA_WRITE ACK from remote \n", getpid());
+
+	if (server)
+		goto smsg;
+acked:
 	stop = get_time();
 	ts.rdma_wr = ((stop - start) * 1.0e6);
 
@@ -1870,6 +2040,181 @@ done:
 	return (DAT_SUCCESS);
 }
 
+#define PAT_NUM 5
+unsigned char pat[PAT_NUM] = { 0, 0xff, 0x55, 0xaa, 0 };
+
+void set_pat(unsigned int len, unsigned int pat_num)
+{
+	if (len <= 1)
+		return;
+
+	if (pat_num >= PAT_NUM) {
+		printf("\n\tpat_num = %d. max valid number is %d.\n\n", pat_num, PAT_NUM - 1);
+		exit(1);
+	}
+
+	if (server) {
+		/* server */
+		if (pat_num == PAT_NUM - 1) {
+			/* future: random data, add checksum */
+			;
+		} else {
+			/* check first byte only for some speed */
+			if ((unsigned char)rbuf[0] != (unsigned char)pat[pat_num]) {
+				fprintf(stderr,"%d: ERR: message len is %d,"
+						" location 0. Rx 0x%x expected"
+						" 0x%x, pat %d\n",
+						getpid(), len, (unsigned char)rbuf[0],
+						(unsigned char)pat[pat_num], pat_num);
+			}
+		}
+		memcpy(sbuf, rbuf, len - 1);
+
+	} else {
+		/* client */
+		int i;
+
+		if (pat_num == PAT_NUM - 1) { /* set random values */
+			struct timeval tv;
+
+			gettimeofday(&tv, NULL);
+			srand((unsigned int)tv.tv_usec);
+			for (i = 0; i < len - 1; i++)
+				sbuf[i] = (unsigned char)rand();
+		} else {
+			memset(sbuf, (unsigned char)pat[pat_num], len - 1);
+		}
+	}
+}
+
+
+/* always uni-direction */
+DAT_RETURN do_rdma_write_ping_pong(int p2, int bytes)
+{
+	DAT_EVENT event;
+	DAT_LMR_TRIPLET l_iov[MSG_IOV_COUNT];
+	DAT_RMR_TRIPLET r_iov;
+	DAT_DTO_COOKIE cookie;
+	DAT_RETURN ret;
+	int i, len, suppress = DAT_COMPLETION_SUPPRESS_FLAG;
+	DAT_DTO_COMPLETION_EVENT_DATA *dto_event =
+		&event.event_data.dto_completion_event_data;
+	volatile char *tx_buf, *rx_buf;
+	uint32_t rx_cnt = 0;
+	uint32_t tx_cnt = 0;
+	unsigned char rx_idx = 0;
+
+	len = bytes ? bytes : 1 << p2;
+
+	tx_buf = (char*)&sbuf[len-1];
+	rx_buf = (char*)&rbuf[len-1];
+
+	/* RMR information from previously received message */
+	r_iov = p_rmr_rcv[recv_msg_index - 1];
+
+	for (i = 0; i < MSG_IOV_COUNT; i++) {
+		l_iov[i].lmr_context = lmr_context_send;
+		l_iov[i].segment_length = len / MSG_IOV_COUNT;
+		l_iov[i].virtual_address = (DAT_VADDR) (uintptr_t)
+					   (&sbuf[l_iov[i].segment_length*i]);
+		LOGPRINTF("%d rdma_write iov[%d] buf=%p,len=%d\n",
+			  getpid(), i,
+			  &sbuf[l_iov[i].segment_length * i],
+			  l_iov[i].segment_length);
+	}
+	start = get_time();
+	for (i = 0; i <= burst; i++) {
+		if (rx_cnt < burst && !(!server && !tx_cnt)) {
+			rx_cnt++;
+			while (*rx_buf != (char)rx_cnt);
+			rx_idx = (unsigned char)*rx_buf;
+
+			if (data_check && !server && memcmp(sbuf, rbuf, len)) {
+				int l=0, ll;
+				fprintf(stderr, "%d: ERR: Tx data from server wrong\n", getpid());
+
+				while (sbuf[l] == rbuf[l] && l < len)
+					l++;
+
+				fprintf(stderr,"%d: len %d, 1st error at %d. Tx 0x%x Rx 0x%x\n",
+						getpid(), len, l, (unsigned char)sbuf[l],
+						(unsigned char)rbuf[l]);
+				fprintf(stderr,"%d: rcnt %d (char = %d), tcnt %d, *rbuf %d\n",
+						getpid(), rx_cnt, (char)rx_cnt, tx_cnt,
+						(unsigned char)*rx_buf);
+				fprintf(stderr, "Send:");
+
+				for (ll=l; ll < len && ll < 1 + 64; ll++)
+					fprintf(stderr,"%02x", (unsigned char)sbuf[ll]);
+
+				fprintf(stderr, "\nRecv:");
+
+				for (ll=l; ll < len && ll < 1 + 64; ll++)
+					fprintf(stderr,"%02x", (unsigned char)rbuf[ll]);
+
+				fprintf(stderr, "\n");
+				return (DAT_ABORT);
+			}
+		}
+
+		if (!((i+1) % signal_rate))
+			suppress =  DAT_COMPLETION_DEFAULT_FLAG;
+		else
+			suppress = DAT_COMPLETION_SUPPRESS_FLAG;
+
+		if (tx_cnt == burst)
+			break;
+
+		if (data_check)
+			set_pat(len, tx_cnt % PAT_NUM);
+
+		*tx_buf = (char)++tx_cnt;
+		cookie.as_64 = tx_cnt;
+		ret = dat_ep_post_rdma_write(h_ep, MSG_IOV_COUNT,
+					     l_iov, cookie, &r_iov,
+					     suppress);
+		if (ret) {
+			fprintf(stderr, "%d: ERROR: dat_rdma_write() %s\n",
+					getpid(), DT_RetToStr(ret));
+			return (DAT_ABORT);
+		}
+		if (!suppress) {
+			while (dat_evd_dequeue(h_dto_req_evd, &event));
+			if (dto_event->status) {
+				fprintf(stderr,
+					"ERROR rdma_write: status=0x%x ck="
+					" "F64x " exp 0x%x\n",
+					dto_event->status,
+					dto_event->user_cookie.as_64, tx_cnt);
+				return (DAT_ABORT);
+			}
+		}
+		LOGPRINTF("%d %s RW pingpong: %p, *rbuf %d rcnt %d\n",
+			  getpid(), server ? "SERVER:" : "CLIENT:",
+			  rx_buf, (unsigned char)*rx_buf,
+			  (unsigned char)rx_cnt);
+	}
+	stop = get_time();
+	ts.rtt = ((stop - start) * 1.0e6);
+
+	if (rx_idx != (unsigned char)rx_cnt) {
+		printf( "%d %s RW pingpong: %p, last *buf %d != cnt %d\n",
+			getpid(), server ? "SERVER:" : "CLIENT:",
+			rx_buf, (unsigned char)*rx_buf,
+			(unsigned char)rx_cnt);
+		return (DAT_ABORT);
+	}
+
+	if (all_data_sizes) {
+		printf( "%d: RDMA write PingPong: (%d x %d) Total %6.2lf us:"
+			" latency %3.2lf us, BW %4.2lf MB/s\n",
+			getpid(), burst, len, ts.rtt, ts.rtt/burst/2,
+			(double)(1/(ts.rtt/burst/2/len)));
+	}
+
+	return (DAT_SUCCESS);
+}
+
 DAT_RETURN do_rdma_read_with_msg(void)
 {
 	DAT_EVENT event;
@@ -2056,15 +2401,15 @@ DAT_RETURN do_ping_pong_msg()
 	rcv_buf = rbuf;
 
 	/* pre-post all buffers */
-	for (i = 0; i < burst; i++) {
+	for (i = 0; i < msg_burst; i++) {
 		burst_msg_posted++;
 		cookie.as_64 = i;
 		l_iov.lmr_context = lmr_context_recv;
 		l_iov.virtual_address = (DAT_VADDR) (uintptr_t) rcv_buf;
 		l_iov.segment_length = buf_len;
 
-		LOGPRINTF("%d Pre-posting Receive Message Buffers %p\n",
-			  getpid(), rcv_buf);
+		LOGPRINTF("%d Pre-posting Receive Message Buffer[%d] %p\n",
+			  getpid(), i, rcv_buf);
 
 		ret = dat_ep_post_recv(h_ep,
 				       1,
@@ -2096,14 +2441,14 @@ DAT_RETURN do_ping_pong_msg()
 
 	/* client ping 0x55, server pong 0xAA in first byte */
 	start = get_time();
-	for (i = 0; i < burst; i++) {
+	for (i = 0; i < msg_burst; i++) {
 		/* walk the send and recv buffers */
 		if (!server) {
 			*snd_buf = 0x55;
 
 			LOGPRINTF("%d %s SND buffer %p contains: 0x%x len=%d\n",
 				  getpid(), server ? "SERVER:" : "CLIENT:",
-				  snd_buf, *snd_buf, buf_len);
+				  snd_buf, *(unsigned char *)snd_buf, buf_len);
 
 			ret = send_msg(snd_buf,
 				       buf_len,
@@ -2120,6 +2465,7 @@ DAT_RETURN do_ping_pong_msg()
 		}
 
 		/* recv message, send completions suppressed */
+		event.event_number = 0;
 		if (collect_event(h_dto_rcv_evd, 
 				  &event, 
 				  DTO_TIMEOUT, 
@@ -2134,8 +2480,9 @@ DAT_RETURN do_ping_pong_msg()
 		/* validate event number and status */
 		LOGPRINTF("%d inbound message; message arrived!\n", getpid());
 		if (event.event_number != DAT_DTO_COMPLETION_EVENT) {
-			fprintf(stderr, "%d Error unexpected DTO event : %s\n",
-				getpid(), DT_EventToStr(event.event_number));
+			fprintf(stderr, "%d Error DTO event (0x%x): %s\n",
+				getpid(), event.event_number,
+				DT_EventToStr(event.event_number));
 			return (DAT_ABORT);
 		}
 		if ((event.event_data.dto_completion_event_data.
@@ -2154,9 +2501,9 @@ DAT_RETURN do_ping_pong_msg()
 			return (DAT_ABORT);
 		}
 
-		LOGPRINTF("%d %s RCV buffer %p contains: 0x%x len=%d\n",
+		LOGPRINTF("%d %s RCV buffer[%d] %p contains: 0x%x len=%d\n",
 			  getpid(), server ? "SERVER:" : "CLIENT:",
-			  rcv_buf, *rcv_buf, buf_len);
+			  i, rcv_buf, *(unsigned char *)rcv_buf, buf_len);
 
 		burst_msg_index++;
 
@@ -2164,9 +2511,9 @@ DAT_RETURN do_ping_pong_msg()
 		if (server) {
 			*snd_buf = 0xaa;
 
-			LOGPRINTF("%d %s SND buffer %p contains: 0x%x len=%d\n",
+			LOGPRINTF("%d %s SND buffer[%d] %p contains: 0x%x len=%d\n",
 				  getpid(), server ? "SERVER:" : "CLIENT:",
-				  snd_buf, *snd_buf, buf_len);
+				  i, snd_buf, *(unsigned char *)snd_buf, buf_len);
 
 			ret = send_msg(snd_buf,
 				       buf_len,
@@ -2203,7 +2550,7 @@ DAT_RETURN register_rdma_memory(void)
 	ret = dat_lmr_create(h_ia,
 			     DAT_MEM_TYPE_VIRTUAL,
 			     region,
-			     buf_len * (burst+1),
+			     buf_len * rq_cnt,
 			     h_pz,
 			     DAT_MEM_PRIV_ALL_FLAG,
 			     DAT_VA_TYPE_VA,
@@ -2230,7 +2577,7 @@ DAT_RETURN register_rdma_memory(void)
 	ret = dat_lmr_create(h_ia,
 			     DAT_MEM_TYPE_VIRTUAL,
 			     region,
-			     buf_len * (burst + 1),
+			     buf_len * rq_cnt,
 			     h_pz,
 			     DAT_MEM_PRIV_ALL_FLAG,
 			     DAT_VA_TYPE_VA,
@@ -2552,17 +2899,22 @@ void print_usage(void)
 	printf("\n DAPL USAGE \n\n");
 	printf("s: server\n");
 	printf("u: unidirectional bandwidth (default=bidirectional\n");
-	printf("w: rdma write only\n");
+	printf("w: rdma write only, streaming\n");
+	printf("W: rdma write only, ping pong\n");
+	printf("D: validate data in ping pong test\n");
 	printf("t: performance times\n");
 	printf("c: use cno\n");
+	printf("a: all data sizes with rdma write pingpong \n");
+	printf("i: increment size for all data size option\n");
 	printf("v: verbose\n");
 	printf("p: polling\n");
 	printf("d: delay before accept\n");
-	printf("b: buf length to allocate\n");
+	printf("b: buf length, upper bound for -W -a -i (WR_pp, all sizes, increment)\n");
 	printf("B: burst count, rdma and msgs \n");
 	printf("h: hostname/address of server, specified on client\n");
 	printf("P: provider name (default = ofa-v2-mlx4_0-1u)\n");
 	printf("S: signal_rate (default=10, completion every 10 iterations\n");
+	printf("U: print this Usage page\n");
 	printf("\n");
 }
 
